/*
*
* versuch project
* AAA module
* $Id: Aaa.cxx,v 1.6 2006/05/19 15:21:25 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#include "Aaa.h"

#include "radlib_vs.h"
#include "Utils.h"
#include "AddrUtils.h"

#include <assert.h>
#include <fstream>
#include <pcre++.h>

TAuth::TAuth(TGlobal * menv)
{
 env = menv;
}

TAuth::~TAuth()
{
// cleanup replies table
 map<string, TCiscoCommandCode*>::iterator it;
 for (it = Replies.begin(); it != Replies.end(); ++it)
 {
   delete it->second;
 }
// close radius handle
 if ( env -> AuthType == "radius" )
  rad_close(RadHandle);
}

void TAuth::InitRadius(const char * Host, int Port, const char * Secret)
{
 if (env -> AuthType != "radius")
 {
   return;
 }
 RadHandle = rad_auth_open();
 if (RadHandle == NULL)
 {
    cerr << endl << "rad_acct_open fails." << endl;
    assert (false);
 }
 if ( rad_add_server(RadHandle, Host, Port , Secret, 5, 2) != 0 )
 {
     cerr << endl << "rad_add_server fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
     assert (false);
 }
 return;
}

void TAuth::FetchPeers(const string & CallingNumber, const string & CalledNumber)
{
// do an actual copy
  WTRACE(1, "Fetching for Calling " << CallingNumber << ", Called " << CalledNumber );
  map<string, TVoicePeer*>::iterator it;
  for (it = env -> DialPeers->List.begin(); it != env -> DialPeers->List.end(); ++it )
  {
    pcrepp::Pcre spDest(it->second->DestPattern);
    if ( spDest.search(CalledNumber) )
    {
       WTRACE(1, "Peer " << it->first << " select by DestPattern " << it->second->DestPattern );
       it->second->Key = it->first;
       PlainPeers.push_back(it->second);
       continue;
    }
    if ( ! CallingNumber.empty() )
    {
      pcrepp::Pcre spSrc(it->second->SrcPattern);
      if ( spSrc.search(CallingNumber) )
      {
        WTRACE(1, "Peer " << it->first << " select by SrcPattern " << it->second->SrcPattern );
        it->second->Key = it->first;
        PlainPeers.push_back(it->second);
        continue;
      }
    } // If CallingNumber Exist
  } // for
// sorting by preference
  sort(PlainPeers.begin(), PlainPeers.end(), RuntimePeerCmp(CalledNumber) );
// huntstop proceed
  vector<TVoicePeer*>::iterator i;
  for (i = PlainPeers.begin(); i != PlainPeers.end(); ++i)
  {
    if ( (*i) -> HuntStop == 1 )
    {
	++i;
	PlainPeers.erase(i, PlainPeers.end());
	break;
    }
  }
}

bool TAuth::FirstAuthorize(WTCPSocket::Address PeerAddri, const string CallingNumber, const string CalledNumber, 
  const string ConferenceID)
{
 WTRACE(1, "FirstAuth. " << PeerAddri << " Calling:" << CallingNumber << " Called: " << CalledNumber );
 if (env -> AuthType == "text")
 {
   FetchPeers(CallingNumber, CalledNumber);
   if (PlainPeers.size() == 0)
     return false;
   unsigned int i;
   for ( i = 0; i < PlainPeers.size(); i++)
   {
     WTRACE(1, "Proceed peer " << PlainPeers[i] -> Key);
     Routes.push_back(PlainPeers[i] -> TargetGW);
     TCiscoCommandCode *Reply = new TCiscoCommandCode();
     Reply -> MediaProxy = ( PlainPeers[i] -> Proxy == 1 ? true : false);
     if (PlainPeers[i] -> src_translate != "")
	Reply -> CallingNumber =  AddrUtils::PerlTranslate( PlainPeers[i] -> src_translate, CallingNumber);
     else
	Reply -> CallingNumber =  CallingNumber;
     if (PlainPeers[i] -> dst_translate != "")
	Reply -> CalledNumber = AddrUtils::PerlTranslate( PlainPeers[i] -> dst_translate, CalledNumber);
     else
	Reply -> CalledNumber = CalledNumber;
     if (PlainPeers[i] -> src_bill_translate != "" && CallingNumber != "")
        Reply -> BCallingNumber = AddrUtils::PerlTranslate( PlainPeers[i] -> src_bill_translate, CallingNumber);
     else
	Reply -> BCallingNumber = CallingNumber;
     if (PlainPeers[i] -> dst_bill_translate != "")
	Reply -> BCalledNumber = AddrUtils::PerlTranslate( PlainPeers[i] -> dst_bill_translate, CalledNumber);
     else
	Reply -> BCalledNumber = CalledNumber;
     Replies[PlainPeers[i]->TargetGW] = Reply;
   }
   return true;
 }
 int RadResponse;
 const void *RadAttrData;
 size_t RadAttrLen;
 char * s;
 u_int32_t vendor;
 if ( rad_create_request(RadHandle, RAD_ACCESS_REQUEST  ) != 0)
 {
    cerr << endl << "rad_create_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
 }
 rad_put_string(RadHandle, RAD_USER_NAME, PeerAddri.AsString());
 rad_put_string(RadHandle, RAD_USER_PASSWORD, env->AuthVSAPrefix.c_str() );
 rad_put_addr(RadHandle, RAD_NAS_IP_ADDRESS, (in_addr)WSocket::Address("0.0.0.0"));
 rad_put_int(RadHandle, RAD_NAS_PORT_TYPE, RAD_ASYNC);
 rad_put_int(RadHandle, RAD_SERVICE_TYPE, RAD_LOGIN);
 rad_put_string(RadHandle, RAD_CALLED_STATION_ID, CalledNumber.c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"request-type=route").c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"routing-request=1").c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONF_ID, string("h323-conf-id="+ConferenceID).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string("h323-call-id="+ConferenceID).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_GW_ID, string("h323-gw-id="+string(PeerAddri.AsString()) ).c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string("h323-gw-address="+string(PeerAddri.AsString()) ).c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"dst-number-in="+CalledNumber).c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"dst-number-out="+
	 CalledNumber).c_str());
 if (CallingNumber != "" )
 {
  rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"src-number-in="+
	  CallingNumber).c_str()  );
  rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"src-number-out="+
	  CallingNumber).c_str()  );
 }
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(env->AuthVSAPrefix+"route-retries=1").c_str()  );
 RadResponse = rad_send_request(RadHandle);
 if( RadResponse == -1 ) 
 {
    cerr << endl << "FirstAuthorize: rad_send_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
    return false;
 }
 if ( RadResponse != RAD_ACCESS_ACCEPT )
 {
    cerr << endl << "Radius authorize response is negative. RadResponse = " << RadResponse << endl;
    return false;
 }
 while ( (RadResponse = rad_get_attr(RadHandle, & RadAttrData, & RadAttrLen )) > 0 )
 {
   if ( RadResponse  == RAD_VENDOR_SPECIFIC ) 
   {
      rad_get_vendor_attr(& vendor, & RadAttrData, & RadAttrLen );
      s = rad_cvt_string( RadAttrData, RadAttrLen);
      if ( s == NULL )
      {
	 cerr << endl << "rad_cvt_string fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
      }
      WTRACE( 1, "Rad Replay got. Attribute is " << s );
      vector<string> tokens;
      tokens.clear();
      Tokenize(s, tokens, ":/,=", true);
      if (tokens[0] == "kill")
      {
	WTRACE(1, "Kill attribute got");
	KillList.push_back(tokens[1]);
      }
      else
      {
        Routes.push_back(tokens[0]);
        TCiscoCommandCode *Reply = new TCiscoCommandCode();
        Reply -> MediaProxy = (tokens[1] == "1" ? true : false);
        Reply -> CallingNumber = tokens[2];
        Reply -> CalledNumber  = tokens[3];
        Reply -> BCallingNumber = tokens[4];
        Reply -> BCalledNumber  = tokens[5];
        Replies[tokens[0]] = Reply;
      }
   } // if is vendor specific attribute
 }  // while has attribute
 return true;
}

//
// Accounting block
//
//

static string TimeStamp();

void TCDR::ToThreadCDR(TThreadCDR & ThrCDR)
{
 ThrCDR.CallRef =       CallRef;
 ThrCDR.StartTime =     StartTime;
 ThrCDR.StopTime =      StopTime;
 ThrCDR.IncOctets =     IncOctets;
 ThrCDR.OutOctets =     OutOctets;
 ThrCDR.DisconnectCause = DisconnectCause;
 ThrCDR.InternalCause = InternalCause;
 ThrCDR.FastStart = 	FastStart;
 ThrCDR.Tunneling = 	Tunneling;
 ThrCDR.RouteRetries = 	RouteRetries;
 strncpy(ThrCDR.SetupTime, SetupTime.c_str(), 50);
 strncpy(ThrCDR.ConnectTime, ConnectTime.c_str(), 50);
 strncpy(ThrCDR.DisconnectTime, DisconnectTime.c_str(), 50);
 strncpy(ThrCDR.CallID, CallID.c_str(), 50);
 strncpy(ThrCDR.gwID, gwID.c_str(), 50);
 strncpy(ThrCDR.ConferenceID, ConferenceID.c_str(), 50);
 strncpy(ThrCDR.Codec, Codec.c_str(), 50);
 strncpy(ThrCDR.CalledStationId, CalledStationId.c_str(), 30);
 strncpy(ThrCDR.CallingStationId, CallingStationId.c_str(), 30);
 strncpy(ThrCDR.AcctSessionID, AcctSessionID.c_str(), 40);
 strncpy(ThrCDR.RemoteAddr, RemoteAddr.c_str(), 20);
 strncpy(ThrCDR.LocalAddr, LocalAddr.c_str(), 20);
 strncpy(ThrCDR.BCalled, CalledBill.c_str(), 30);
 strncpy(ThrCDR.BCalling, CallingBill.c_str(), 30);
}

TAcct::TAcct(TGlobal * menv)
{
 env = menv;
 pthread_mutex_init(&mtx, NULL);
 pthread_attr_init(&attr);
 pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
 if ( env-> AcctType == "both" || env-> AcctType == "radius" )
 {
  RadHandle = rad_acct_open();
  if (RadHandle == NULL) 
  {
    cerr << endl << "rad_acct_open fails." << endl;
    assert (false);
  }
  if ( rad_add_server(RadHandle, env->AcctRadHost.c_str(), env->AcctRadPort , env->AcctRadSecret.c_str(), 5, 2) != 0 )
  {
    cerr << endl << "rad_add_server fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
    assert (false);
  }
  return;
 } // Radius init
}

TAcct::~TAcct()
{
 int status;
 unsigned int i;
 for (i = 0; i < threads.size(); i++) 
 {
   if (pthread_join(threads[i], (void **)&status) != 0 )
       WTRACE(0, "Error\terrno from pthread_join radius_thread = " << errno << ". Text = " << strerror(errno) << 
	" . thread_id = " << threads[i]);
 }
 if ( env-> AcctType == "both" || env-> AcctType == "radius" )
 {
   rad_close(RadHandle);
 }
 if (pthread_mutex_destroy(&mtx) != 0)
      cerr << "pthread_mutex_destroy error " << errno << " : " << strerror(errno) << endl;
  pthread_attr_destroy(&attr);
}

void AcctRadRequest(TThreadCDR * CDR, bool isStartReq, bool CallOrigin, struct rad_handle * RadHandle, const TGlobal * env)
{
 stringstream sCallID, sSetupTime, sConnectTime, sDisconnectTime, sOrigin, sRemoteAddr, sDC;
 if ( rad_create_request(RadHandle, RAD_ACCOUNTING_REQUEST) != 0)
 {
    cerr << endl << "rad_create_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
 }
 sCallID << "call-id=" << CDR -> CallID ;
 if (CallOrigin)
   sOrigin << "h323-call-origin=" << "originate" ;
  else
   sOrigin << "h323-call-origin=" << "answer" ;
 if (isStartReq )
 {
    rad_put_int(RadHandle, RAD_ACCT_STATUS_TYPE, RAD_START);
    rad_put_string(RadHandle, RAD_ACCT_SESSION_ID, CDR -> AcctSessionID);
    rad_put_string(RadHandle, RAD_CALLED_STATION_ID, CDR -> CalledStationId);
    rad_put_string(RadHandle, RAD_CALLING_STATION_ID, CDR -> CallingStationId);
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, sCallID.str().c_str());
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_GW_ID, CDR -> gwID);
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONF_ID, CDR -> ConferenceID);
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_ORIGIN, sOrigin.str().c_str() );
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_TYPE, "VoIP" );
    rad_put_string(RadHandle, RAD_USER_NAME, CDR -> CallingStationId);
    rad_put_int(RadHandle, RAD_SERVICE_TYPE, RAD_LOGIN);
 }
 else
 {
    unsigned int SessionTime;
    sSetupTime     << "h323-setup-time=" << CDR -> SetupTime ;
    sConnectTime   << "h323-connect-time =" << CDR -> ConnectTime;
    sDisconnectTime <<"h323-disconnect-time =" << CDR -> DisconnectTime;
    sRemoteAddr << "h323-remote-address=" << CDR -> RemoteAddr;
    sDC << "h323-disconnect-cause=" << CDR -> DisconnectCause ;
    rad_put_int(RadHandle, RAD_ACCT_STATUS_TYPE, RAD_STOP);
    rad_put_string(RadHandle, RAD_ACCT_SESSION_ID, CDR -> AcctSessionID);
    rad_put_string(RadHandle, RAD_CALLED_STATION_ID, CDR -> CalledStationId);
    rad_put_string(RadHandle, RAD_CALLING_STATION_ID, CDR -> CallingStationId);
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, sCallID.str().c_str());
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_SETUP_TIME, sSetupTime.str().c_str() );
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONNECT_TIME, sConnectTime.str().c_str() );
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_DISCONNECT_TIME, sDisconnectTime.str().c_str() );
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_GW_ID, CDR -> gwID);
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONF_ID, CDR -> ConferenceID);
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_ORIGIN, sOrigin.str().c_str() );
    rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_TYPE, "VoIP" );
    rad_put_string(RadHandle, RAD_USER_NAME, CDR -> CallingStationId);
    if ((CDR -> StopTime != 0 ) && (CDR -> StartTime != 0) )
       SessionTime = (CDR -> StopTime) - (CDR -> StartTime);
    else
       SessionTime = 0;
    rad_put_int(RadHandle, RAD_ACCT_SESSION_TIME, SessionTime );
   rad_put_int(RadHandle, RAD_ACCT_INPUT_OCTETS, CDR -> IncOctets);
   rad_put_int(RadHandle, RAD_ACCT_OUTPUT_OCTETS, CDR -> OutOctets);
   rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_REMOTE_ADDRESS, sRemoteAddr.str().c_str() );
   rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_DISCONNECT_CAUSE, sDC.str().c_str() );
 }
 if( rad_send_request(RadHandle) == -1 ) {
    cerr << endl << "AcctRadRequest: rad_send_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
 }
}

const char * GetInternalCause(int Code)
{
 map<int, string> IntCause;
 IntCause[0] = "CallInProgress";
 IntCause[1] = "CallerNormal";
 IntCause[2] = "CalledNormal";
 IntCause[3] = "CallerDrop";
 IntCause[4] = "CalledDrop";
 IntCause[5] = "AuthReject";
 IntCause[6] = "InternalError";
 IntCause[20] = "ForceTermination";
 IntCause[22] = "TimeoutConnectMsg";
 IntCause[23] = "TimeoutAlertingMsg";
 IntCause[24] = "TimeoutMedia";
 IntCause[25] = "NoRouteToDest";
 IntCause[26] = "DestUnreach";
 IntCause[27] = "InvalidTPKTCaller";
 IntCause[28] = "InvalidTPKTCalled";
 IntCause[29] = "InvalidSetupMsg";
 IntCause[30] = "CallLoop";
 IntCause[31] = "H245fail";
 IntCause[32] = "ExceedRouteAttmpts";
 IntCause[33] = "SessionLimit";
 IntCause[34] = "HungCall";
 IntCause[35] = "RestrictedInPeer";
 IntCause[36] = "RestrictedTarget";
 IntCause[37] = "RestrictedCalledN";
 IntCause[38] = "TimeoutRlsCaller";
 IntCause[39] = "TimeoutRlsCalled";
 IntCause[40] = "TimeoutProceeding";
 return IntCause[Code].c_str();
}

void LoggingData(TThreadCDR * inCDR, TThreadCDR * outCDR, char logfile[50])
{
WTRACE(1, "Logging data to " << logfile);
 fstream lf( logfile, fstream::out | fstream::app );
 if ( lf == 0 )
 {
   WTRACE(0, "Error\t Error in openig log file " << logfile);
   return;
 }
 int SessionTime;
 if ( (strlen( outCDR -> ConnectTime ) != 0) && ( ( inCDR -> StopTime > 100 ) && ( inCDR -> StartTime > 100 ) ) )
    SessionTime = (inCDR -> StopTime) - (inCDR -> StartTime);
 else
    SessionTime = 0;
 WTRACE(1, "inCDR -> StartTime = " << inCDR -> StartTime << "; inCDR -> StopTime = " << inCDR -> StopTime );
 lf << TimeStamp() << ", " ;
 lf << "HOST=" << inCDR->gwID << ", DST-NUMBER-IN=" << inCDR->CalledStationId << ", SRC-NUMBER-IN=" << inCDR->CallingStationId ; 
 lf << ", DST-NUMBER-OUT=" << outCDR->CalledStationId << ", SRC-NUMBER-OUT=" << outCDR->CallingStationId;
 lf << ", SRC-NUMBER-BILL=" << outCDR->BCalling << ", DST-NUMBER-BILL=" << outCDR->BCalled;
 lf << ", SRC-IP=" << inCDR->RemoteAddr << ", DST-IP=" << outCDR->RemoteAddr ;
 lf << ", SRC-RTP-IP=" << inCDR->RemoteAddr << ", DST-RTP-IP=" << outCDR->RemoteAddr ;
 lf << ", SRC-USER=" << inCDR->RemoteAddr ;
 lf << ", INITIAL-INCOMING-LOCAL-ADDRESS=" <<  inCDR->LocalAddr << ", OUTGOING-LOCAL-ADDRESS=" << outCDR->LocalAddr;
 lf << ", SETUP-TIME=" << inCDR -> SetupTime << ", CONNECT-TIME=" 
	 << ( SessionTime == 0 ? inCDR -> SetupTime : outCDR -> ConnectTime);
 if ( strlen(inCDR -> DisconnectTime) == 0 )
   lf << ", DISCONNECT-TIME=" << outCDR -> DisconnectTime ;
 else
   lf << ", DISCONNECT-TIME=" << inCDR -> DisconnectTime ;
 if (inCDR -> DisconnectCause > 0 || inCDR -> DisconnectCause < 128 )
   lf << ", DISCONNECT-CODE-Q931=" << inCDR -> DisconnectCause ;
 else
   lf << ", DISCONNECT-CODE-Q931=" << outCDR -> DisconnectCause ;
 lf << ", INTERNALCAUSE=" << GetInternalCause(outCDR->InternalCause) << ", ROUTE-RETRIES=" << outCDR -> RouteRetries;
 lf << ", SRC-BYTES-IN=" << inCDR -> IncOctets << ", DST-BYTES-IN=" << outCDR -> IncOctets;
 lf << ", SRC-BYTES-OUT=" << inCDR -> OutOctets << ", DST-BYTES-OUT=" << outCDR -> OutOctets;
 lf << ", SRC-CODEC=" << inCDR -> Codec << ", DST-CODEC=" << outCDR -> Codec;
 lf << ", CALLID=" << inCDR -> CallID << ", CONFID=" << inCDR -> ConferenceID;
 lf << ", PROXY-MODE=1" << ", ELAPSED-TIME=" << SessionTime;
 lf << ", SOURCE-FASTSTART=" << inCDR -> FastStart << ", DESTINATION-FASTSTART=" << outCDR -> FastStart ;
 lf << ", SOURCE-TUNNELLING=" << inCDR -> Tunneling << ", DESTINATION-TUNNELLING=" << outCDR -> Tunneling ;
 lf << endl;
 lf.close();
}


void RadLoggingData(TThreadCDR * inCDR, TThreadCDR * outCDR,  struct rad_handle * RadHandle, const TGlobal * env)
{
 stringstream sCallID, sSetupTime, sConnectTime, sDisconnectTime, sRemoteAddr, sDC, sRouteRetries;
 if ( rad_create_request(RadHandle, RAD_ACCOUNTING_REQUEST) != 0)
 {
    cerr << endl << "Error\trad_create_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
 }
 int SessionTime;
 if ( (strlen( outCDR -> ConnectTime ) != 0) && ( ( inCDR -> StopTime > 100 ) && ( inCDR -> StartTime > 100 ) ) )
    SessionTime = (inCDR -> StopTime) - (inCDR -> StartTime);
 else
    SessionTime = 0;
 sCallID << "h323-call-id=" << inCDR -> CallID ;
 sSetupTime     << "h323-setup-time=" << inCDR -> SetupTime ;
 if (SessionTime == 0 )
   sConnectTime   << "h323-connect-time=" << outCDR -> SetupTime;
 else
   sConnectTime   << "h323-connect-time=" << outCDR -> ConnectTime;
 if ( strlen(inCDR -> DisconnectTime) != 0 )
   sDisconnectTime <<"h323-disconnect-time=" << inCDR -> DisconnectTime;
 else
   sDisconnectTime <<"h323-disconnect-time=" << outCDR -> DisconnectTime;
 if (inCDR -> DisconnectCause > 0 || inCDR -> DisconnectCause < 128 )
   sDC << "h323-disconnect-cause=" << inCDR -> DisconnectCause ;
 else
   sDC << "h323-disconnect-cause=" << outCDR -> DisconnectCause ;
 sRemoteAddr << "h323-remote-address=" << outCDR -> RemoteAddr;
 rad_put_int(RadHandle, RAD_ACCT_STATUS_TYPE, RAD_STOP);
 rad_put_addr(RadHandle, RAD_NAS_IP_ADDRESS, (in_addr)WSocket::Address("0.0.0.0"));
 rad_put_int(RadHandle, RAD_NAS_PORT_TYPE, RAD_ASYNC);
 rad_put_int(RadHandle, RAD_SERVICE_TYPE, RAD_LOGIN);
 if ( strlen(inCDR -> BCalled ) != 0 )
  rad_put_string(RadHandle, RAD_CALLED_STATION_ID, inCDR -> BCalled);
 else
   rad_put_string(RadHandle, RAD_CALLED_STATION_ID, inCDR -> CalledStationId);
 if ( strlen(inCDR -> BCalling ) != 0 )
   rad_put_string(RadHandle, RAD_CALLING_STATION_ID, outCDR -> BCalling );
 else
   rad_put_string(RadHandle, RAD_CALLING_STATION_ID, outCDR -> CallingStationId);
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_ORIGIN, "h323-call-origin=originate" );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, 
	 string(string("h323-incoming-conf-id=") + inCDR -> CallID).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, 
         string(string("h323-incoming-call-id=") + inCDR -> CallID).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, sCallID.str().c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_SETUP_TIME, sSetupTime.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONNECT_TIME, sConnectTime.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_DISCONNECT_TIME, sDisconnectTime.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_GW_ID, string(string("h323-gw-id=") + inCDR->RemoteAddr).c_str()); 
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(string("h323-gw-address=") + inCDR->RemoteAddr).c_str()); 
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONF_ID, inCDR -> ConferenceID);
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_TYPE, "VoIP" );
 rad_put_string(RadHandle, RAD_USER_NAME, inCDR -> RemoteAddr);
 rad_put_int(RadHandle, RAD_ACCT_SESSION_TIME, SessionTime );
 rad_put_int(RadHandle, RAD_ACCT_DELAY_TIME, 0 );
 rad_put_int(RadHandle, RAD_ACCT_INPUT_OCTETS, inCDR -> IncOctets);
 rad_put_int(RadHandle, RAD_ACCT_OUTPUT_OCTETS, inCDR -> OutOctets);
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_REMOTE_ADDRESS, sRemoteAddr.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(string("h323-remote-id=")+outCDR -> RemoteAddr ).c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_DISCONNECT_CAUSE, sDC.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, 
         string(env->AcctVSAPrefix + "local-disconnect-cause=" + GetInternalCause(outCDR->InternalCause)).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "src-codec=" + inCDR->Codec ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "initial-incoming-local-address=" + inCDR->LocalAddr ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "selected-incoming-local-address=" + inCDR->LocalAddr ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "outgoing-local-address=" + outCDR->LocalAddr ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "source-faststart=" + (inCDR -> FastStart == 0 ? "0" : "1" ) ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "source-tunneling=" + (inCDR -> Tunneling == 0 ? "0" : "1" ) ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "src-number-in=" + inCDR -> CallingStationId ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "src-number-out=" + outCDR -> CallingStationId ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "dst-number-in=" + inCDR -> CalledStationId ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
         string(env->AcctVSAPrefix + "dst-number-out=" + outCDR -> CalledStationId ).c_str());
 sRouteRetries << env->AcctVSAPrefix << "route-retries=" << outCDR -> RouteRetries ;
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, sRouteRetries.str().c_str());

 if( rad_send_request(RadHandle) == -1 ) {
     cerr << endl << "RadLoggingData: rad_send_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
     return;
 }
 WTRACE(1, "RadAccounting stop leg done");
}

void RadTransitStart(TThreadCDR * inCDR, TThreadCDR * outCDR,  struct rad_handle * RadHandle, const TGlobal * env)
{
 stringstream sCallID, sSetupTime, sConnectTime, sRemoteAddr, sDC, sRouteRetries;
 if ( rad_create_request(RadHandle, RAD_ACCOUNTING_REQUEST) != 0)
 {
    cerr << endl << "Error\trad_create_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
 }
 sCallID << "h323-call-id=" << inCDR -> CallID ;
 sSetupTime     << "h323-setup-time=" << inCDR -> SetupTime ;
 if ( strlen(outCDR -> ConnectTime) == 0  )
   sConnectTime   << "h323-connect-time=" << outCDR -> SetupTime;
 else
   sConnectTime   << "h323-connect-time=" << outCDR -> ConnectTime;
 sRemoteAddr << "h323-remote-address=" << outCDR -> RemoteAddr;
 rad_put_int(RadHandle, RAD_ACCT_STATUS_TYPE, RAD_START);
 rad_put_addr(RadHandle, RAD_NAS_IP_ADDRESS, (in_addr)WSocket::Address("0.0.0.0"));
 rad_put_int(RadHandle, RAD_NAS_PORT_TYPE, RAD_ASYNC);
 rad_put_int(RadHandle, RAD_SERVICE_TYPE, RAD_LOGIN);
 if ( strlen(inCDR -> BCalled ) != 0 )
     rad_put_string(RadHandle, RAD_CALLED_STATION_ID, inCDR -> BCalled);
 else
     rad_put_string(RadHandle, RAD_CALLED_STATION_ID, inCDR -> CalledStationId);
 if ( strlen(inCDR -> BCalling ) != 0 )
     rad_put_string(RadHandle, RAD_CALLING_STATION_ID, outCDR -> BCalling );
 else
     rad_put_string(RadHandle, RAD_CALLING_STATION_ID, outCDR -> CallingStationId);
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_ORIGIN, "h323-call-origin=originate" );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, 
 string(string("h323-incoming-conf-id=") + inCDR -> CallID).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, 
    string(string("h323-incoming-call-id=") + inCDR -> CallID).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, sCallID.str().c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_SETUP_TIME, sSetupTime.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONNECT_TIME, sConnectTime.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_GW_ID, string(string("h323-gw-id=") + inCDR->RemoteAddr).c_str()); 
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(string("h323-gw-address=") + inCDR->RemoteAddr).c_str()); 
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CONF_ID, inCDR -> ConferenceID);
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_CALL_TYPE, "VoIP" );
 rad_put_string(RadHandle, RAD_USER_NAME, inCDR -> RemoteAddr);
 rad_put_int(RadHandle, RAD_ACCT_DELAY_TIME, 0 );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, H323_REMOTE_ADDRESS, sRemoteAddr.str().c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, string(string("h323-remote-id=")+outCDR -> RemoteAddr ).c_str() );
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "src-codec=" + inCDR->Codec ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "initial-incoming-local-address=" + inCDR->LocalAddr ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "selected-incoming-local-address=" + inCDR->LocalAddr ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "outgoing-local-address=" + outCDR->LocalAddr ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "source-faststart=" + (inCDR -> FastStart == 0 ? "0" : "1" ) ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "source-tunneling=" + (inCDR -> Tunneling == 0 ? "0" : "1" ) ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "src-number-in=" + inCDR -> CallingStationId ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "src-number-out=" + outCDR -> CallingStationId ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "dst-number-in=" + inCDR -> CalledStationId ).c_str());
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR,
    string(env->AcctVSAPrefix + "dst-number-out=" + outCDR -> CalledStationId ).c_str());
 sRouteRetries << env->AcctVSAPrefix << "route-retries=" << outCDR -> RouteRetries ;
 rad_put_vendor_string(RadHandle, CISCO_VENDOR_ID, CISCO_AVPAIR, sRouteRetries.str().c_str());

 if( rad_send_request(RadHandle) == -1 ) 
 {
    cerr << endl << "Error\tRadTransitStart: rad_send_request fails. rad_strerror = " << rad_strerror(RadHandle) << endl;
    return;
 }
 WTRACE(1, "RadAccounting start leg done");
}

void * AcctThreadFunction( void *ptr )
{
 struct acct_thread_data * tdata;
 tdata = (struct acct_thread_data *) ptr;
 pthread_mutex_lock(tdata -> mtx);
 AcctRadRequest(& (tdata -> CDR), tdata -> isStartReq, tdata -> CallOrigin, tdata -> rh, tdata -> env );
 free(tdata);
 pthread_mutex_unlock(tdata -> mtx);
 pthread_exit(0);
}

void * LogThreadFunction( void *ptr )
{
  struct log_thread_data * logdata;
  logdata = (struct log_thread_data *) ptr;
  pthread_mutex_lock(logdata -> mtx);
  LoggingData(& (logdata -> inCDR), & (logdata -> outCDR), logdata -> log_file);
  free(logdata);
  pthread_mutex_unlock(logdata -> mtx);
  pthread_exit(0);
}

void * RadTransitThreadFunction( void *ptr ) 
{
 struct rad_log_thread_data * rad_logdata;
 rad_logdata = (struct rad_log_thread_data *) ptr;
 pthread_mutex_lock(rad_logdata -> mtx);
 if (rad_logdata -> isStopLeg )
   RadLoggingData(& (rad_logdata -> inCDR), & (rad_logdata -> outCDR), rad_logdata -> rh, rad_logdata -> env );
 else
   RadTransitStart(& (rad_logdata -> inCDR), & (rad_logdata -> outCDR), rad_logdata -> rh, rad_logdata -> env );
 free(rad_logdata);
 pthread_mutex_unlock(rad_logdata -> mtx);
 pthread_exit(0);
}

void TAcct::StartAcct(TCDR & CDR, bool CallOrigin)
{
 if ( env -> RadStopOnly )
    return;
 if (env -> AcctType == "both" || env -> AcctType == "radius" ) 
 {
   if (env -> AcctRadMode == 2 ) // records for both incomming and outgoing legs
   {
     pthread_t AcctThreadID;
     acct_thread_data * acct_tdata;
     acct_tdata = (struct acct_thread_data *) malloc(sizeof(struct acct_thread_data));
     if (acct_tdata == NULL) 
     {
       cerr << "Thread dat allocation fail " << endl;
       return;
     }
     CDR.ToThreadCDR(acct_tdata -> CDR);
     acct_tdata -> CallOrigin = CallOrigin;
     acct_tdata -> rh = RadHandle;
     acct_tdata -> mtx = &mtx;
     acct_tdata -> isStartReq = true;
     acct_tdata -> env = env ;
     if (  pthread_create( &AcctThreadID,  &attr, AcctThreadFunction, (void *) acct_tdata  ) != 0 )
     {
      cerr << endl << "Error in RadStartAcct thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
      return;
     }
     threads.push_back(AcctThreadID);
     return;
   } // if AcctRadMode == 2
 } // if both & Radius
}

void TAcct::StopAcct(TCDR & inCDR, TCDR & outCDR, TCallOrigin CallOrigin)
{
 WTRACE(1, "Acct\tAcctType = " << env -> AcctType << " AcctRadMode = " << env -> AcctRadMode );
 if ( env -> AcctType == "both" || env -> AcctType == "log" )
 {
    pthread_t LogThreadID;
    log_thread_data * log_tdata;
    log_tdata = (struct log_thread_data*) malloc(sizeof(struct log_thread_data));
    if ( log_tdata == NULL)
    {
      cerr << "Thread data allocation fails " << endl;
      return;
    }
    inCDR.ToThreadCDR( log_tdata -> inCDR );
    outCDR.ToThreadCDR( log_tdata -> outCDR );
    log_tdata -> mtx = &(env->global_mtx);
    strncpy(log_tdata -> log_file, env->LogFile.c_str(), 49);
    if (  pthread_create( &LogThreadID,  &attr, LogThreadFunction, (void *) log_tdata  ) != 0 )
    {
      cerr << endl << "Error in Log thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
      return;
    }
    threads.push_back(LogThreadID);
 } // log or both
 if (env -> AcctType == "both" || env -> AcctType == "radius" )
 {
  if (env -> AcctRadMode == 2 ) // record for both incomming and outgoing legs (cisco mode)
  {
    if ( CallOrigin == Both || CallOrigin == Answer )
    {
      pthread_t inAcctThreadID;
      acct_thread_data * in_acct_tdata;
      in_acct_tdata = (struct acct_thread_data *) malloc(sizeof(struct acct_thread_data));
      if (in_acct_tdata == NULL)
      {
         cerr << "Thread data allocation fails " << endl;
         return;
      }
      inCDR.ToThreadCDR( in_acct_tdata -> CDR );
      in_acct_tdata -> CallOrigin = false;
      in_acct_tdata -> rh = RadHandle;
      in_acct_tdata -> mtx = &mtx;
      in_acct_tdata -> isStartReq = false;
      in_acct_tdata -> env = env;
      if (  pthread_create( &inAcctThreadID,  &attr, AcctThreadFunction, (void *) in_acct_tdata  ) != 0 )
      {
         cerr << endl << "Error in RadStartAcct thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
         return;
      }
      threads.push_back(inAcctThreadID);
    } // Both & Answer
    if ( CallOrigin == Both || CallOrigin == Origin )
    {
      pthread_t outAcctThreadID;
      acct_thread_data * out_acct_tdata;
      out_acct_tdata = (struct acct_thread_data *) malloc(sizeof(struct acct_thread_data));
      if (out_acct_tdata == NULL)
      {   
         cerr << "Thread data allocation fails " << endl;
         return;
      }   
      outCDR.ToThreadCDR( out_acct_tdata -> CDR );
      out_acct_tdata -> CallOrigin = true;
      out_acct_tdata -> rh = RadHandle;
      out_acct_tdata -> mtx = &mtx;
      out_acct_tdata -> isStartReq = false;
      out_acct_tdata -> env = env;
      if (  pthread_create( &outAcctThreadID,  &attr, AcctThreadFunction, (void *) out_acct_tdata  ) != 0 )
      {       
        cerr << endl << "Error in RadStartAcct thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
        return;
      }       
      threads.push_back(outAcctThreadID);
    } // Both & Origin
  } // RadMode & cisco mode
  if (env -> AcctRadMode == 1 ) // transit mode
  {
     pthread_t RadLogThreadID;
     rad_log_thread_data * rad_log_tdata;
     rad_log_tdata = (struct rad_log_thread_data*) malloc(sizeof(struct rad_log_thread_data));
     if ( rad_log_tdata == NULL)
     {
          cerr << "Thread data allocation fails (rad_log_tdata)" << endl;
          return; 
     } 
     inCDR.ToThreadCDR( rad_log_tdata -> inCDR );
     outCDR.ToThreadCDR( rad_log_tdata -> outCDR );
     rad_log_tdata -> mtx = &mtx;
     rad_log_tdata -> rh = RadHandle;
     rad_log_tdata -> isStopLeg = true;
     rad_log_tdata -> env = env;
     if (  pthread_create( &RadLogThreadID,  &attr, RadTransitThreadFunction, (void *) rad_log_tdata  ) != 0 )
     {
        cerr << endl << "Error in Log thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
        return;
     }
     threads.push_back(RadLogThreadID);
  } // Radmode & transit mode
 } // Radius or both
 return;
}

void TAcct::StartAcctTransit(TCDR & inCDR, TCDR & outCDR)
{
 if ( (env -> AcctType == "both" || env -> AcctType == "radius" ) && env -> AcctRadMode == 1 )
 {
   pthread_t RadLogThreadID;
   rad_log_thread_data * rad_log_tdata;
   rad_log_tdata = (struct rad_log_thread_data*) malloc(sizeof(struct rad_log_thread_data));
   if ( rad_log_tdata == NULL)
   {
      cerr << "Thread data allocation fails (rad_log_tdata)" << endl;
      return;
   }
   inCDR.ToThreadCDR( rad_log_tdata -> inCDR );
   outCDR.ToThreadCDR( rad_log_tdata -> outCDR );
   rad_log_tdata -> mtx = &mtx;
   rad_log_tdata -> rh = RadHandle;
   rad_log_tdata -> isStopLeg = false;
   rad_log_tdata -> env = env;
   if (  pthread_create( &RadLogThreadID,  &attr, RadTransitThreadFunction, (void *) rad_log_tdata  ) != 0 )
   {
       cerr << endl << "Error in Log thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
       return;
   }
   threads.push_back(RadLogThreadID);
 }
}

const char WeekDay[7][4] = {"Sun", "Mon", "Tue", "Web", "Thu", "Fri", "Sat"};
const char Month[12][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

string TAcct::GetLocalTime()
{
 stringstream sstr;
 struct tm tmclock;
 time_t clock = time(NULL);
 localtime_r(&clock, &tmclock);
 sstr << setfill('0') << tmclock.tm_hour << ":" << setw(2) << tmclock.tm_min << ":" << setw(2) << tmclock.tm_sec
   << ".000" << " " << (tmclock.tm_gmtoff > 0 ? "+" : "-" ) << setw(4) << (tmclock.tm_gmtoff / 3600 * 100) 
   << " " << WeekDay[tmclock.tm_wday] << " " << Month[tmclock.tm_mon]
   << " " << tmclock.tm_mday << " " << tmclock.tm_year+1900 ;
 return sstr.str();
}

static string TimeStamp()
{
 stringstream sstr;
 struct tm tmclock;
 time_t clock = time(NULL);
 localtime_r(&clock, &tmclock);
 sstr << setfill('0') << WeekDay[tmclock.tm_wday] << " " << Month[tmclock.tm_mon] << " " << tmclock.tm_mday << " " 
    << tmclock.tm_hour << ":" << setw(2) << tmclock.tm_min << ":" << setw(2) << tmclock.tm_sec << " " << tmclock.tm_year+1900 ;
 return sstr.str();
}
