/*
*
* versuch project
* Control & Status 
* $Id: ControlThread.cxx,v 1.3 2006/07/27 12:15:34 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#include <sstream>

#include <stdlib.h>
#include <limits.h>

#include "ControlThread.h"
#include "AddrUtils.h"

static char * IntStates[] = { "NoActive", "Signalls", "Establish", "Shutting", "MediaRun" };

void * status_thread_function( void *ptr )
{
 struct T_status_thread_data * thread_data;
 WTRACE(3, "StatusThread\tEntering to status_thread_function. ptr= " << ptr);
 thread_data = (struct T_status_thread_data *) ptr;
 thread_data -> Table -> Add(thread_data -> ch);
 thread_data -> ch -> Main();
 WTRACE(4, "StatusThread\tExiting from status_thread_function");
 thread_data -> Table -> Del(thread_data -> ch);
 delete thread_data -> ch ;
 pthread_exit(0);
}

ControlThread::ControlThread(CallThread * Parent, unsigned int Port)
 : ParentCallThread(Parent), StausPort(Port)
{
 if ( !Listener.Listen( Parent->env.ListenAddr, Port, true ) )
    WTRACE(1, "Listen failed on Status Thread listen socket");
 WTRACE(3, "ControlThread started. Listen started on port " << Port);
 control_tdata = ( struct T_control_thread_data *) malloc (sizeof (struct T_control_thread_data));
}

ControlThread::~ControlThread()
{
 Listener.Close();
 if (control_tdata != NULL) 
 {
   WTRACE(1, "Free pointer " << control_tdata);
   free(control_tdata);
 }
}

void ControlThread::Main()
{
 while ( Listener.IsOpen() )
 {
   Listener.SetReadTimeout(0);
   WTelnetSocket2 * StatusChannel = new WTelnetSocket2;
   if ( StatusChannel->Accept( Listener ) )
   {
     WTCPSocket::Address PeerAddr;
     StatusChannel->GetPeerAddress( PeerAddr );

     WTRACE( 3, "New incoming connection to status port " << PeerAddr );
     pthread_t StatThreadID;
     TStatusThread * StatThread = new TStatusThread( ParentCallThread, StatusChannel, & StatusTable );
     StatThread -> status_tdata -> ch = StatThread;
     StatThread -> status_tdata -> Table = & StatusTable;
     WTRACE(3, "StatThread\tStarting Status thread");
     if ( pthread_create( &StatThreadID,  NULL, status_thread_function, (void *) StatThread -> status_tdata  ) != 0 )
     {
       cerr << endl << "Error in Status thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
       return;
     }
     if ( (pthread_detach(StatThreadID)) != 0)  // there are several reasons
         WTRACE(1, "Status pthread_detach error" << StatThreadID);
   }
   else
   {
     WTRACE( 1, "Error in accepting status connection. errno = " << errno << ". text is " << strerror(errno) );
     delete StatusChannel;
   }
 }
}

TStatusThread::TStatusThread(CallThread * Parent, WTCPSocket *  vSocket, TStatusTable * vTable)
 : ParentCallThread(Parent)
{
 Table = vTable;
 StatSocket = (WTelnetSocket2 *) vSocket;
 status_tdata = ( struct T_status_thread_data *) malloc (sizeof (struct T_status_thread_data));
}

TStatusThread::~TStatusThread()
{
 if (StatSocket != NULL) 
 {
   StatSocket->Close();
   delete StatSocket;
 }
 if (status_tdata != NULL)
 {
   WTRACE(2, "Free pointer " << status_tdata);
   free(status_tdata);
 }
 WTRACE(1,"StatusThread destroy");
}

void TStatusThread::Main()
{
 int sel;
 while ( StatSocket->IsOpen() ) {
  sel = WTCPSocket::Select( *StatSocket, 50 );
  if( sel != -1 )
  {
    WSocket::Address  PeerAddr;
    unsigned          PeerPort;
    StatSocket->GetPeerAddress( PeerAddr, PeerPort );
    WTRACE(1, "Proceed status connection from " << PeerAddr << ":" << PeerPort);
    if (!ReadMsg()) 
      StatSocket -> Close();
  }
  else
  {
    WTRACE(1, "Select Error");
    StatSocket->Close();
    return;
  } // if
 } // while
}

bool TStatusThread::ReadCommand(string& cmd)
{                                       
  string Command, Last3;                        
  int     CharRead;               
  while (true) { 
    CharRead = StatSocket->ReadChar();
    WTRACE(3, "GotChar = " << CharRead << "(" << ")" );
    if ( CharRead < 0 )     
         return false;   
    if ( CharRead == '\n' )         
         break;          
    if( CharRead == '\b' ) // BackSpace
    {
	if (Command.size() > 0 )
          Command.erase(Command.size()-1,1);
        StatSocket -> WriteString("\b \b"); // clear the last character on telnet connection
        continue;       
    } 
    if ( CharRead != '\r' && CharRead != 27) // ignore carriage return and arrow keys
    {
       Command += (char)CharRead;
       StatSocket -> WriteString(string(1, (char)CharRead));
    }
    else { 
       if (CharRead == 27 )
       {
          CharRead = StatSocket->ReadChar();
          if ( CharRead == 91 )
	  {
	    WTRACE(3,"Arrow pressed");
	    CharRead = StatSocket->ReadChar();
	    if (CharRead == 65 ) // Up arrow
	    {
	      if (LastCmd > 1)
              {
                WTRACE(2, "Up Arrow pressed. string = "<< Command << "size = " << Command.size());
                for(unsigned int i = 0; i < Command.size(); i ++)
                {
                  StatSocket -> WriteString("\b \b");
                }
                StatSocket -> WriteString(CommandHistory[LastCmd-1]);
                Command = CommandHistory[LastCmd-1];
                --LastCmd;
	      }
              continue;
	    }
            if (CharRead == 66 ) // Down arrow
            {
               WTRACE(1, "Down Arrow pressed. string = "<< Command << "size = " << Command.size() 
			       << ". lastCmdINdex = " << LastCmd);
	       if (LastCmd+1 < CommandHistory.size() )
	       {
		 for(unsigned int i = 0; i < Command.size(); i ++)
                 {             
                    StatSocket -> WriteString("\b \b");
                 } 
                 WTRACE(3,"Procced down arrow. LastCmd = " << (CommandHistory[LastCmd+1]));
                 StatSocket -> WriteString(CommandHistory[LastCmd+1]);
                 Command = CommandHistory[LastCmd+1];
                 ++LastCmd;
	       }
               continue;
            }
	    StatSocket -> WriteString(string(1, (char)91) + string(1, (char)CharRead));
	  }
	  else
          {
            StatSocket -> WriteString(string(1, (char)CharRead));
          }
	  StatSocket -> WriteString(string(1, (char)27) );
	  continue;
       }
       StatSocket -> WriteString("\r\n");
       break;
    }  // else
  } // while
  cmd = Command;
  if (CommandHistory.size() > 20)
    CommandHistory.erase( CommandHistory.begin() );
  if (CommandHistory.size() == 0 || cmd != CommandHistory[CommandHistory.size()-1])
  {
    CommandHistory.push_back(cmd);
  }
  LastCmd = CommandHistory.size();
  return true;            
} 

bool TStatusThread::ReadMsg()
{
 string Msg;
 if (! ReadCommand ( Msg) )
 {
   StatSocket->Close();
   return false;
 }
 HandleRequest(Msg);
 return true;
}

bool TStatusThread::WriteMsg(string Msg)
{
 Msg = Msg + "\n\r";
 size_t MsgLength = Msg.length();
 WTRACE(2, "Writing message: '" << Msg << "' :" << MsgLength);
 if  ( !StatSocket->Write(Msg.c_str(), MsgLength) ) {
   WTRACE(1, "Error in write Message into status channel");
   return false;
 }
 return true;
}

bool TStatusThread::HandleRequest(string Mess)
{
 WTRACE(1, "Handle Status Request:" << Mess << ":" << Mess.length());
 ostringstream OutMsg;
 string GWaddr;
 int CallRef, TraceLevel, i;
 if (Mess == "q" || Mess == "Q" )
 {
   WTRACE(1,"Close connection");
   StatSocket->Close();
   return false;
 }
 if (Mess == "counters") {
   OutMsg << ParentCallThread->counters.attempts << ":" << ParentCallThread->counters.calls << ":" << ParentCallThread->counters.wm;
   WriteMsg(OutMsg.str());
   return true;
 }
 if (Mess == "uptime") {
   OutMsg << GetUptime();
   WriteMsg(OutMsg.str());
   return true;
 }
 if (Mess == "list") {
   OutMsg << ParentCallThread->ClTbl->ListCalls();
   WriteMsg(OutMsg.str());
   return true;
 }
 if (Mess == "showlist") {
    OutMsg << ParentCallThread->ClTbl->ColumnListCalls();
    WriteMsg(OutMsg.str());
    return true; 
 }
 if (Mess == "reloadconfig") {
    if ( ParentCallThread -> env.ReadFile()  )
      OutMsg << "reload succeful";
    else
      OutMsg << "config reload fails";
    WriteMsg(OutMsg.str());
    return true;
 }
 if (Mess == "gwusage") {
   OutMsg << ParentCallThread->ClTbl->GWusage();
   WriteMsg(OutMsg.str());
   return true;
 }
 CallRef = IsDetailRequest(Mess);
 if (CallRef != 0) {
   OutMsg << ParentCallThread->ClTbl->CallDetails(CallRef);
   WriteMsg(OutMsg.str());
   return true;
 }
 CallRef = IsKillRequest(Mess);
 if (CallRef != 0) {
   OutMsg << ParentCallThread->ClTbl->KillCall(CallRef);
   WriteMsg(OutMsg.str());
   return true;
 }
 CallRef = IsShutdownRequest(Mess);
 if (CallRef != 0) {
   OutMsg << ParentCallThread->ClTbl->ShutdownCall(CallRef);
   WriteMsg(OutMsg.str()); 
   return true;
 }
 GWaddr = IsGWstatusRequest(Mess);
 if ( ! GWaddr.empty() ) {
   OutMsg << ParentCallThread->ClTbl->GWstatus(GWaddr);
   WriteMsg(OutMsg.str());
   return true;
 }
 GWaddr = IsGWcountersRequest(Mess);
 if ( ! GWaddr.empty() ) {
   OutMsg << ParentCallThread->ClTbl->GWcounters(GWaddr);
   WriteMsg(OutMsg.str());
   return true;
 }
 TraceLevel = IsTraceLevelRequest(Mess);
 if ( TraceLevel >= 0 ) {
   i = WTrace::GetLevel();
   OutMsg << "Old level is " << i;
   WTrace::SetLevel(TraceLevel);
   i = WTrace::GetLevel();
   OutMsg << ": NewLevel is " << i;
   WriteMsg(OutMsg.str());
   return true;
 }
 if (Mess == "who" || Mess == "w") {
   WriteMsg(Table->who());
   return true;
 }
 OutMsg << "Uknown request :" << Mess << "\r\n\r\n"
	<< "Try use:" << "\r\n" 
	<< "showlist\t\t"	<< "list of active calls in human readable form" << "\r\n"
	<< "reloadconfig\t\t"	<< "reread config files" << "\r\n"
	<< "q/Q\t\t\t"		<< "close console connection" << "\r\n"
	<< "counters\t\t"	<< "counters of attempts, calls and watermark" << "\r\n"
	<< "uptime\t\t\t"	<< "uptime in seconds" << "\r\n"
	<< "list\t\t\t" 	<< "list of active calls" << "\r\n"
	<< "details:CallRef\t\t"	<< "details of call" << "\r\n"
	<< "gwusage\t\t\t"	<< "info about every gateway usage" << "\r\n"
	<< "kill:CallRef\t\t"	<< "kill call by force to close sockets" << "\r\n"
	<< "shutdown:CallRef\t"	<< "shutdown call with correct messanging" << "\r\n"
	<< "gwstatus:IP\t\t"	<< "request for staus about the gateway" << "\r\n"
	<< "gwcounters:IP\t\t"	<< "request for counters about the gateway" << "\r\n"
	<< "TraceLevel:level\t"	<< "set/get trace level" << "\r\n" << "\r\n";
 WriteMsg(OutMsg.str());
 return false;
}

unsigned TStatusThread::IsDetailRequest(string Msg)
{
 vector<string> tokens;
 Tokenize(Msg, tokens, ",.:-/'");
 if (tokens.size() >= 2 && tokens[0] == "details")
   return strtoll(tokens[1].c_str(), NULL,10);
 return 0;
}

string TStatusThread::IsGWstatusRequest(string Msg)
{
 vector<string> tokens;
 Tokenize(Msg, tokens, ",.:-/'");
 if (tokens.size() >= 2 && tokens[0] == "gwstatus")
   return tokens[1];
 return "";
}

string TStatusThread::IsGWcountersRequest(string Msg)
{
 vector<string> tokens;
 Tokenize(Msg, tokens, ",.:-/'");
 if (tokens.size() >= 2 && tokens[0] == "gwcounters")
   return tokens[1];
 return "";
}

unsigned TStatusThread::IsKillRequest(string Msg)
{
 vector<string> tokens;
 Tokenize(Msg, tokens, ",.:-/'");
 if (tokens.size() >= 2 && tokens[0] == "kill")
   return strtoll(tokens[1].c_str(), NULL,10);
 return 0;
}

unsigned TStatusThread::IsShutdownRequest(string Msg)
{
 vector<string> tokens;
 Tokenize(Msg, tokens, ",.:-/'");
 if (tokens.size() >= 2 && tokens[0] == "shutdown")
     return strtoll(tokens[1].c_str(), NULL,10);
 return 0;
}

int TStatusThread::IsTraceLevelRequest(string Msg)
{
 vector<string> tokens;
 Tokenize(Msg, tokens, ",.:-/'");
 if (tokens.size() >= 2 && tokens[0] == "TraceLevel")
   return strtoll(tokens[1].c_str(), NULL,10);
 return -1;
}

unsigned TStatusThread::GetUptime()
{
 struct timezone tmpTZ;
 struct timeval NowTime;
 gettimeofday(&NowTime, &tmpTZ);
 return (NowTime.tv_sec - ParentCallThread->ssStartTime.tv_sec);
}


CallTable::CallTable()
{
 mList.clear();
}

void CallTable::AddCall(SignallingThread * SThread)
{
 mList.push_back(SThread);
}

void CallTable::DelCall(SignallingThread * SThread)
{
 SigItr it;
 it = find(mList.begin(),mList.end(), SThread);
 if (it == mList.end()) 
 {
  WTRACE (1, "Error\tWrong finding SignallingThread in list") ;
  return ;
 }
 mList.erase(it);
}

string CallTable::ListCalls()
{
 SignallingThread * S;
 ostringstream res, res1;
 map<unsigned, unsigned> st;
 map<unsigned, unsigned>::iterator it;
 unsigned int i, j, Now =  time(NULL);
 j=0;
 for (i = 0; i < mList.size() ; i++) {
   S = mList[i];
   st = S->CallPool->StartTime;
   for (it = st.begin(); it != st.end() ; it++) {
     res << it->first << "," << S->IncAddr <<  "," << S->CallPool->OutAddr[it->first] << "," 
	     << S->CallPool->CalledNumber[it->first] << ","
        << (Now - (S->CallPool->StartTime[it->first])) << "," << S->CallPool->cStatesList[it->first] << ","
	<< S->CallPool->InternTC[it->first] << "," << GetNumber(S) << ":" ;
     j++;
   }  // in CallPool
 }  // per Thread
 res1 << j << "|" << res.str();
 return res1.str();
}

int CallTable::GetUsage(WSocket::Address Addr)
{
 int Cap = 0;
 unsigned int i;
 SignallingThread * S;
 map<unsigned, unsigned> st;
 map<unsigned, unsigned>::iterator it;
 for (i = 0; i < mList.size() ; i++) {
   S = mList[i];
   for (it = st.begin(); it != st.end() ; it++) {
     if (  S->IncAddr == Addr || S->CallPool->OutAddr[it->first] == Addr )
     {
	 Cap++;
     }
   }  // in CallPool
 }  // per Thread
 return Cap;
}

string CallTable::ColumnListCalls()
{
 SignallingThread * S;
 ostringstream res, res1;
 string CalledNumber, CallingNumber;
 map<unsigned, unsigned> st;
 map<unsigned, unsigned>::iterator it;
 unsigned int i, j, Now =  time(NULL);
 j=0;
 res << setfill(' ') << setw(6) << "CallRe" << "|" << setw(15) << "From" <<  "|" << setw(15) << "To" << "|"
     << setw(20) << "CalledNumber" << "|" << setw(20) << "Callingnumber" << "|"
     << setw(5)  << "Time" << "|"
     << setw(9) << " State " << "|"
     << setw(2) << "TC" << "|" << "Thread" << "\r\n" ;
 for (i = 0; i < mList.size() ; i++) {
   S = mList[i];
   st = S->CallPool->StartTime;
   for (it = st.begin(); it != st.end() ; it++) {
     if (S->CallPool->OutCDR[it->first].CalledBill != "")
     {
       CalledNumber = S->CallPool->OutCDR[it->first].CalledBill;
     }
     else
     {
       CalledNumber =  S->CallPool->OutCDR[it->first].CalledStationId;
     }
     if (CalledNumber.size() > 20 )
     {
	CalledNumber.resize(19);
        CalledNumber = CalledNumber + "<" ;
     }
     if (S->CallPool->OutCDR[it->first].CallingBill != "")
     {
       CallingNumber = S->CallPool->OutCDR[it->first].CallingBill;
     }
     else
     {
       CallingNumber =  S->CallPool->OutCDR[it->first].CallingStationId;
     }
     if (CallingNumber.size() > 20 )
     {
       CallingNumber.resize(19);
       CallingNumber = CallingNumber + "<" ;
     }
     res << setfill(' ') << setw(6) << it->first << "|" << setw(15) << S->IncAddr <<  "|" << setw(15) 
          << S->CallPool->OutAddr[it->first] << "|" 
          << setw(20) << CalledNumber << "|" << setw(20) << CallingNumber << "|"
          << setw(5)  << (Now - (S->CallPool->StartTime[it->first])) << "|" 
	  << setw(9) << IntStates[S->CallPool->cStatesList[it->first]] << "|"
          << setw(2) << S->CallPool->InternTC[it->first] << "|" << GetNumber(S) << "\r\n"  ;
     j++;
    }  // in CallPool
 }  // per Thread
 res1 << res.str() << "\r\n" << "Total : " << j;
 return res1.str();
}

bool CallTable::isUniqCallIdent(H225_CallIdentifier CallIdent)
{
 SignallingThread * S;
 unsigned i;
 bool WasFirst;
 WasFirst = false;
 map<unsigned, H225_CallIdentifier>::iterator it;
 for (i = 0; i < mList.size() ; i++) {
   S = mList[i];
   for (it = S->CallPool->CallIdent.begin(); it != S->CallPool->CallIdent.end() ; it++) {
    if ( S->CallPool->CallIdent[it->first] == CallIdent ) {
      if (WasFirst)
        return false;
      else
        WasFirst = true;
    }
   }
 }
 return true;
}

string CallTable::GetNumber(SignallingThread * S)
{
 ostringstream strm;
 strm << hex << setfill('0') << setw(7) << (void *)S;
 return strm.str().c_str();
}

string CallTable::CallDetails(unsigned int CallRef)
{
 SignallingThread * S;
 std::ostringstream ThreadsInfo, res;
 unsigned int Now =  time(NULL);
 S = ResolveRef(CallRef);
 if ( S == NULL) 
   return "Erorr in resolving thread by CallRef";
 ThreadsInfo << ",SinallingThread=Id " << S->GetThreadName() << " ptr " << S;
 if (S->CallPool->H245ThreadList[CallRef] != NULL) {
   ThreadsInfo << ",H245Thread=Id " << S->CallPool->H245ThreadList[CallRef]->GetThreadName() << " ptr " << S->CallPool->H245ThreadList[CallRef] ;
 }
 res << "CallRef=" << CallRef << ",IncomeIP=" << S->IncAddr << ",OutIP=" << S->CallPool->OutAddr[CallRef]
	<< ",CalledNumber=" << S->CallPool->CalledNumber[CallRef] << ",Duration=" << (Now - (S->CallPool->StartTime[CallRef])) 
	<< ",ConnectionStatus=" << S->CallPool->cStatesList[CallRef] << ",FastStartS=" << S->CallPool->fStatesList[CallRef] 
	<< ",CallRef=" << CallRef << ",CallIdent=" << AddrUtils::AsString(S->CallPool->CallIdent[CallRef].m_guid) 
	<< ",IncomeBytes=" << S->CallPool->IncBytes[CallRef] << ",OutBytes=" << S->CallPool->OutBytes[CallRef]
	<< ",ProductId=" << S->IncProductId << ",VersionId=" << S->IncVersionId
	<< ",Tunneling=" << S->CallPool->Tunneling[CallRef] << ThreadsInfo.str() << S->MessLog->Output();
 return res.str();
}

SignallingThread * CallTable::ResolveRef(unsigned CallRef)
{
 unsigned i;
 map<unsigned, unsigned> StrTm;
 map<unsigned, unsigned>::iterator it;
 for (i = 0; i < mList.size() ; i++) {
  StrTm = mList[i]->CallPool->StartTime;
  for (it = StrTm.begin(); it != StrTm.end(); it++) {
    if (it->first == CallRef) {
       return mList[i];
    } // if
  } // by callpool
 } // by signal threads
 return NULL;
}

bool CallTable::FindCall(const string CallId, SignallingThread *SigThread, unsigned int & CallRef)
{
 unsigned i;
 map<unsigned, H225_CallIdentifier> StrTm;
 map<unsigned, H225_CallIdentifier>::iterator it;
 string strId;
 for (i = 0; i < mList.size() ; i++) {
   StrTm = mList[i]->CallPool->CallIdent;
   for (it = StrTm.begin(); it != StrTm.end(); it++) {
    strId = (const char *)AddrUtils::AsString(it->second.m_guid);
    if (strId == CallId) {
	SigThread = mList[i];
	CallRef = it->first ;
        return true;
    } // if
   } // for by callpool
 } // for by signal threads
 return false;
}

string CallTable::KillCall(unsigned int CallRef)
{
 SignallingThread * S;
 ostringstream res;
 S = ResolveRef(CallRef);
 if ( S == NULL) {
   res << "Erorr in resolving thread by CallRef=" << CallRef;
   return res.str() ;
 }
 if (S != NULL) {
   S->CallPool->InternTC[CallRef]= TCallPool::ForceTermination;
   S->CallPool->StopAcctCall(CallRef);
   S->ClearCall(CallRef);
 }
 res << "Thread has been killed : CallRef=" << CallRef;
 return res.str();
}

string CallTable::ShutdownCall(unsigned int CallRef)
{
 SignallingThread * S;
 ostringstream res;
 S = ResolveRef(CallRef);
 if ( S == NULL) 
 {
     res << "Erorr in resolving Signal thread by CallRef=" << CallRef;
     return res.str() ;
 }
 if (S != NULL) 
 {
   S->ShutdownCall(CallRef);
 }
 res << "Call has been shutted: CallRef=" << CallRef;
 return res.str();
}

void CallTable::KillAllCalls()
{
 unsigned int i, CallRef;
 map<unsigned, unsigned> RefMap;
 map<unsigned, unsigned>::iterator it;
 for (i=0; i < mList.size() ; i++) {
   if ( mList[i] != NULL ) {
     RefMap = mList[i] -> CallPool-> StartTime;
     for (it = RefMap.begin(); it != RefMap.end(); it++) {
       CallRef = it->first;
       WTRACE(1, "CallTable::KillAllCalls : Killing Calls with CallRef = " << CallRef );
       mList[i]->CallPool->InternTC[CallRef]= TCallPool::ForceTermination;
       mList[i]->CallPool->StopAcctCall(CallRef);
       // mList[i]->ClearCall(CallRef);  // other call aspect will stop 
     } // foreach StartTime maps for killing
   } // if
 } // for in SignalThread list
}

string CallTable::GWusage()
{
 ostringstream res;
 map<unsigned, int>::iterator it;
 for (it = DGWusage.begin(); it != DGWusage.end(); it++) 
 {
   res << WSocket::Address(it->first) << "," << it->second << ":";
 }
 res << "|";
 for (it = SGWusage.begin(); it != SGWusage.end(); it++)
 {
   res << WSocket::Address(it->first) << "," << it->second << ":";
 }
 WTRACE(3, "GWusage() request. Dsize:" << DGWusage.size() << " Csize: " << SGWusage.size());
 return res.str();
}

string CallTable::GWstatus(string GWaddr)
{
 ostringstream res;
 WSocket::Address addr1 = GWaddr.c_str();
 res << GWaddr << "," << DGWusage[(DWORD)addr1] << "," << SGWusage[(DWORD)addr1] << "," << WMGWusage[(DWORD)addr1];
 return res.str();
}

int CallTable::GWLinesUsage(string GWaddr)
{
 WSocket::Address addr1 = GWaddr.c_str();
 return DGWusage[(DWORD)addr1] + SGWusage[(DWORD)addr1] ;
}

string CallTable::GWcounters(string GWaddr)
{
 ostringstream res;
 WSocket::Address addr1 = GWaddr.c_str();
 res << GWaddr << "," << CntInAttempts[(DWORD)addr1] << "," << CntInCalls[(DWORD)addr1]
	<< "," << CntOutAttempts[(DWORD)addr1] << "," << CntOutCalls[(DWORD)addr1]  
	<< "," << WMGWusage[(DWORD)addr1] ;
 return res.str();
}

void CallTable::CheckWMGWusage(WSocket::Address GWaddr)
{
 int CurWM;
 CurWM = WMGWusage[(DWORD)GWaddr];
 WTRACE(6, "CheckWMGWusage for " << GWaddr );
 if (SGWusage[(DWORD)GWaddr] + DGWusage[(DWORD)GWaddr] > CurWM) {
   WMGWusage[(DWORD)GWaddr] = SGWusage[(DWORD)GWaddr] + DGWusage[(DWORD)GWaddr];
   WTRACE(6, "WM for " << GWaddr << " inreases ");
 }
}

//  TStatusTable 
TStatusTable::~TStatusTable()
{
 mList.clear();
}

void TStatusTable::Add(TStatusThread * StatThread)
{
 mList.push_back(StatThread);
}

void TStatusTable::Del(TStatusThread * StatThread)
{
 vector<TStatusThread*>::iterator it;
 for (it = mList.begin(); it < mList.end(); it++)
 {
   if (*it == StatThread) {
     mList.erase(it);
     return;
   } // if
 } // for
}

void TStatusTable::PrintForEach(string Info)
{
 
}

string TStatusTable::who()
{
 unsigned int i;
 ostringstream res;
 WSocket::Address PeerAddr;
 unsigned int Port;
 for (i = 0 ; i < mList.size(); i++) {
   mList[i] -> StatSocket -> GetPeerAddress(PeerAddr, Port);
   res << PeerAddr << " : " << Port << " " << "\r\n";
 }
 return res.str();
}
