/*
*
* versuch project
* RTP/RTCP channel
* $Id: MediaProxy.cxx,v 1.2 2006/05/06 13:17:54 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#define BUFFLEN 2048

#include <assert.h>

#include "MediaProxy.h"
#include "CallThread.h"
#include "AddrUtils.h"

void * udp_thread_function( void *ptr )
{
 struct T_udp_thread_data * thread_data;
 thread_data = (struct T_udp_thread_data *) ptr;
 UDPProxyLChannel * ch = new UDPProxyLChannel(thread_data -> ParenSigThread, thread_data -> mCallRef);
 thread_data -> MediaParent -> rtp_ch = ch;
 WTRACE(1, "MEDIA\tfork rtp channel. ptr=" << ch);
 ch->Started = false;
 ch->Running = true;
 ch->Main();
 WTRACE(1, "MEDIA\tExiting from udp_thread_function");
 pthread_exit(0);
}

void * rtcp_thread_function( void *ptr )
{
 struct T_udp_thread_data * thread_data;
 thread_data = (struct T_udp_thread_data *) ptr;
 RTCPProxyChannel * ch = new RTCPProxyChannel(thread_data -> ParenSigThread, thread_data -> mCallRef);
 thread_data -> MediaParent -> rtcp_ch = ch;
 WTRACE(2, "MEDIA\tfork rtcp channel. ptr=" << ch);
 ch->Started = false;
 ch->Running = true;
 ch->Main();
 WTRACE(2, "MEDIA\tExiting from rtcp_thread_function");
 pthread_exit(0);
}

MediaThread::MediaThread(const TGlobal & vEnv, SignallingThread * ParentSigThread, unsigned mCallRef)
    : Parent(ParentSigThread), env(vEnv)
{
 pthread_attr_t attr;
 int rtp_res, rtcp_res;
 struct timespec nano_time;
 IsRunning = false;
 CallRef = mCallRef;
 InMediaChannelOpened = false;
 OutMediaChannelOpened = false;

 rtp_ch = NULL;
 rtcp_ch = NULL;

 rtp_tdata  = (struct T_udp_thread_data *) malloc (sizeof (struct T_udp_thread_data));
 if (rtp_tdata == NULL )
 {
  WTRACE(1, "Error\tMemory allocation for media rtp thread fails: errno = " << errno << ". text is " << strerror(errno));
  return;
 }

 rtcp_tdata = (struct T_udp_thread_data *) malloc (sizeof (struct T_udp_thread_data));
 
 if (rtcp_tdata == NULL) 
 {
  WTRACE(1, "Error\tMemory allocation for media rtcp thread fails: errno = " << errno << ". text is " << strerror(errno));
  return;
 }
 
 rtp_tdata->ParenSigThread =  Parent;
 rtp_tdata->mCallRef =        CallRef;
 rtp_tdata->MediaParent =     this;

 rtcp_tdata->ParenSigThread = Parent;
 rtcp_tdata->mCallRef =       CallRef;
 rtcp_tdata->MediaParent =    this;

// Starting RTP/RTCP threads

 pthread_attr_init(&attr);
 pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
 
 WTRACE(1, "Starting media thread");
 rtp_res = pthread_create( &rtp_thread,  &attr, udp_thread_function, (void *) rtp_tdata  );
 rtcp_res = pthread_create( &rtcp_thread, &attr, rtcp_thread_function, (void *) rtcp_tdata );
 if ((rtp_res != 0) || (rtcp_res != 0)) {
   cerr << "Error in create Media channels: " << __func__ << endl;
   cerr << "rtp_res = " << rtp_res << "; rtcp_res = " << rtcp_res << endl;
   cerr << "errno = " << errno << ". Text is " << strerror(errno) << endl;
   assert(false);
 }
 pthread_attr_destroy(&attr);
 nano_time.tv_sec = 0;
 nano_time.tv_nsec = 50000000; // 1/20 sec
 while (rtp_ch == NULL || rtcp_ch== NULL ) {
   nanosleep(&nano_time, NULL);
 }
 WTRACE(3, "UDP threads forked. rtp_ptr=" << rtp_ch << " and rtcp_ch=" << rtcp_ch );
}

MediaThread::~MediaThread()
{
 WTRACE(2,"Mediathread terminate");
 if (rtp_tdata != NULL)
 {
   WTRACE(2, "Free pointer " << rtp_tdata );
   free(rtp_tdata);
 }
 if (rtcp_tdata != NULL) 
 {
   WTRACE(2, "Free pointer " << rtcp_tdata );
   free(rtcp_tdata);
 }
 if (rtp_ch != NULL) {
   delete rtp_ch;
 }
 if (rtcp_ch != NULL) {
   delete rtcp_ch;
 }
}

void MediaThread::Stop()
{
 int status;
 WTRACE(2, "MediaThread stopping" );
 PrintOn();
// Stopping
 if (rtp_ch != NULL) {
   rtp_ch->Running = false;
   WTRACE (2, "Stoping rtp_ch. Running=" << rtp_ch->Running << " :Started=" << rtp_ch->Started);
 }
 if (rtcp_ch != NULL) {
   rtcp_ch->Running = false;
   WTRACE (2, "Stoping rtcp_ch. Running=" << rtcp_ch->Running << " :Started=" << rtcp_ch->Started);
 }
// Joining
 if (pthread_join(rtp_thread, (void **)&status) != 0 )
 {
   WTRACE(0, "Error\terrno from pthread_join rtp_thread = " << errno << ". Text = " << strerror(errno));
 }
 if (pthread_join(rtcp_thread, (void **)&status) != 0 )
 {
   WTRACE(0, "Error\terrno from pthread_join rtcp_thread = " << errno << ". Text = " << strerror(errno));
 }
 WTRACE(2,"Media\tMedia threads was joined");


return;
 WTRACE(3, "WaitForTermination for Media Proxy");
 if (rtp_ch != NULL) {
   rtp_ch->WaitForTermination();
   delete rtp_ch;
 }
 if (rtcp_ch != NULL) {
   rtcp_ch->WaitForTermination();
   delete rtcp_ch;
 }
 WTRACE(1, "MediaThread stopped");
}

void MediaThread::Start( )
{
 in_addr s1, s2;
 WSocket::Address CallerAddr, CalledAddr;
 Parent->GetCallSignallingAddr( s1, s2, CallRef );
 CallerAddr = s1;
 CalledAddr = s2;
 
 WTRACE(3, "Starting MediaProxy");
 PrintOn();

 rtp_ch->setDestination( CalledAddr, rtp_ch->dst_port );
 rtcp_ch->setDestination( CalledAddr, rtcp_ch->dst_port );

 rtp_ch->setSource( CallerAddr, rtp_ch->src_port );
 rtcp_ch->setSource( CallerAddr, rtcp_ch->src_port );
 	
 WSocket::Address proxyIP;
 unsigned proxyPort;

 rtp_ch->getDestinationSideAddr(proxyIP, proxyPort);
 WTRACE(2, "getDestinationSideAddr rtp = "<< proxyIP <<" port "<< proxyPort );
 rtcp_ch->getDestinationSideAddr(proxyIP, proxyPort);	
 WTRACE(2, "getDestinationSideAddr rtcp = "<< proxyIP <<" port "<< proxyPort );
 
 rtp_ch->getSourceSideAddr(proxyIP, proxyPort);
 WTRACE(2, "getSourceSideAddr rtp = "<< proxyIP <<" port "<<proxyPort );
 rtcp_ch->getSourceSideAddr(proxyIP, proxyPort);
 WTRACE(2, "getSourceSideAddr rtcp = "<< proxyIP <<" port "<<proxyPort );

// Starting RTP/RTCP threads

 rtp_ch->Started = true;
 rtcp_ch->Started = true;
}

void MediaThread::SetControlChannel(H245_TransportAddress & vAddr, bool FromCaller)
{
 WSocket::Address LocalAddr;
 RTCPProxyChannel* rtcpch;
 if ( Parent == NULL ) {
   Stop();
 }
 if (FromCaller)
   LocalAddr = Parent -> CallPool ->OutLocalAddr[CallRef];
 else
   LocalAddr = Parent->IncLocalAddr;
 if ( vAddr.GetTag() == H245_TransportAddress::e_unicastAddress )
 {
    H245_UnicastAddress & UnicastAddr = vAddr;
    if ( UnicastAddr.GetTag() == H245_UnicastAddress::e_iPAddress )
    {
       H245_UnicastAddress_iPAddress & IP = UnicastAddr;
       WSocket::Address IPAddr(IP.m_network[0], IP.m_network[1], IP.m_network[2], IP.m_network[3] );
       WTRACE( 2, "RTP\tControlChannel addr is " << IPAddr << " port " << IP.m_tsapIdentifier
                         << " packet from" << (FromCaller? "Caller" : "Called") << " CallRef = " << CallRef );
       WSocket::Address proxyIP;
       unsigned proxyPort;
       rtcpch = getRTCP_LCCh();
       if ( rtcpch != NULL)
       {
        if (FromCaller)
        {
         rtcpch -> setSource( IPAddr, IP.m_tsapIdentifier );
         rtcpch -> getDestinationSideAddr(proxyIP, proxyPort);
        }
        else
        {
         rtcpch -> setDestination( IPAddr, IP.m_tsapIdentifier );
         rtcpch -> getSourceSideAddr(proxyIP, proxyPort);
        }
       }
       AddrUtils::setIPAddress( IP, (in_addr) LocalAddr, proxyPort );
    } // if H245_UnicastAddress::e_iPAddress
 }  // if H245_TransportAddress::e_unicastAddress
}

void MediaThread::SetMediaChannel(H245_TransportAddress & vAddr, bool FromCaller)
{
 WSocket::Address LocalAddr;
 UDPProxyLChannel* rtpch;
 if (FromCaller)
   LocalAddr = Parent -> CallPool ->OutLocalAddr[CallRef];
 else
   LocalAddr = Parent->IncLocalAddr;
 if ( vAddr.GetTag() == H245_TransportAddress::e_unicastAddress )
 {
    H245_UnicastAddress & UnicastAddr = vAddr;
    if ( UnicastAddr.GetTag() == H245_UnicastAddress::e_iPAddress )
    {
       H245_UnicastAddress_iPAddress & IP = UnicastAddr;
       WSocket::Address IPAddr(IP.m_network[0], IP.m_network[1], IP.m_network[2], IP.m_network[3] );
       WTRACE( 2, "MediaChannel OLC addr is " << IPAddr << " port " << IP.m_tsapIdentifier
                        << " packet from" << (FromCaller? "Caller" : "Called") << " CallRef = " << CallRef );
       WSocket::Address proxyIP;
       unsigned proxyPort;
       rtpch = getRTP_LCCh();
       if ( rtpch != NULL )
       {
         if (FromCaller)
         {
          rtpch -> setSource( IPAddr, IP.m_tsapIdentifier );
          rtpch -> getDestinationSideAddr(proxyIP, proxyPort);
         }
         else
         {
          rtpch -> setDestination(IPAddr, IP.m_tsapIdentifier);
          rtpch -> getSourceSideAddr(proxyIP, proxyPort);
         }
       }
       AddrUtils::setIPAddress( IP, (in_addr) LocalAddr, proxyPort );
    } // if H245_UnicastAddress::e_iPAddress
 } // if H245_TransportAddress::e_unicastAddress
}

void MediaThread::PrintOn() const
{
 if (!WTrace::CanTrace(1))
    return;
 ostream & trace = WTrace::Begin(1, __FILE__, __LINE__);
 trace << "MediaProxy status: CallRef = " << CallRef << endl;
 WSocket::Address IP;
 unsigned Port;

 rtp_ch->getSourceSideAddr(IP, Port);
 trace << "\tSourceSideAddr      rtp = " << IP << " / " << Port ;
 rtcp_ch->getSourceSideAddr(IP, Port);
 trace << "\tSourceSideAddr      rtcp = "<< IP << " / " << Port << endl;
 rtp_ch->getDestinationSideAddr(IP, Port);
 trace << "\tDestinationSideAddr rtp = "<< IP << " / " << Port;
 rtcp_ch->getDestinationSideAddr(IP, Port);
 trace << "\tDestinationSideAddr rtcp = "<< IP <<" / " << Port << endl;

 rtp_ch->getSource(IP, Port);
 trace << "\tRemote Source       rtp = " << IP << " / " << Port ;
 rtcp_ch->getSource(IP, Port);
 trace << "\tRemote Source       rtcp = " << IP << " / " << Port << endl;
 rtp_ch->getDestination(IP, Port);
 trace << "\tRemote Destination  rtp = " << IP << " / " << Port ;
 rtcp_ch->getDestination(IP, Port);
 trace << "\tRemote Destination  rtcp = " << IP << " / " << Port << endl;

 trace << WTrace::End;
}

///////////////////////////////////////////////////

UDPProxyLChannel::UDPProxyLChannel(SignallingThread * ParenSigThread, unsigned mCallRef)
{
 Started = false;
 Running = true;
 SigParent = ParenSigThread;
 CallRef = mCallRef;
}

UDPProxyLChannel::~UDPProxyLChannel()
{
  s_src.Close();
  s_dst.Close();
}

void UDPProxyLChannel::Main()
{
 struct timeval timeout;
 unsigned int i, Limit;
 Limit = 200; // 20 sec
 i = 0;
 WTRACE(1, "MEDIA\tStart main cycle in UDPChannel");
 while ( (! Started) && Running && (i < Limit)) {
   timeout.tv_sec = 0;
   timeout.tv_usec = 100000;  //    1/10 sec
   pthread_yield();
   select(0, NULL, NULL, NULL, &timeout);
   i++;
 }
 if ( i == Limit ) 
 {
   WTRACE(1, "Error\tError in UDPProxyLChannel::Main waiting timeout. CallRef " << CallRef );
   Running = false;
   SigParent -> CallPool -> InternTC[CallRef] = TCallPool::InternalError;
   SigParent -> Stop = true;
   Running = false;
 }
 while ( Running ) {
   proxyMsg(10);
 }
 Started = false;
}

bool UDPProxyLChannel::bindSockets(unsigned srcPort, unsigned dstPort, WSocket::Address & inAddr, WSocket::Address & outAddr)
{
 WTRACE(2, "bindSockets. inAddr " << inAddr.AsString() << ", outAddr " << outAddr.AsString()  );
 a_in = inAddr; 
 a_out = outAddr;
 if (!s_src.Listen(inAddr, srcPort)) {
   WTRACE(1, "Error\tFailed listen on s_src socket. errno = " << strerror(errno) << "(" << errno << ")" );
   WTRACE(1, "Error\tAttempt Listen " << inAddr << ":" << srcPort << ". CallRef = " << CallRef );
   return false;
 }
 if (!s_dst.Listen(outAddr, dstPort)) {
   WTRACE(1, "Error\tFailed listen on s_dst socket. errno = " << strerror(errno) << "(" << errno << ")" );
   WTRACE(1, "Error\tAttempt Listen " << outAddr << ":" << dstPort << ". CallRef = " << CallRef );
   return false;
 }
 return true;
}

void UDPProxyLChannel::setSource(const WSocket::Address & addr, unsigned port)
{
 if( addr == a_out || addr == a_in )
 {
   WTRACE(2,"Error\tsetSource conflict. addr = " << WSocket::Address(addr).AsString()  
         << ", a_in = " << a_in << ", a_out = " << a_out );
   return;
 }
 if ( (a_src != addr) || (src_port != port) ) {
  a_src = addr; src_port = port;
  WTRACE(3, "RTP\tUDPProxyLChannel::setSource to " << a_src << " port " << src_port);
  s_src.SetSendAddress( a_src, src_port);
 }
}

void UDPProxyLChannel::setDestination(const WSocket::Address & addr, unsigned port)
{
  if( addr == a_out || addr == a_in )
  {
    WTRACE(2,"Error\tsetDestination conflict. addr = " << WSocket::Address(addr).AsString() 
	    << ", a_in = " << a_in << ", a_out = " << a_out );
    return;  
  }
  if ( (a_dst != addr) || (dst_port != port) ) {
   a_dst = addr; dst_port = port;
   WTRACE(3, "UDPProxyLChannel::setDestination to " << a_dst << " port " << dst_port);
   s_dst.SetSendAddress( a_dst, dst_port);
 }
}

void UDPProxyLChannel::getSourceSideAddr(WSocket::Address & addr, unsigned & port)
{ 
   s_src.GetLocalAddress(addr, port);  
}

void UDPProxyLChannel::getDestinationSideAddr(WSocket::Address & addr, unsigned & port)
{ 
   s_dst.GetLocalAddress(addr, port);  
}

void UDPProxyLChannel::proxyMsg(const int sec)
{
 int selres = WUDPSocket::Select(s_src, s_dst, sec);   
 if (! Running ) {
    // race condition possible
    return;
 }
 switch (selres) {
  case -1: // src socket
         do_proxy(s_src, s_dst, true);
         return ;
  case -2: // dst socket
         do_proxy(s_dst, s_src, false);
         return ;
  case -3: // both sockets
         do_bidirectional_proxy();
         return ;
  case 0: // timeout
         return ;
         break;
  default: // some error
         WTRACE( 1, "Error " << selres << " on Media select : " << strerror(selres) << " . Thread is stoping");
         Running = false;
         SigParent -> CallPool -> InternTC[CallRef] = TCallPool::InternalError;
	 SigParent -> Stop = true;
         break;
 }
}

void UDPProxyLChannel::do_proxy(WUDPSocket & rd, WUDPSocket & wr, bool Caller)
{
 WSocket::Address addr;
 unsigned port;
 unsigned len = 0;
 unsigned char buff[BUFFLEN];

 if ( rd.ReadFrom(buff, (unsigned)BUFFLEN, addr, port) && (len = rd.GetLastReadCount()) > 0 ) 
 {
   if (Caller) {
      SigParent->CallPool->IncBytes[CallRef] += len;
      SigParent->CallPool->LastIncByte[CallRef] = time(NULL);
   }
   else {
      SigParent->CallPool->OutBytes[CallRef] += len;
      SigParent->CallPool->LastOutByte[CallRef] = time(NULL);
   }
   if (WTrace::CanTrace(4)) {
     WSocket::Address laddr, saddr, sladdr;
     unsigned lport, sport, slport;
     rd.GetLocalAddress(laddr, lport);
     wr.GetSendAddress(saddr, sport);
     wr.GetLocalAddress(sladdr, slport);
   }
   if (len == BUFFLEN) 
		WTRACE(2, "Reading buffer filled");
    // must check if addr and port are the same of source
    if (!wr.Write(buff, len)) {
      // must throw an exception
      WSocket::Address laddr, saddr, sladdr;
      unsigned lport, sport, slport;
      rd.GetLocalAddress(laddr, lport);
      wr.GetSendAddress(saddr, sport);
      wr.GetLocalAddress(sladdr, slport);
      cerr <<  "proxyMsg: udp write error on port " << slport << " destination " 
                     << saddr << ":" << sport << " (data read from " << addr << ":"
                     << port << " on port " << lport << ")" << endl ; 
    }
 } 
 else 
 {
   cerr << "proxyMsg: read error. len is " << len << endl;
   WSocket::Address laddr;
   unsigned lport;
   rd.GetLocalAddress(laddr, lport);
   WTRACE(1, "UDPProxyLChannel::do_proxy() status : GetLastReadCount " << rd.GetLastReadCount() 
	   << ". Socket addr " << laddr << ":" << lport << endl 
           << "Read data from " << addr << ":" << port << endl
	   << "Socket handle is " << rd.GetHandler() );
 }
}

void UDPProxyLChannel::do_bidirectional_proxy()
{
   unsigned char buff_s[BUFFLEN], buff_d[BUFFLEN]; 
   WSocket::Address addr_s, addr_d;
   unsigned port_s, port_d;
   PINDEX len_s = 0, len_d = 0;
   BOOL f_s, f_d;

   f_s = s_src.ReadFrom(buff_s, (unsigned)BUFFLEN, addr_s, port_s);
   f_d = s_dst.ReadFrom(buff_d, (unsigned)BUFFLEN, addr_d, port_d);

   if ( f_s && (len_s = s_src.GetLastReadCount()) > 0 ) 
   {
      SigParent->CallPool->IncBytes[CallRef] += len_s;
      SigParent->CallPool->LastIncByte[CallRef] = time(NULL);
      if (len_s == BUFFLEN) WTRACE(2, "Reading buffer filled");
         if (WTrace::CanTrace(4)) {
           WSocket::Address laddr, saddr, sladdr;
           unsigned lport, sport, slport;
           s_src.GetLocalAddress(laddr, lport);
           s_dst.GetSendAddress(saddr, sport);
           s_dst.GetLocalAddress(sladdr, slport);
         }
      // must check if addr and port are the same of source
      if (!s_dst.Write(buff_s, len_s)) {
         // must throw an exception
         WTRACE(3, "proxyMsg: udp write error (data read from " << addr_s << " : " << port_s << ")");
      }
   } else {
      //cerr <<  "proxyMsg: bidirectional_proxy: read error on source socket" << endl;
   }
   if ( f_d && (len_d = s_dst.GetLastReadCount()) > 0 ) 
   {
      SigParent->CallPool->OutBytes[CallRef] += len_d;
      SigParent->CallPool->LastOutByte[CallRef] = time(NULL);
      if (len_d == BUFFLEN) WTRACE(2, "Reading buffer filled");
         if (WTrace::CanTrace(4)) {
          WSocket::Address laddr, saddr, sladdr;
          unsigned lport, sport, slport;
          s_dst.GetLocalAddress(laddr, lport);
          s_src.GetSendAddress(saddr, sport);
          s_src.GetLocalAddress(sladdr, slport);
          WTRACE(4, "UDPProxyLChannel::do_bidirectional_proxy() letti " <<len_d<< " byte port " << slport << " destination " 
                     << saddr << ":" << sport << " (data read from " << addr_s << ":"
                     << port_d << " on port " << lport << ")" );
         }
      // must check if addr and port are the same of source
      if (!s_src.Write(buff_d, len_d)) {
         // must throw an exception
         WTRACE(3, "proxyMsg: udp write error (data read from " << addr_d << " : " << port_d << ")");
      }
   } else {
      // cerr << "proxyMsg: bidirectional_proxy: read error on dest socket " << endl;
   }
}

void UDPProxyLChannel::WaitForTermination()
{
 unsigned int i, Limit;
 Limit = 200; // 20 sec
 i = 0;
 struct timeval timeout;
 timeout.tv_sec = 0;
 timeout.tv_usec = 100000;  //    1/10 sec

 WTRACE (3, "WaitForTermiantion. Running=" << Running << " :Started=" << Started);
 while ((Running || Started) && (i < Limit)) {
  pthread_yield();
  select(0, NULL, NULL, NULL, &timeout);
  i++;
 }
 if ( i == Limit )
   WTRACE(1, "Error in UDPProxyLChannel:WaitForTermination() waiting timeout. CallRef = " << CallRef);
 WTRACE(1, "UDP Thread Terminated. CallRef = " << CallRef);
}

void RTCPProxyChannel::proxyMsg(const int sec)
{
 SigParent->CallPool->RTCPcount[CallRef] ++;
 int selres = WUDPSocket::Select(s_src, s_dst, sec);
 switch (selres) {
   case -1: // src socket
    do_proxy(s_src, s_dst, true);
    return ;
   case -2: // dst socket
    do_proxy(s_dst, s_src, false);
    return ;
   case -3: // both sockets
    do_bidirectional_proxy();
    return ;
   case 0: // timeout
    return ;
    break;
   default: // some error?
    WTRACE( 1, "Error " << selres << " on Media select : " << strerror(selres) << " . Thread is stoping");
    Running = false;
    SigParent -> CallPool -> InternTC[CallRef] = TCallPool::InternalError;
    SigParent -> Stop = true;
    break;
 }
}
