/*
*
* versuch project
* RTP/RTCP channel
* $Id: MediaProxy.h,v 1.1 2006/03/06 13:07:56 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#ifndef _MEDIATHREAD_H
#define _MEDIATHREAD_H

#include "Sockets.h"
#include "global.h"
#include "CallThread.h"

#include <ptlib.h>
#include <pthread.h>

class RTCPProxyChannel;
class UDPProxyLChannel;
class SignallingThread; 
class MediaThread;  

struct T_udp_thread_data
{
  SignallingThread * ParenSigThread;
  unsigned mCallRef;
  MediaThread	   * MediaParent;
};

class MediaThread 
{
 public:
  MediaThread(const TGlobal & env, SignallingThread * ParenSigThread, unsigned mCallRef);
  virtual ~MediaThread();
  void Stop();
  void Start();
  UDPProxyLChannel*  getRTP_LCCh(){ return rtp_ch; };
  RTCPProxyChannel*  getRTCP_LCCh(){ return rtcp_ch; };
 
  void SetControlChannel(H245_TransportAddress & vAddr, bool FromCaller);
  void SetMediaChannel(H245_TransportAddress & vAddr, bool FromCaller);
  void PrintOn() const;
  bool IsRunning;
  bool OutMediaChannelOpened, InMediaChannelOpened;  // for EstablishedConnectionCheck
  UDPProxyLChannel * rtp_ch;
  RTCPProxyChannel * rtcp_ch;
  SignallingThread * Parent;
 protected:
  struct T_udp_thread_data * rtp_tdata, * rtcp_tdata;
  pthread_t rtp_thread, rtcp_thread;
  const TGlobal & env;
  unsigned CallRef;
};

class UDPProxyLChannel : public PObject
{
PCLASSINFO(UDPProxyLChannel, PObject)
public:
 UDPProxyLChannel(SignallingThread * ParenSigThread, unsigned mCallRef);
 virtual ~UDPProxyLChannel();
 virtual void Main();
 virtual bool bindSockets(unsigned srcPort, unsigned dstPort, WSocket::Address & inAddr, WSocket::Address & outAddr );
 virtual void setSource(const WSocket::Address & addr, unsigned port);
 virtual void setDestination(const WSocket::Address & addr, unsigned port);
 virtual void getSourceSideAddr(WSocket::Address & addr, unsigned & port);
 virtual void getDestinationSideAddr(WSocket::Address & addr, unsigned & port);
 void getSource(WSocket::Address & addr, unsigned & port) const
          {    addr = a_src; port = src_port; };
 void getDestination(WSocket::Address & addr, unsigned & port) const
          {    addr = a_dst; port = dst_port; };
 void WaitForTermination();   
 WSocket::Address    a_src, a_dst;       // endpoint addresses of channel
 unsigned            src_port, dst_port; // ...
 WSocket::Address    a_in,  a_out;       // addresses of sockets
 SignallingThread    * SigParent;  
 bool                Running, Started;
 MediaThread	     * MediaParent;
protected:
 virtual void proxyMsg(const int sec = 0);
 void do_proxy(WUDPSocket & rd, WUDPSocket & wr, bool Caller);
 void do_bidirectional_proxy();
 WUDPSocket   s_src, s_dst;
 unsigned     CallRef;
} ;

class RTCPProxyChannel : public UDPProxyLChannel
{
public:
  RTCPProxyChannel(SignallingThread * ParenSigThread, unsigned mCallRef) : UDPProxyLChannel(ParenSigThread, mCallRef) {};
protected:
  virtual void proxyMsg(const int sec = 0);
};

#endif // _MEDIATHREAD_H
