/*
* 
* versuch project
* Socket routines
* $Id: Sockets.cxx,v 1.2 2006/05/06 13:17:54 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#include "Sockets.h"

#include <assert.h>

#include "Utils.h"

static WSocket::Address loopback4(127,0,0,1);

WSocket::Address::Address()
{
 *this = loopback4;
}

WSocket::Address::Address(unsigned char b1, unsigned char b2, unsigned char b3, unsigned char b4)
{
 unsigned char * p = (unsigned char *)&v.s_addr;
 p[0] = b1;
 p[1] = b2;
 p[2] = b3;
 p[3] = b4;
}

WSocket::Address::Address(const in_addr & addr)
{
 v = addr;
}

WSocket::Address::Address(const char * str_addr)
{
 v.s_addr = inet_addr(str_addr);
 if ( v.s_addr == INADDR_NONE )
   cerr << "Constructor fails with wrong address : " << str_addr << endl ;
}

WSocket::Address::Address(const u_int32_t & addr)
{
 v.s_addr = addr;
}

WSocket::Address::Address & WSocket::Address::operator=(const in_addr & addr)
{
 v = addr;
 return *this;
}

WSocket::Address::operator in_addr() const
{
 return v;
}

ostream & operator<<(ostream & s, WSocket::Address & a)
{
  return s << a.AsString();
}

const char * WSocket::Address::AsString()
{
 // inet_ntoa is thread unsafe
 bzero(strbuf, sizeof(strbuf));
 return inet_ntop(AF_INET, & v, strbuf, sizeof(strbuf));
}

bool WSocket::Address::operator==(in_addr & addr) const
{
 Address other = addr;
 return ( (unsigned) *this == (unsigned) other);
}

bool WSocket::Address::operator==(const Address & addr) const
{
 Address a1 = addr;
 return (unsigned) *this  == (unsigned) addr;
}

bool WSocket::Address::operator==(u_int32_t & addr) const
{
 return (unsigned) *this == addr;
}

unsigned char WSocket::Address::operator[](unsigned idx) const
{
 if (idx > 3)
   assert(false);
 return ((unsigned char *)&v.s_addr)[idx];
}

WSocket::Address::operator unsigned() const
{
  return (unsigned)v.s_addr;
}

unsigned char WSocket::Address::Byte1() const
{
  return *(((char *)&v.s_addr)+0);
}

unsigned char WSocket::Address::Byte2() const
{
  return *(((char *)&v.s_addr)+1);
}

unsigned char WSocket::Address::Byte3() const
{
  return *(((char *)&v.s_addr)+2);
}
unsigned char WSocket::Address::Byte4() const
{
  return *(((char *)&v.s_addr)+3);
}

/////////////////////////////////////////////////

WSocket::WSocket()
{
 os_handle = -1;
 localPort = 0;
 backlogSize = 4096;
}

void WSocket::Close()
{
 if (os_handle != -1 ) 
 {
   if (::close(os_handle) != 0 )
   {  
     // cerr << "Error on close operation for handle. errno = " << errno << endl;
   }
   localPort = peerPort = 0;
   os_handle = -1;
 }
}

bool WSocket::Listen(const Address & bindAddr, unsigned newPort, bool Reusability)
{
  // make sure we have a port
  if (newPort != 0)
    localPort = newPort;

  if (!IsOpen()) {
    // attempt to create a socket
    if (!OpenSocket())
      return false;
  }
  // attempt to listen
  if (!SetOption(SO_REUSEADDR, (Reusability ? 1 : 0) ) ) {
    Close();
    return false;
  }

  // attempt to bind
  sockaddr_in sin;
  memset(&sin, 0, sizeof(sin));
  sin.sin_family      = AF_INET;
  sin.sin_addr.s_addr = bindAddr;
  sin.sin_port        = htons(localPort);

  if (::bind(os_handle, (struct sockaddr*)&sin, sizeof(sin)) == 0)
  {
    socklen_t size = sizeof(sin);
    if (::getsockname(os_handle, (struct sockaddr*)&sin, &size) == 0) {
      localAddress = sin.sin_addr;
      localPort = ntohs(sin.sin_port);
      return true;
    }
  }
  ::close(os_handle);
  return false;
}

bool WSocket::Bind(const Address & bindAddr, unsigned Port)
{
  if (!IsOpen()) {
    if (!OpenSocket())
      return false;
  }

  sockaddr_in sin;
  memset(&sin, 0, sizeof(sin));
  sin.sin_family      = AF_INET;
  sin.sin_addr.s_addr = bindAddr;
  sin.sin_port        = htons(localPort);
  if (::bind(os_handle, (struct sockaddr*)&sin, sizeof(sin)) == 0)
  {
    socklen_t size = sizeof(sin);
    if (::getsockname(os_handle, (struct sockaddr*)&sin, &size) == 0) {
      localAddress = sin.sin_addr;
      localPort = ntohs(sin.sin_port);
      return true;
    }
  }
  ::close(os_handle);
  return false;
}

bool WSocket::Bind(const Address & iFace)
{
 return Bind(iFace, 0);
}

bool WSocket::GetLocalAddress(Address & addr, unsigned & portNum)
{
  sockaddr_in address;
  socklen_t size = sizeof(address);
  if ( ::getsockname(os_handle,(struct sockaddr*)&address,&size) == -1)
    return false;

  addr = address.sin_addr;
  portNum = ntohs(address.sin_port);
  return true;
}

bool WSocket::GetLocalAddress(Address & addr)
{
  sockaddr_in address;
  socklen_t size = sizeof(address);
  if ( ::getsockname(os_handle,(struct sockaddr*)&address,&size) == -1)
    return false;
  addr = address.sin_addr;
  return true;
}

bool WSocket::GetPeerAddress(Address & addr, unsigned & portNum)
{
  sockaddr_in address;
  socklen_t size = sizeof(address);
  if ( ::getpeername(os_handle, (struct sockaddr*)&address, &size) == -1)
    return false;

  addr = address.sin_addr;
  portNum = ntohs(address.sin_port);
  return true;
}

bool WSocket::GetPeerAddress(Address & addr)
{
  sockaddr_in address;
  socklen_t size = sizeof(address);
  if ( ::getpeername(os_handle, (struct sockaddr*)&address, &size) == -1)
    return false;
  addr = address.sin_addr;
  return true;
}

int WSocket::Select(WSocket & sock1, WSocket & sock2, const int sec)
{
  struct timeval timeout;
  int rval;
  int h1,h2;
  if (!sock1.IsOpen() && !sock2.IsOpen())
    return 5;

  h1 = sock1.GetHandler();
  h2 = sock2.GetHandler();

  int plus_one;
  if (h1 > h2)
    plus_one = h1;
  else
    plus_one = h2;
  plus_one++;
  timeout.tv_usec = 0;
  timeout.tv_sec = sec;

  fd_set readfds;
  FD_ZERO(&readfds);
  if (h1 != -1) {
    FD_SET(h1, &readfds);
  }
  if (h2 != -1) {
    FD_SET(h2, &readfds);
  }

  if (sec == 0)
    rval = ::select(plus_one, &readfds, NULL, NULL, NULL);
  else
    rval = ::select(plus_one, &readfds, NULL, NULL, &timeout);

  if ( rval == -1 )
    return errno;

  rval = 0;
  if ((h1 != -1) && FD_ISSET(h1, &readfds)) {
    rval -= 1;
  }
  if ((h2 != -1) && FD_ISSET(h2, &readfds)) {
    rval -= 2;
  }
  return rval;
}

int WSocket::Select(WSocket & sock,  const int sec )
// Return true if socket get data
{
 struct timeval timeout;
 int shandler, plus_one, retval;
 fd_set readfds;
 if ( ! sock.IsOpen() )
   return -1;
 shandler = sock.GetHandler();
 plus_one = shandler + 1;
 FD_ZERO(&readfds);
 FD_SET(shandler, &readfds);
 if ( sec != 0 )
 {
   timeout.tv_usec = 0;
   timeout.tv_sec = sec;
   retval = ::select(plus_one, &readfds, NULL, NULL, &timeout);
 }
 else
 {
   retval = ::select(plus_one, &readfds, NULL, NULL, NULL);
 }
 if ( retval == -1 )
 {
    WTRACE(1, "Select error. errno = " << errno << " : " << strerror(errno) );
    return -1;
 }
 if ( retval == 0 )
 {
    // No data avaliable. Timeout expires.
    return 0;
 }
 if ( FD_ISSET( shandler, &readfds) )
 {
    // Socket ready
    return 1;
 }
 return -1;
}

int WSocket::Select( std::vector<WSocket *> & SocketList, const int sec )
{
 WSocket * ptr;
 struct timeval timeout;
 int shandler, plus_one = 0, retval;
 fd_set readfds;
 FD_ZERO(&readfds);
 std::vector<WSocket *>::iterator it;
 it = SocketList.begin();
 while ( it != SocketList.end() )
 {
   ptr = *it; 
   if ( ptr == NULL || ( ! ptr -> IsOpen() ) )
   {
      SocketList.erase(it);
      continue;
   }
   shandler = ptr -> GetHandler();
   if ( shandler > plus_one )
         plus_one = shandler;
   if (shandler > 0) // possible race condition
     FD_SET(shandler, &readfds);
   it++;
 } // while
 plus_one++;
 if ( sec != 0 )
 {
   timeout.tv_usec = 0;
   timeout.tv_sec = sec;
   retval = ::select(plus_one, &readfds, NULL, NULL, &timeout);
 }
 else
 {
   retval = ::select(plus_one, &readfds, NULL, NULL, NULL);
 } 
 if ( retval == -1 )
 {
   WTRACE(1, "Select error. errno = " << errno << " : " << strerror(errno) );
   return -1;
 }
 if ( retval == 0 )
 {
   // No data avaliable. Timeout expires.
   return 0;
 }
 it = SocketList.begin();
 while ( it != SocketList.end() )
 {
   ptr = *it;
   if ( ptr == NULL || ( ! ptr -> IsOpen() ) ) 
   {
	SocketList.erase(it);
        continue;
   }
   shandler = ptr -> GetHandler();
   if ((shandler < 0) || (! FD_ISSET( shandler, &readfds)) )
      SocketList.erase(it);
   else
      it++;
 } // while
 return retval;
}

bool WSocket::SetOption(int option, int value, int level)
{
 return (::setsockopt(os_handle, level, option, (char *)&value, sizeof(value)) ==0);
}

bool WSocket::Accept(WSocket & Listener)
{
  int resSocket;
  unsigned pp, lp;
  sockaddr_in address;
  address.sin_family = AF_INET;
  socklen_t addrLen = sizeof(address);
  resSocket = ::accept(Listener.os_handle, (struct sockaddr *)&address, &addrLen) ;
  if (resSocket == -1)
    return false;
  os_handle = resSocket;
  GetPeerAddress(peerAddress, pp);
  GetLocalAddress(localAddress, lp);
  peerPort = pp;
  localPort = lp;
  return true;
}

bool WSocket::Connect(const Address & addr, unsigned Port)
{
  if (!IsOpen()) {
    if (!OpenSocket())
      return false;
  }

  sockaddr_in sin;
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port   = htons(Port);
  sin.sin_addr   = addr;
  if (::connect(os_handle, (struct sockaddr *)&sin, sizeof(sin)) == 0) 
  {
    peerPort = Port;
    peerAddress = addr;
    GetLocalAddress(localAddress, localPort);
//    GetLocalAddress(localAddress, lp);
    return true;
  }
  os_close();
  return false;
}


void WSocket::os_close()
{
 close(os_handle);
}

string WSocket::OnPrint()
{
 std::stringstream strm;
 Address laddr, paddr;
 unsigned lport, pport;
 GetLocalAddress(laddr, lport);
 GetPeerAddress(paddr, pport);
 strm << endl << 
	"\tSocket. Object data. Handle = " << os_handle << endl << 
	"\tlocalPort = " << localPort << " ; peerPort = " << peerPort << endl <<
	"\tlocalAddress = " << localAddress << " ; peerAddress = " << peerAddress << endl <<
	"\tActual data:" << endl <<
        "\tlocalPort = " << lport << " ; peerPort = " << pport << endl <<
        "\tlocalAddress = " << laddr << " ; peerAddress = " << paddr << endl;
 
 return strm.str() ;
}

bool WSocket::GetHostAddress(const char * name, Address & addr)
{
 struct hostent * host_info;
 struct in_addr ip_addr;
 host_info = ::gethostbyname(name);
 if ( ! host_info )
   return false;
 ip_addr = *(struct in_addr *)(host_info->h_addr);
 addr = ip_addr;
 return true;
}

bool WSocket::SetReadTimeout(unsigned int sec)
{
 struct timeval tv;
 tv.tv_usec = 0;
 tv.tv_sec  = sec;
 if (setsockopt(os_handle, SOL_SOCKET, SO_RCVTIMEO, (void *) & tv, sizeof(tv) ) == -1)
   return false;
 return true;
}

bool WSocket::SetWriteTimeout(unsigned int sec)
{
 struct timeval tv;
 tv.tv_usec = 0;
 tv.tv_sec  = sec;
 if (setsockopt(os_handle, SOL_SOCKET, SO_SNDTIMEO, (void *) & tv, sizeof(tv) ) == -1)
   return false;
 return true;
}

bool WSocket::GetReadTimeout(unsigned int & sec)
{
 struct timeval tv;
 socklen_t optlen;
 optlen = sizeof(struct timeval);
 if (getsockopt(os_handle, SOL_SOCKET, SO_SNDTIMEO, &tv, &optlen ) == -1 )
  return false;
 sec = tv.tv_sec;
 return true;
}

//////////////////////////////////////////////////////

WPipe::WPipe(int handle)
{
 os_handle = handle;
 localPort = 0;
 backlogSize = 4096;
 lastWriteCount = lastReadCount = 0;
}

WPipe::~WPipe()
{
 WTRACE(1, "Pipe destroy");
}

bool WPipe::Write( const char * buf, unsigned int len )
{
 int res;
 res = write(os_handle, (const void *) buf, len);
 if ( res == -1 )
 {
   cerr << "Pipe write error. errno = " << errno << " : " << strerror(errno) << endl;
   return false;
 }
 lastWriteCount = res;
 return true;
}

bool WPipe::Read( char * buf, unsigned & len )
{
 int res;
 res = read(os_handle, ( void *) buf, len);
 if ( res == -1 )
 {
   cerr << "Pipe read error. errno = " << errno << " : " << strerror(errno) << endl;
   return false;
 }
 len = lastReadCount = res;
 return true;
}

//////////////////////////////////////////////////////

WUDPSocket::WUDPSocket()
 : WSocket()
{
 os_handle = -1;
 peerPort = 0;
}

bool WUDPSocket::ReadFrom(void * buf, unsigned len, Address & addr, unsigned & port)
{
  lastReadCount = 0;
  sockaddr_in sockAddr;
  socklen_t addrLen = sizeof(sockAddr);

  int r = ::recvfrom(os_handle, (char *)buf, len, 0, (struct sockaddr *)&sockAddr, &addrLen);
  if (r == -1) {
    cerr << "UDPsocket error. errno is " << errno << ". text is " << strerror(errno) << endl;
    return FALSE;
  }
  lastReadCount = r;
  addr = sockAddr.sin_addr;
  port = ntohs(sockAddr.sin_port);
  return lastReadCount > 0;
}

bool WUDPSocket::OpenSocket()
{
 os_handle = ::socket(AF_INET, SOCK_DGRAM, 0);
 if ( os_handle == -1 ) {
  WTRACE(1, "Error in socket openning : errno=" << errno);
  return false;
 }
 else {
  return true;
 }
}

bool WUDPSocket::OpenSocket(int ipAdressFamily)
{
 os_handle = ::socket(ipAdressFamily, SOCK_DGRAM, 0);
 if ( os_handle == -1 ) {
  WTRACE(1, "Error in socket openning : errno=" << errno);
  return false;
 }
 else {
  return true;
 }
}

void WUDPSocket::SetSendAddress(const Address & newAddress, unsigned newPort)
{
  peerAddress = newAddress;
  peerPort    = newPort;
}

void WUDPSocket::GetSendAddress(Address & address, unsigned & port)
{
  address = peerAddress;
  port    = peerPort;
}

bool WUDPSocket::Write( const void * buf, unsigned len )
{
  if (peerPort == 0)
    return false;
  lastWriteCount = 0;
  sockaddr_in sockAddr;
  sockAddr.sin_family = AF_INET;
  sockAddr.sin_addr = peerAddress;
  sockAddr.sin_port = htons(peerPort);
  int r = sendto(os_handle, buf, len, 0, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
  if (r == -1 || r < (int)len )
    return false;
  lastWriteCount = r;
  return true;
}

bool WUDPSocket::WriteTo( const void * buf, unsigned len, Address addr, unsigned int port )
{
 sockaddr_in sockAddr;
 sockAddr.sin_family = AF_INET;
 sockAddr.sin_addr = addr;
 sockAddr.sin_port = htons(port);
 int r = sendto(os_handle, buf, len, 0, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
 if (r == -1 )
 {
   cerr << "UDP socket WriteTo fails. errno = " << errno << " : " << strerror(errno) << endl;
   return false;
 }
 lastWriteCount = r;
 return true;
}

///////////////////////////////////////////////////////////////////

WTCPSocket::WTCPSocket()
 : WSocket()
{
 os_handle = -1;
 localPort = 0;
}

WTCPSocket::~WTCPSocket()
{
//cerr << "Try to destroy TCP socket. PeerAddr = " << peerAddress << ", PeerPort = " << peerPort << endl;
//cerr << "LocalPort = " << localPort << endl;
}

WTCPSocket::WTCPSocket(unsigned Port)
 : WSocket()
{
 os_handle = -1;
 localPort = Port;
}

bool WTCPSocket::OpenSocket()
{
 os_handle = ::socket(AF_INET, SOCK_STREAM, 0);
 if ( os_handle == -1 ) {
  WTRACE(1, "Error in socket openning : errno=" << errno);
  return false;
 }
 else {
  SetReadTimeout(30);
  SetWriteTimeout(30);
  return true;
 }
}

bool WTCPSocket::ReadBlock(void * buf, unsigned len)
{
  int remain = len;
  lastReadCount = 0;

  while ( remain != 0) {
   int r = ::recv(os_handle, (char *)buf + lastReadCount, remain, 0);
   if (r == -1) {
    cerr << "TCPSocket recv. errno is " << errno << ". text is " << strerror(errno) << endl;
    return false;
   }
   if (r == 0 ) {
    return false;
   }
   lastReadCount += r;
   remain = len - lastReadCount;
  }
  return lastReadCount > 0;
}

int WTCPSocket::ReadChar()
{
  lastReadCount = 0;
  unsigned char c;
  bool retVal = ReadBlock(&c, 1);
  return (retVal && lastReadCount == 1) ? c : -1;
}

bool WTCPSocket::Write( const void * buf, unsigned len )
{
  if (peerPort == 0)
    return false;
  lastWriteCount = 0;
  sockaddr_in sockAddr;
  sockAddr.sin_family = AF_INET;
  sockAddr.sin_addr = peerAddress;
  sockAddr.sin_port = htons(peerPort);
  int r = sendto(os_handle, buf, len, 0, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
  if ( r == -1 ) {
     cerr << "WTCPSocket::Write error on sendto. errno = " << errno << ". text is " << strerror(errno) << endl;
     return false;
  }
  if ( r < (int) len ) {
     cerr << "BLIA!!! " << endl << "r= " << r << ", len = " << len << endl;
    return false;
  }
  lastWriteCount = r;
  return true;
}

bool WTCPSocket::WriteString( const std::string Str0 )
{
 return( Write( Str0.c_str(), Str0.size() ) );
}

bool WTCPSocket::Listen(const Address & bindAddr, unsigned newPort, bool Reusability)
{
 if ( !WSocket::Listen(bindAddr, newPort, Reusability) )
  return false;
 if ( ::listen(os_handle, backlogSize) == -1) 
  return false;
 return true;
}

bool WTCPSocket::WriteOutOfBand(void const * buf, int len)
{
  int count = ::send(os_handle, (const char *)buf, len, MSG_OOB);
  if (count < 0) {
    lastWriteCount = 0;
    return false;
  }
  else {
    lastWriteCount = count;
    return true;
  }
}

///////////////////////////////////////////////////////////////////

WTelnetSocket::WTelnetSocket()
 : WTCPSocket()
{
 os_handle = -1;
 peerPort = 0;
}

bool WTelnetSocket::ReadBlock(void * buf, unsigned len)
{
  lastReadCount = 0;
  int r = ::recv(os_handle, (char *)buf, len, 0);
// cerr << "Socket got " << r << " bytes.";
  if (r == -1) {
    cerr << "TelnetSocket recv. errno is " << errno << ". text is " << strerror(errno) << endl;
    return false;
  }
  if (r == 0 ) {
    return false;
  }
  lastReadCount = r;
  return lastReadCount > 0;
}

///////////////////////////////////////////////////////////////////

WTelnetSocket2::WTelnetSocket2()
 : WTCPSocket()
{
 os_handle = -1;
 peerPort = 0;
 Construct();
}

bool WTelnetSocket2::ReadBlock(void * buf, unsigned len)
{
 lastReadCount = 0;
 int r = ::recv(os_handle, (char *)buf, len, 0);
  if (r == -1) 
  {
    cerr << "TelnetSocket recv. errno is " << errno << ". text is " << strerror(errno) << endl;
    return false;
  }
  if (r == 0 ) 
  {
      return false;
  }
  lastReadCount = r;
  return lastReadCount > 0;
}

bool WTelnetSocket2::SendCommand(Command cmd, int opt)
{
WTRACE(1, "SendCommamd. Command = " << cmd << ", Code = " << opt);
  unsigned char buffer[3];
  buffer[0] = IAC;
  buffer[1] = (BYTE)cmd;

  switch (cmd) {
    case DO :
    case DONT :
    case WILL :
    case WONT :
       buffer[2] = (BYTE)opt;
       return WTCPSocket::Write(buffer, 3);

    case InterruptProcess :
    case Break :
    case AbortProcess :
    case SuspendProcess :
    case AbortOutput :
       if (opt) {
          // Send the command
         if (!WTCPSocket::Write(buffer, 2))
                return false;
          // Send a TimingMark for output flush.
          buffer[1] = TimingMark;
          if (!WTCPSocket::Write(buffer, 2))
                return false;
          // Send a DataMark for synchronisation.
          if (cmd != AbortOutput) {
                buffer[1] = DataMark;
                if (!WTCPSocket::Write(buffer, 2))
                      return false;
          // Send the datamark character as the only out of band data byte.
                if (!WriteOutOfBand(&buffer[1], 1))
                      return false;
       }
       // Then flush any waiting input data.
       unsigned int oldTimeout;
       GetReadTimeout(oldTimeout);
       SetReadTimeout(0);
       while (WTCPSocket::ReadBlock(buffer, sizeof(buffer)))
          ;
       SetReadTimeout(oldTimeout);
    }
    break;

    default :
WTRACE(1, "Sending buffer = " << buffer);
    return WTCPSocket::Write(buffer, 2);
  }
  return true;
}

void WTelnetSocket2::Construct()
{
  synchronising = 0;
  terminalType = "UNKNOWN";
  windowWidth = windowHeight = 0;
  state = StateNormal;
	   
  memset(option, 0, sizeof(option));
  SetOurOption(TransmitBinary);
  SetOurOption(SuppressGoAhead);
  SetOurOption(StatusOption);
  SetOurOption(TimingMark);
  SetOurOption(TerminalSpeed);
  SetOurOption(TerminalType);
  SetTheirOption(TransmitBinary);
  SetTheirOption(SuppressGoAhead);
  SetTheirOption(StatusOption);
  SetTheirOption(TimingMark);
//  SetTheirOption(EchoOption);
}

bool WTelnetSocket2::Accept(WSocket & sock)
{
  SetReadTimeout(0);
  if (!WTCPSocket::Accept(sock))
     return false;

// original from pwlib
//  SendDo(SuppressGoAhead);
//  SendWill(StatusOption);
// end

//  SendDont(EchoOption);
  SendWill(SuppressGoAhead);
//  SendWill(StatusOption);
  SendWill(EchoOption);

// From GkStatus
// SendDo(PTelnetSocket::SuppressGoAhead);
// SendWill(PTelnetSocket::StatusOption);
// SendDont(PTelnetSocket::EchoOption);


//  SendDo(SuppressGoAhead);
//  SendDo(StatusOption);
//  SendWill(TerminalSpeed);
//  SendDo(EchoOption);
//  SendWill(EchoOption);

  return true;
}

bool WTelnetSocket2::Write(void const * buffer, int length)
{ 
  const unsigned char * base = (const unsigned char *)buffer;
  const unsigned char * next = base;
  int count = 0;
	        
   while (length > 0) {
    if (*next == '\r' && 
         !(length > 1 && next[1] == '\n') && !IsOurOption(TransmitBinary)) 
         {
         // send the characters
           if (!WTCPSocket::Write(base, (next - base) + 1))
              return false;
           count += lastWriteCount;
                     
           char null = '\0';
           if (!WTCPSocket::Write(&null, 1))
	      return false;
           count += lastWriteCount;
                                                          
           base = next+1;
         }
	                                                                 
         if (*next == IAC) 
	 {
           // send the characters
	   if (!WTCPSocket::Write(base, (next - base) + 1))
               return FALSE;
           count += lastWriteCount;
           base = next;
         }
                                                                                                                  
         next++;
         length--;
     }
     if (next > base) {
        if (!WTCPSocket::Write(base, next - base))
          return false;
       count += lastWriteCount;
                                                                                                                                        }
    lastWriteCount = count;
    return true;
}

bool WTelnetSocket2::Read(void * data, int bytesToRead)
{
 unsigned char buffer[256];
 int charsLeft = bytesToRead;
 unsigned char * dst = (unsigned char *)data;

 while (charsLeft > 0) 
 {
   unsigned char * src = buffer;
   if (!ReadBlock(src, charsLeft)) 
   {
      lastReadCount = bytesToRead - charsLeft;
      return lastReadCount > 0;
   }
   while (lastReadCount > 0) {
     unsigned int currentByte = *src++;
     lastReadCount--;
     switch (state) 
     {
         case StateCarriageReturn :
              state = StateNormal;
              if (currentByte == '\0')
                break; // Ignore \0 after CR
            // Else, fall through for normal processing

         case StateNormal :
              if (currentByte == IAC)
                state = StateIAC;
              else 
	      {
                if (currentByte == '\r' && !IsTheirOption(TransmitBinary))
                     state = StateCarriageReturn;
                *dst++ = currentByte;
                charsLeft--;
              }
	      break;

         case StateIAC :
             switch (currentByte) {
               case IAC :
                  state = StateNormal;
                  *dst++ = IAC;
                  charsLeft--;
                  break;
               case DO :
                  state = StateDo;
                  break;
               case DONT :
                  state = StateDont;
                  break;
               case WILL :
                  state = StateWill;
                  break;
               case WONT :
	          state = StateWont;
	          break;
               case DataMark :    // data stream portion of a Synch
               /* We may have missed an urgent notification, so make sure we
                 flush whatever is in the buffer currently.
               */
                   if (synchronising > 0)
                           synchronising--;
                   break;
               case SB :          // subnegotiation start
                   state = StateSubNegotiations;
                   subOption.resize(0);
                   break;
               default:
                    state = StateNormal;
                    break;
             } // switch currentByte
             break;
          case StateDo :
              OnDo(currentByte);
	      state = StateNormal;
              break;
          case StateDont :
              OnDont(currentByte);
              state = StateNormal;
              break;
          case StateWill :
              OnWill(currentByte);
              state = StateNormal;
              break;
          case StateWont :
              OnWont(currentByte);
              state = StateNormal;
              break;
          case StateSubNegotiations :
              if (currentByte == IAC)
                state = StateEndNegotiations;
              else
                subOption[subOption.size()] = currentByte;
              break;
          case StateEndNegotiations :
              if (currentByte == SE)
                state = StateNormal;
              else if (currentByte != IAC) 
	      {
                 /* This is an error.  We only expect to get "IAC IAC" or "IAC SE".
                    Several things may have happend.  An IAC was not doubled, the
                    IAC SE was left off, or another option got inserted into the
                    suboption are all possibilities. If we assume that the IAC was
                    not doubled, and really the IAC SE was left off, we could get
                    into an infinate loop here.  So, instead, we terminate the
                    suboption, and process the partial suboption if we can.
                  */
                  state = StateIAC;
                  src--;  // Go back to character for IAC ccommand
              }
              else 
	      {
                  subOption[subOption.size()] = currentByte;
                  state = StateSubNegotiations;
                  break;  // Was IAC IAC, subnegotiation not over yet.
              }
              if (subOption.size() > 1 && IsOurOption(subOption[0]))
                  OnSubOption(subOption[0],
                           &(subOption[1]), subOption.size()-1);
                  break;
          default :
             state = StateNormal;
     }
     if (synchronising > 0) 
     {
         charsLeft = bytesToRead;    // Flush data being received.
         dst = (unsigned char *)data;
     }
   } // while  LastReadCount
 } // while
 lastReadCount = bytesToRead;
 return true;
}

int WTelnetSocket2::ReadChar()
{ 
  lastReadCount = 0;
  unsigned char c;
  bool retVal = Read(&c, 1);
  return (retVal && lastReadCount == 1) ? c : -1;
}   

void WTelnetSocket2::OnDont(unsigned char code)
{
WTRACE(3, "OnDont. code = " << (int)code );
  OptionInfo & opt = option[code];
  switch (opt.ourState) {
    case OptionInfo::IsNo :
      break;
    case OptionInfo::IsYes :
      opt.ourState = OptionInfo::IsNo;
      SendCommand(WONT, code);
      break;
    case OptionInfo::WantNo :
      opt.ourState = OptionInfo::IsNo;
      break;
    case OptionInfo::WantNoQueued :
      opt.ourState = OptionInfo::WantYes;
      SendCommand(DO, code);
      break;
    case OptionInfo::WantYes :
      opt.ourState = OptionInfo::IsNo;
      break;
    case OptionInfo::WantYesQueued :
      opt.ourState = OptionInfo::IsNo;
      break;
   }
}

void WTelnetSocket2::OnDo(unsigned char code)
{
WTRACE(3, "OnDo. code = " << (int)code );
 OptionInfo & opt = option[code];
 switch (opt.ourState) {
    case OptionInfo::IsNo :
       if (opt.weCan) 
       {
         SendCommand(WILL, code);
         opt.ourState = OptionInfo::IsYes;
       } 
       else 
       {
          SendCommand(WONT, code);
       } 
       break;
    case OptionInfo::IsYes :
       break;
    case OptionInfo::WantNo :
       opt.ourState = OptionInfo::IsNo;
       break;
    case OptionInfo::WantNoQueued :
       opt.ourState = OptionInfo::IsYes;
       break;
    case OptionInfo::WantYes :
       opt.ourState = OptionInfo::IsYes;
       break;
    case OptionInfo::WantYesQueued :
       opt.ourState = OptionInfo::WantNo;
       SendCommand(WONT, code);
       break;
 }   
 if (IsOurOption(code)) {
      switch (code) {
         case TerminalSpeed : {
           static unsigned char defSpeed[] = "38400,38400";
           SendSubOption(TerminalSpeed,defSpeed,sizeof(defSpeed)-1,SubOptionIs);
           }
           break;
         case TerminalType :
           SendSubOption(TerminalType, (const unsigned char *)terminalType.c_str() , terminalType.size(), SubOptionIs);
           break;
         case WindowSize :
           SetWindowSize(windowWidth, windowHeight);
           break;
     } // switch
 }  // if
}

void WTelnetSocket2::OnWill(unsigned int code)
{     
 WTRACE(3, "OnWill. code = " << (int)code );
 OptionInfo & opt = option[code];
 switch (opt.theirState) {
     case OptionInfo::IsNo :
        if (opt.theyShould) 
	{
           SendCommand(DO, code);
           opt.theirState = OptionInfo::IsYes;
        }
        else 
	{
           SendCommand(DONT, code);
        }
        break;
     case OptionInfo::IsYes :
        break;
     case OptionInfo::WantNo :
        opt.theirState = OptionInfo::IsNo;
        break;
     case OptionInfo::WantNoQueued :
        opt.theirState = OptionInfo::IsYes;
        break;
     case OptionInfo::WantYes :
        opt.theirState = OptionInfo::IsYes;
        break;
     case OptionInfo::WantYesQueued :
        opt.theirState = OptionInfo::WantNo;
        SendCommand(DONT, code);
        break;
 }
}

void WTelnetSocket2::OnWont(unsigned char code)
{
 WTRACE(3, "OnWont. code = " << (int)code );
 OptionInfo & opt = option[code];
 switch (opt.theirState) {
    case OptionInfo::IsNo :
      break;
    case OptionInfo::IsYes :
      opt.theirState = OptionInfo::IsNo;
      SendCommand(DONT, code);
      break;
    case OptionInfo::WantNo :
      opt.theirState = OptionInfo::IsNo;
      break;
    case OptionInfo::WantNoQueued :
      opt.theirState = OptionInfo::WantYes;
      SendCommand(DO, code);
      break;
    case OptionInfo::WantYes :
      opt.theirState = OptionInfo::IsNo;
      break;
    case OptionInfo::WantYesQueued :
      opt.theirState = OptionInfo::IsNo;
      break;
 }
}

void WTelnetSocket2::OnSubOption(unsigned char code, const unsigned char * info, int len)
{
 WTRACE(3, "OnSubOptions. code = " << (int)code );
 switch (code) {
   case TerminalType :
     if (*info == SubOptionSend)
        SendSubOption(TerminalType, (const unsigned char *) terminalType.c_str(), terminalType.size(), SubOptionIs);
        break;
   case TerminalSpeed :
     if (*info == SubOptionSend) 
     {
        static unsigned char defSpeed[] = "38400,38400";
        SendSubOption(TerminalSpeed,defSpeed,sizeof(defSpeed)-1,SubOptionIs);
     }
     break;
 }
}

bool WTelnetSocket2::SendDo(unsigned char code)
{
  if (!IsOpen())
  { 
    WTRACE(1, "SenDo fails for code " << code);
    return false;
  }
  WTRACE(3, "SendDO. code = " << (int)code);
  OptionInfo & opt = option[code];
	      
   switch (opt.theirState) {
       case OptionInfo::IsNo :
          SendCommand(DO, code);
          opt.theirState = OptionInfo::WantYes;
          break;
							      
       case OptionInfo::IsYes :
          return false;
						          
       case OptionInfo::WantNo :
          opt.theirState = OptionInfo::WantNoQueued;
          break;
						    
       case OptionInfo::WantNoQueued :
          opt.theirState = OptionInfo::IsNo;
          return false;
												      
       case OptionInfo::WantYes :
          opt.theirState = OptionInfo::IsNo;
          return false;

       case OptionInfo::WantYesQueued :
          opt.theirState = OptionInfo::WantYes;
          break;
   }
  return true;
}

bool WTelnetSocket2::SendDont(unsigned char code)
{
 if (!IsOpen())
 {
    WTRACE(1, "SenDont fails for code " << code);
    return false;
 }

 OptionInfo & opt = option[code];

WTRACE(3, "SendDont. code = " << (int)code << ", theirOption = " << opt.theirState) ;
WTRACE(3, "IsNo = " << OptionInfo::IsNo);

 switch (opt.theirState) {
     case OptionInfo::IsNo :
        return false;
     case OptionInfo::IsYes :
         SendCommand(DONT, code);
         opt.theirState = OptionInfo::WantNo;
         break;
     case OptionInfo::WantNo :
         opt.theirState = OptionInfo::IsNo;
         return false;
     case OptionInfo::WantNoQueued :
         opt.theirState = OptionInfo::WantNo;
         break;
     case OptionInfo::WantYes :
         opt.theirState = OptionInfo::WantYesQueued;
         break;
     case OptionInfo::WantYesQueued :
         opt.theirState = OptionInfo::IsYes;
         return false;
 }
 return true;
}

bool WTelnetSocket2::SendWill(unsigned char code)
{   
  if (!IsOpen())
  {
      WTRACE(1, "SenWill fails for code " << code);
      return false;
  }
WTRACE(3, "SendWill. code = " << (int)code);
	          
  OptionInfo & opt = option[code];
	            
  switch (opt.ourState) {
    case OptionInfo::IsNo :
      SendCommand(WILL, code);
      opt.ourState = OptionInfo::WantYes;
      break;

    case OptionInfo::IsYes :
      return false;
									  
    case OptionInfo::WantNo :
      opt.ourState = OptionInfo::WantNoQueued;
      break;

   case OptionInfo::WantNoQueued :
      opt.ourState = OptionInfo::IsNo;
      return false;
													        
   case OptionInfo::WantYes :
      opt.ourState = OptionInfo::IsNo;
      return false;

   case OptionInfo::WantYesQueued :
      opt.ourState = OptionInfo::WantYes;
      break;
  }
  return true;
}

bool WTelnetSocket2::SendWont(unsigned char code)
{
  if (!IsOpen())
  {
    WTRACE(1, "SenWont fails for code " << (int)code);
    return false;
  }
  WTRACE(3, "SendWont. code = " << (int)code);

  OptionInfo & opt = option[code];

  switch (opt.ourState) {
     case OptionInfo::IsNo :
        return false;
     case OptionInfo::IsYes :
        SendCommand(WONT, code);
        opt.ourState = OptionInfo::WantNo;
        break;
     case OptionInfo::WantNo :
        opt.ourState = OptionInfo::IsNo;
        return false;
     case OptionInfo::WantNoQueued :
        opt.ourState = OptionInfo::WantNo;
        break;
     case OptionInfo::WantYes :
        opt.ourState = OptionInfo::WantYesQueued;
        break;
     case OptionInfo::WantYesQueued :
        opt.ourState = OptionInfo::IsYes;
        return false;
   }
  return true;
}

bool WTelnetSocket2::SendSubOption(unsigned char code, const unsigned char * info, int len, int subCode)
{
WTRACE(3, "SendSubOption. info = " << info);
 if (!IsOpen())
   return FALSE;
 unsigned char buffer[512];
 buffer[0] = IAC;
 buffer[1] = SB;
 buffer[2] = code;
 int i = 3;
 if (subCode >= 0)
   buffer[i++] = (unsigned char)subCode;
 while (len-- > 0) 
 {
   if (*info == IAC)
      buffer[i++] = IAC;
   buffer[i++] = *info++;
 }
 buffer[i++] = IAC;
 buffer[i++] = SE;
 return WTCPSocket::Write((const unsigned char *)buffer, i);
}

void WTelnetSocket2::SetWindowSize(int width, int height)
{
  windowWidth = width;
  windowHeight = height;
  if (IsOurOption(WindowSize)) {
     unsigned char buffer[4];
     buffer[0] = (unsigned char)(width >> 8);
     buffer[1] = (unsigned char)width;
     buffer[2] = (unsigned char)(height >> 8);
     buffer[3] = (unsigned char)height;
     SendSubOption(WindowSize, buffer, 4);
  }
  else {
    SetOurOption(WindowSize);
    SendWill(WindowSize);
  }
}
