/*
*
* versuch project
* Sockets header
* $Id: Sockets.h,v 1.2 2006/05/06 13:17:54 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#ifndef _WSOCKETS_H
#define _WSOCKETS_H

// #define    FD_SETSIZE      4098U
// old limit in 1024 for select() is too low

#include <string>
#include <iostream>
#include <vector>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

class WSocket
{
public:
 class Address {
  public:
   Address();
   Address(unsigned char b1, unsigned char b2, unsigned char b3, unsigned char b4);
   Address(const in_addr & addr);
   Address(const char * str_addr);
   Address(const u_int32_t & addr);
   Address & operator=(const in_addr & addr);
   operator in_addr() const;
   const char * AsString();
   bool operator==(in_addr & addr) const;
   bool operator==(const Address & addr) const;
   bool operator==(u_int32_t & addr) const;
   bool operator!=(const Address & addr)   const { return !operator==(addr);} ;
   bool operator!=(const in_addr & addr)   const { return !operator==(addr);} ;
   bool operator!=(const u_int32_t & addr) const { return !operator==(addr);} ;
   operator unsigned() const;
   unsigned char operator[](unsigned idx) const;
   unsigned char Byte1() const;
   unsigned char Byte2() const;
   unsigned char Byte3() const;
   unsigned char Byte4() const;
  protected:
   in_addr v;
   char strbuf[16];
   friend std::ostream & operator<<(std::ostream & s, Address & a);
 };
 WSocket();
 WSocket(int handle);
 virtual bool OpenSocket() = 0; ;
 void Close();
 bool Listen(const Address & bindAddr, unsigned newPort, bool Reusability = false);
 bool Bind(const Address & bindAddr, unsigned Port);
 bool Bind(const Address & iFace);
 bool GetLocalAddress(Address & addr, unsigned & portNum);
 bool GetLocalAddress(Address & addr);
 static int Select(WSocket & sock1, WSocket & sock2, const int sec);
 static int Select(WSocket & sock1,  const int sec = 0 );
 static int Select( std::vector<WSocket *> & SocketList, const int sec = 0 );
 int GetHandler() { return os_handle; };
 int GetLastReadCount() { return lastReadCount; };
 int GetLastWriteCount() { return lastWriteCount; };
 bool IsOpen() { return os_handle >= 0; };
 bool SetOption( int option, int value, int level = SOL_SOCKET);
 bool GetPeerAddress(Address & addr, unsigned & portNum);
 bool GetPeerAddress(Address & addr);
 bool Accept (WSocket & Listener);
 bool Connect(const Address & addr, unsigned Port);
 unsigned GetLocalPort() { return localPort ;};
 std::string OnPrint();
 static bool GetHostAddress(const char * name, Address & addr);
 bool SetReadTimeout(unsigned int sec);
 bool SetWriteTimeout(unsigned int sec);
 bool GetReadTimeout(unsigned int & sec);
 virtual const char * GetProtocolName() const = 0 ;
// virtual const char * GetProtocolName() const {  return "" ; } ;
protected:
 int            os_handle;
 unsigned       localPort, peerPort;
 Address	localAddress, peerAddress;
 int            lastReadCount, lastWriteCount ;
 void		os_close();
 int		backlogSize;
};

class WPipe : public WSocket
{
public:
 WPipe(int handle);
 virtual ~WPipe();
 virtual const char * GetProtocolName() const {  return "pipe" ; } ;
 virtual bool OpenSocket() { return true; } ;
 bool Write( const char * buf, unsigned len );
 bool Read( char * buf, unsigned & len );
};

class WUDPSocket : public WSocket
{
public:
 WUDPSocket();
 virtual ~WUDPSocket() {};
 bool ReadFrom(void * buf, unsigned len, Address & addr, unsigned & port);
 virtual bool OpenSocket() ;
 virtual bool OpenSocket(int ipAdressFamily);
 void SetSendAddress(const Address & newAddress, unsigned newPort);
 void GetSendAddress(Address & address, unsigned & port);
 bool Write( const void * buf, unsigned len );
 bool WriteTo( const void * buf, unsigned len, Address addr, unsigned int port);
 virtual const char * GetProtocolName() const {  return "udp" ; } ;
};

class WTCPSocket : public WSocket
{
public:
 WTCPSocket();
 WTCPSocket(unsigned Port);
 virtual ~WTCPSocket();
 virtual bool OpenSocket();
 bool ReadBlock(void * buf, unsigned len);
 bool Write( const void * buf, unsigned len );
 bool WriteString( const std::string Str0 );
 bool Listen(const Address & bindAddr, unsigned newPort, bool Reusability = false);
 int ReadChar();
 bool WriteOutOfBand(void const * buf, int len);
 virtual const char * GetProtocolName() const {  return "tcp" ; } ;
};

class WTelnetSocket : public WTCPSocket
{
public:
 WTelnetSocket();
 virtual ~WTelnetSocket() {};
 bool ReadBlock(void * buf, unsigned len);
};

//
// PWlib TelnetSocket based
//

class WTelnetSocket2 : public WTCPSocket
{
public:
 WTelnetSocket2();
 virtual ~WTelnetSocket2() {};

 enum Command {
   IAC           = 255,    // Interpret As Command - escape character.
   DONT          = 254,    // You are not to use option.
   DO            = 253,    // Request to use option.
   WONT          = 252,    // Refuse use of option.
   WILL          = 251,    // Accept the use of option.
   SB            = 250,    // Subnegotiation begin.
   GoAhead       = 249,    // Function GA, you may reverse the line.
   EraseLine     = 248,    // Function EL, erase the current line.
   EraseChar     = 247,    // Function EC, erase the current character.
   AreYouThere   = 246,    // Function AYT, are you there?
   AbortOutput   = 245,    // Function AO, abort output stream.
   InterruptProcess = 244, // Function IP, interrupt process, permanently.
   Break         = 243,    // NVT character break.
   DataMark      = 242,    // Marker for connection cleaning.
   NOP           = 241,    // No operation.
   SE            = 240,    // Subnegotiation end.
   EndOfReccord  = 239,    // End of record for transparent mode.
   AbortProcess  = 238,    // Abort the entire process
   SuspendProcess= 237,    // Suspend the process.
   EndOfFile     = 236     // End of file marker.
 };
 enum Options {
   TransmitBinary      = 0,    // Assume binary 8 bit data is transferred.
   EchoOption          = 1,    // Automatically echo characters sent.
   ReconnectOption     = 2,    // Prepare to reconnect
   SuppressGoAhead     = 3,    // Do not use the GA protocol.
   MessageSizeOption   = 4,    // Negatiate approximate message size
   StatusOption        = 5,    // Status packets are understood.
   TimingMark          = 6,    // Marker for synchronisation.
   RCTEOption          = 7,    // Remote controlled transmission and echo.
   OutputLineWidth     = 8,    // Negotiate about output line width.
   OutputPageSize      = 9,    // Negotiate about output page size.
   CRDisposition       = 10,   // Negotiate about CR disposition.
   HorizontalTabsStops = 11,   // Negotiate about horizontal tabstops.
   HorizTabDisposition = 12,   // Negotiate about horizontal tab disposition
   FormFeedDisposition = 13,   // Negotiate about formfeed disposition.
   VerticalTabStops    = 14,   // Negotiate about vertical tab stops.
   VertTabDisposition  = 15,   // Negotiate about vertical tab disposition.
   LineFeedDisposition = 16,   // Negotiate about output LF disposition.
   ExtendedASCII       = 17,   // Extended ascic character set.
   ForceLogout         = 18,   // Force logout.
   ByteMacroOption     = 19,   // Byte macro.
   DataEntryTerminal   = 20,   // data entry terminal.
   SupDupProtocol      = 21,   // supdup protocol.
   SupDupOutput        = 22,   // supdup output.
   SendLocation        = 23,   // Send location.
   TerminalType        = 24,   // Provide terminal type information.
   EndOfRecordOption   = 25,   // Record boundary marker.
   TACACSUID           = 26,   // TACACS user identification.
   OutputMark          = 27,   // Output marker or banner text.
   TerminalLocation    = 28,   // Terminals physical location infromation.
   Use3270RegimeOption = 29,   // 3270 regime.
   UseX3PADOption      = 30,   // X.3 PAD
   WindowSize          = 31,   // NAWS - Negotiate About Window Size.
   TerminalSpeed       = 32,   // Provide terminal speed information.
   FlowControl         = 33,   // Remote flow control.
   LineModeOption      = 34,   // Terminal in line mode option.
   XDisplayLocation    = 35,   // X Display location.
   EnvironmentOption   = 36,   // Provide environment information.
   AuthenticateOption  = 37,   // Authenticate option.
   EncriptionOption    = 38,   // Encryption option.
   ExtendedOptionsList = 255,  // Code for extended options.
   MaxOptions
 };
 bool SendCommand( Command cmd, int opt = 0 );
 bool SendSubOption(unsigned char code, const unsigned char * info, int len, int subCode = -1);
 void SetOurOption( unsigned char code, bool state = true ) { option[code].weCan = state; }
 bool IsOurOption( unsigned int code ) const { return option[code].ourState == OptionInfo::IsYes; }
 void SetTheirOption( unsigned char code, bool state = true ) { option[code].theyShould = state; }
 bool IsTheirOption( unsigned char code ) const { return option[code].theirState == OptionInfo::IsYes; }

 bool Accept(WSocket & sock);
 bool SendDo(unsigned char code);
 bool SendDont(unsigned char code);
 bool SendWill(unsigned char code);
 bool SendWont(unsigned char code);
 bool Write(void const * buffer, int length);
 bool Read(void * data, int bytesToRead);
 int ReadChar();

 void OnDont(unsigned char code);
 void OnDo(unsigned char code);
 void OnWill(unsigned int code);
 void OnWont(unsigned char code);
 void OnSubOption(unsigned char code, const unsigned char * info, int len);
 enum SubOptionCodes 
 {
   SubOptionIs       = 0,  // Sub-option is...
   SubOptionSend     = 1,  // Request to send option.
 };

 void SetWindowSize(int width, int height);
protected:
 void Construct();
 struct OptionInfo {
   enum {
      IsNo, IsYes, WantNo, WantNoQueued, WantYes, WantYesQueued
   };
   unsigned weCan:1;      // We can do the option if they want us to do.
   unsigned ourState:3;
   unsigned theyShould:1; // They should if they will.
   unsigned theirState:3;
 };
 OptionInfo option[MaxOptions];

private:
 unsigned synchronising;
 std::string terminalType;
 int windowWidth, windowHeight;
 enum State {
   StateNormal,
   StateCarriageReturn,
   StateIAC,
   StateDo,
   StateDont,
   StateWill,
   StateWont,
   StateSubNegotiations,
   StateEndNegotiations
 };
 State state;
 std::vector<unsigned char> subOption;
 bool ReadBlock(void * buf, unsigned len);
};

#endif // _WSOCKETS_H
