/*
*
* versuch project
* several things for replace in pwlibs
* $Id: Utils.cxx,v 1.4 2006/08/24 09:47:48 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#include "Utils.h"

#include <fstream>
#include <assert.h>

void wtrim(string& str) {
 const char* sp_chars = " \t\r\n";
 unsigned int fst_notsp = str.find_first_not_of(sp_chars);
 unsigned int lst_notsp = str.find_last_not_of(sp_chars);
 string res_str;
  if(fst_notsp != string::npos) {
     assert(lst_notsp >= fst_notsp);
     res_str = str.substr(fst_notsp, lst_notsp - fst_notsp + 1);
  }
  str =  res_str;
}

int str2int( const string str) {
 const char* s0 = str.c_str();
 int i = atoi(s0);
 return i;
}

TConfig::TConfig(const string& filename, const string& section)
{
  fname = filename;
  CurrentSection = section;
  string secname;

  ifstream f(filename.c_str());
  if(!f) {
    cerr << "erorr in an opening of file : " << filename << flush << endl;
    assert(false);
  }
  string str;
  while(!f.eof()) {
        getline(f, str);
        wtrim(str);
        if(str.empty() || str[0] == '#')
                continue;
	transform(str.begin(), str.end(), str.begin(), ::tolower );
        unsigned int p0, p1;
        p0 = str.find('[');
        p1 = str.find(']');
        if ( p0 != string::npos && p1 != string::npos) {
          secname = str.substr(p0+1,p1-1);
          continue;
        }
        if ( (p0 != string::npos && p1 == string::npos) || (p0 == string::npos && p1 != string::npos) ) {
                cerr << "erorr in [section] syntax: " << endl <<  str << endl;
        }
        p0 = str.find('=');
        if (p0 != string::npos) {
          string key = str.substr(0, p0);
          string val = "";
          if(p0 != str.length()-1)
            val = str.substr(p0+1);
          m_config[secname].insert(pair<string, string>(key,val));
          continue ;
        }
        cerr << "erorr in syntax: " << endl <<  str << endl;
        assert(false);
  }  // while
}

string TConfig::GetString (const string &key, const string &dflt) const
{
  TSection m_vals;
  multimap<string, TSection>::const_iterator it1 = m_config.find(CurrentSection);
  if ( it1 == m_config.end() )
    return dflt;
  m_vals = it1->second;
  TSection::const_iterator it = m_vals.find(key);
  string str;
  if(it == m_vals.end())
     str = dflt;
  else
    str = it->second;
  return str;
}

void TConfig::SetDefaultSection (const string &section)
{
  CurrentSection = section;
}

long TConfig::GetInteger (const string &key, long dflt) const
{
  TSection m_vals;
  map<string, TSection>::const_iterator it1 = m_config.find(CurrentSection);
  if ( it1 == m_config.end() )
    return dflt;
  m_vals = it1->second;
  TSection::const_iterator it = m_vals.find(key);
  long res;
  if(it == m_vals.end()) 
    res = dflt;
  else
    res = str2int(it->second);
  return res;
}

unsigned TConfig::GetVector(const string &key, vector<string>& params) const
{
  TSection m_vals;
  multimap<string, TSection>::const_iterator it1 = m_config.find(CurrentSection);
  if ( it1 == m_config.end() )
    return 0;
  m_vals = it1->second;
  multimap<string, string>::iterator it;
  for (it = m_vals.begin(); it != m_vals.end(); ++it)
  {
    if ( it->first == key )
      params.push_back(it->second);
  }
  return params.size();
}

bool TConfig::GetBool(const string &key, const bool dflt)
{
  TSection m_vals;
  map<string, TSection>::const_iterator it1 = m_config.find(CurrentSection);
  if ( it1 == m_config.end() )
    return dflt;
  m_vals = it1->second;
  TSection::const_iterator it = m_vals.find(key);
  bool res;
  if(it == m_vals.end())
    res = dflt;
  else 
  {
    if ( (it->second == "YES") || (it->second == "yes") ||(it->second == "OK") ||(it->second == "ok") ) 
    {
      res = true;
    }
    else
    {
      res = false;
    }
  }
  return res;
}

// WTRACE 
static ostream * WTraceStream = &cerr;
static unsigned WTraceOptions = WTrace::FileAndLine;
static unsigned dlevel;
static ofstream * traceOutput;

void WTrace::Initialise(const char * filename, unsigned EDebugOptions)
{
  ios_base::openmode mode;
  WTraceOptions = EDebugOptions;
  if (filename == NULL) 
    return;
  if (EDebugOptions & AppendToFile)
    mode = ofstream::out | ofstream::app;
  else
    mode = ofstream::out | ofstream::trunc;
  traceOutput->open(filename, mode);
  if (traceOutput->is_open())
    WTrace::SetStream(traceOutput);
  else {
    WTRACE(0, "Could not open trace output file \"" << filename << '"');
    delete traceOutput;
  }
}

void WTrace::SetOptions(unsigned EDebugOptions)
{
  WTraceOptions |= EDebugOptions;
}

void WTrace::ClearOptions(unsigned EDebugOptions)
{
  WTraceOptions &= ~EDebugOptions;
}

bool WTrace::CanTrace(unsigned level)
{
 if ( level <= dlevel )
  return true;
 else
  return false;
}

void WTrace::SetLevel(unsigned level)
{
  dlevel = level;
}

unsigned WTrace::GetLevel()
{
 return dlevel;
}

void WTrace::SetStream(ostream * s)
{
 if (s == NULL)
     s = &cerr;
}

ostream & WTrace::Begin(unsigned level, const char * fileName, int lineNum)
{
  if (level == UINT_MAX)
    return *WTraceStream;
  if ((WTraceOptions & DateAndTime) != 0) {
    time_t t;
    t = time(NULL);
    string now = ctime(&t);
    wtrim(now);
    *WTraceStream << now << '\t' ;
  }
  if ((WTraceOptions & Thread) != 0) {
    *WTraceStream << pthread_self();
    *WTraceStream << '\t';
  }
  if ((WTraceOptions & TraceLevel) != 0)
    *WTraceStream << level << '\t';
  if ((WTraceOptions & FileAndLine) != 0 && fileName != NULL) {
    const char * file = strrchr(fileName, '/');
    if (file != NULL)
      file++;
    else {
      file = strrchr(fileName, '\\');
      if (file != NULL)
        file++;
      else
        file = fileName;
    }
    *WTraceStream << setw(16) << file << '(' << setbase(10) << lineNum << ")\t";
  }
  return *WTraceStream;
}

ostream & WTrace::End(ostream & s)
{
  s << endl;
  return s;
}

void put_empty(int n, vector<string>& tokens)
{
 for (int i = 0; i < n; i++) {
  tokens.push_back("");
 }
}

void Tokenize(const string& str,
              vector<string>& tokens,
              const string& delimiters,
	      bool onePerSeparator)
{
  // Skip delimiters at beginning.
  string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  if (onePerSeparator)
    put_empty(lastPos - 1, tokens);
  // Find first "non-delimiter".
  string::size_type pos     = str.find_first_of(delimiters, lastPos);

  while (string::npos != pos || string::npos != lastPos)
  {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(lastPos, pos - lastPos));
    // Skip delimiters.  Note the "not_of"
    lastPos = str.find_first_not_of(delimiters, pos);
    if (onePerSeparator)
    {
// cerr << "pos = " << pos << " lasPos = " << lastPos << endl;
       put_empty(lastPos - pos - 1, tokens);
    }
    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, lastPos);
  }
}


