/*
*
* versuch project
* global objects
* $Id: global.cxx,v 1.11 2006/09/04 14:21:47 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#include <ptlib.h>
#include <assert.h>
#include <fstream>
#include <sstream>

#include "global.h"
#include "Utils.h"

TGlobal::TGlobal() 
{
 Gateways = new TGateways();
 DialPeers = new TDialPeers();
 pthread_mutex_init(&global_mtx, NULL);
}

TGlobal::~TGlobal()
{
 delete Gateways;
 delete DialPeers;
 if (RTPPortRange != NULL)
   delete RTPPortRange;
 if (H245PortRange != NULL)
   delete H245PortRange;
 if (cfg != NULL)
   delete cfg;
 pthread_mutex_destroy(&global_mtx);
}

bool TGlobal::OpenCfg(const PFilePath & fp)
{
 cfg = new TConfig((const char *) fp, "main");
 return true;
}

bool TGlobal::ReadMap(const string FileName, map<string, string> & mMap)
{
 ifstream f(FileName.c_str());
 if(!f) {
   WTRACE(1, "erorr in opening file : " << FileName );
   return(false);
 }
 string Str0;
 vector<string> tokens;
 while(!f.eof()) 
 {
   tokens.clear();
   getline(f, Str0);
   wtrim(Str0);
   if ( Str0.empty() || Str0[0] == '#')
       continue;
   Tokenize(Str0, tokens, "=,:-/'");
   mMap[tokens[0]] = tokens[1];
 }
 return(true);
}

bool TGlobal::IsBlockingCause(int Code)
{
 ostringstream strm0;
 strm0 << Code;
 string sCode = strm0.str();
 unsigned int i;
 for (i = 0; i < BlockRoutingCauses.size(); i++)
 {
   if ( BlockRoutingCauses[i] == sCode ) 
   {
     WTRACE(1, "The cause " << Code << " blocks routing election");
     return true;
   }
 } 
 return false;
}

bool TGlobal::ReadFile()
{
  WTRACE(1, "Reading config file " << fp );
  if ( ! OpenCfg(fp)) {
    cerr << "Config file " << fp << " read error " << endl;
    return false;
  }
  string str0;
// Local address for binding
  cfg->SetDefaultSection("main");
  PString LA = cfg->GetString( "localaddress", "" );
  if ( LA.IsEmpty() )
    ListenAddr = INADDR_ANY;
  else
    WSocket::GetHostAddress(LA, ListenAddr);
// Address for outgoing connections
  LA = cfg->GetString( "outgoingaddress", "" );
  if ( LA.IsEmpty() )
    OutAddr = INADDR_ANY;
  else
    WSocket::GetHostAddress(LA, OutAddr);

// Port bases resolving
  RTPPortRange = new TUDPPortRange(cfg->GetString("rtpportrange","10000-59000"));
  H245PortRange = new TTCPPortRange(cfg->GetString("H245portrange","10000-59000"));
// Other options
  MinDigitsDefault =  cfg->GetInteger("defaultmindigits", 0);
  ID = cfg->GetString( "id", "default" );
  StatusPort = cfg->GetInteger( "statusport", 2016 );
  cfg->GetVector("incomingtranslatecalled", IncTlCalled);
  BlockG711 = cfg->GetBool("blockg711", false);
  GatewaysConf = cfg->GetString( "gatewaysconfig", "" );
  if (BlockRoutingCauses.size() != 0)
     BlockRoutingCauses.clear();
  str0 = cfg->GetString("blockroutingcauses","");
  if (str0 == "") 
  {
     WTRACE(1,"BlockRoutingCauses is empty");
  }
  Tokenize(str0, BlockRoutingCauses, ",:-/'");
  LA = cfg->GetString( "debugsourcehost", "0.0.0.0" );
  WSocket::GetHostAddress(LA, DebugSourceHost);

// Accounting section
  cfg->SetDefaultSection("accounting");
  AcctType = cfg->GetString("type","cdr");
  AcctRadHost = cfg->GetString("radiushost","127.0.0.1");
  AcctRadPort = cfg->GetInteger("port",1813);
  AcctRadSecret = cfg->GetString("secret","");
  AcctRadMode = cfg->GetInteger("radmode", 2);
  RadStopOnly = cfg->GetBool("radstoponly", false);
  AcctVSAPrefix = cfg->GetString("vsaprefix","");
  LogFile = cfg->GetString("logfile","/dev/null");
  LogRotateSpan = cfg->GetInteger("logrotatetime", 60);
  LogUmask = cfg->GetString("logumask","0600");

// Auth Section
  cfg->SetDefaultSection("auth");
  AuthType = cfg->GetString("type","text");
  AuthRadHost = cfg->GetString("radiushost","127.0.0.1");
  AuthRadPort = cfg->GetInteger("port",1812);
  AuthRadSecret = cfg->GetString("secret","");
  AuthVSAPrefix = cfg->GetString("vsaprefix","");
  PeerFile = cfg->GetString("peerfile","");

// Gateways config extra section
  if ( GatewaysConf != "" )
  {
   Gateways->ReadFile( GatewaysConf );
  }
  if (AuthType == "text")
  {
    if ( PeerFile == "")
    {
      cerr << "PeerFile info is empty " << endl;
      return false;
    }
    DialPeers->ReadFile(PeerFile);
  }
 return true;
}

//
// TPortRange
//

TPortRange::TPortRange(const PString sRange)
{
 PStringArray cfgs = sRange.Tokenise(",.:-/'", FALSE);
 if (cfgs.GetSize() >= 2)
 {
    LastPort = MinPort    = cfgs[0].AsUnsigned();
    MaxPort               = cfgs[1].AsUnsigned();
 }
 else
 {
   WTRACE(1, "Wrong format of PortRange in cfg file :" << sRange);
   assert(false);
 }
}

int TPortRange::GetFree(bool IsTCP)
{
 int res, fport, i;
 int sock;
 res = -1;
 fport = LastPort;
 if (IsTCP)
 {
  sock = socket ( AF_INET, SOCK_STREAM, 0 );
 }
 else
 {
  sock = socket ( AF_INET, SOCK_DGRAM, 0 );
 }
 if ( sock == -1 )
   return -1;
 for (i = 0; i < 100; i++)
 {
   sockaddr_in addr;
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = INADDR_ANY;
   addr.sin_port = htons ( fport );
   int bind_return = bind ( sock, ( struct sockaddr * ) &addr, sizeof ( addr ) );
   if( bind_return == -1 )
   {
      switch(errno)
      {
        case EBADF: case EINVAL: case ENOTSOCK: return -1;
        default: break;
      }
   }
   else
     if( bind_return == 0 )
     {
        res = fport;
        break;
     }
   fport++;
   if (fport >= MaxPort)
     fport = MinPort;
 }
 if( close( sock ) != 0 )
     res = -1;
 LastPort = fport;
 return res;
}

TUDPPortRange::TUDPPortRange(const PString sRange) : TPortRange(sRange)
{
}

int TUDPPortRange::GetFreePort()
{
 LastPort = LastPort >> 2;
 LastPort = ++LastPort << 2;
 return GetFree(false);
}

TTCPPortRange::TTCPPortRange(const PString sRange) : TPortRange(sRange)
{
}

int TTCPPortRange::GetFreePort()
{
 return GetFree(true);
}

void TMessLog::Add(const char * Level, const char * Text, bool Sender)
{
 char buf[100];
 sprintf(buf, "%i:%s:%s:%s", (int)time(NULL) , Level, Text, ( Sender ? "Caller" : "Called" ));
 string new_entry= buf;
 mLog.push_back(new_entry);
}

string TMessLog::Output()
{
 string res;
 vector<string>::iterator it;
 res = "|";
 for (it=mLog.begin(); it != mLog.end(); it++){
  res += *it + ",";
 }
 return res;
}

//
//  TGateways
//

TGateways::~TGateways()
{
 EraseData();
}

void TGateways::EraseData()
{
  map<string, TGateDescr*>::iterator it;
  for (it = gwList.begin(); it != gwList.end(); ++it)
  {
     delete it->second;
  }
  gwList.clear();
}

void TGateways::ReadFile(const string FileName)
{
 ifstream f(FileName.c_str());
 if(!f) {
   WTRACE(1, "erorr in opening file : " << FileName );
   assert(false);
 }
 EraseData();
 string Str0, TheGW;
 vector<string> tokens;
 unsigned int p0, p1;
 TheGW = "";
 while(!f.eof())
 {
   tokens.clear();
   getline(f, Str0);
   wtrim(Str0);
   if ( Str0.empty() || Str0[0] == '#')
      continue;
   p0 = Str0.find('[');
   p1 = Str0.find(']');
   if ( (p0 != string::npos && p1 == string::npos) || (p0 == string::npos && p1 != string::npos) ) 
   {
       cerr << "erorr in [section] syntax: " << endl <<  Str0 << endl;
       continue;
   }
   if ( p0 != string::npos && p1 != string::npos)
   {
     TheGW = Str0.substr(p0+1,p1-1);
     if ( gwList[TheGW] == NULL )
     {
        TGateDescr * mGW = new TGateDescr();
        gwList[TheGW] = mGW;
     }
     continue;
   }
   if ( TheGW == "" )
   {
       cerr << "Empty GW defenition " << endl;
       continue;
   }
   Tokenize(Str0, tokens, "=,:-/'");
   if ( ! AddPair( TheGW, tokens[0], tokens[1]) )
   {
     cerr << "An adding field fails. GW = " << TheGW << " Key = " << tokens[0] << " Field " << tokens[1] << endl;
     continue;
   }

  } // while
}

bool TGateways::AddPair(const string GW, string & Key, const string Field)
{
 transform(Key.begin(), Key.end(), Key.begin(), ::tolower );
 if (Key == "address" )
 {
   gwList[GW] -> Address = Field ;
   return true;
 }
 if (Key == "defer" )
 {
   gwList[GW] -> DeferMsg = Field ;
   return true;
 }
 if (Key == "capacity" )
 {
   gwList[GW] -> Capacity = strtoll( Field.c_str(), NULL,10) ;
   return true;
 }
 return false;
}

//
// TDialPeers
//

TVoicePeer::TVoicePeer()
{
// default values
 TargetGW = "";
 DestPattern = "";
 dst_translate = "";
 dst_bill_translate = "";
 src_translate = "";
 src_bill_translate = "";
 Preference = 0;
 HuntStop = 0;
 Proxy = 1;
}

TDialPeers::~TDialPeers()
{
 EraseData();
}

void TDialPeers::EraseData()
{
  map<string, TVoicePeer*>::iterator it;
  for (it = List.begin(); it != List.end(); ++it)
  {
      delete it->second;
  }
  List.clear();
}

bool TDialPeers::ReadFile(const string FileName)
{
 ifstream f(FileName.c_str());
 if(!f) { 
    WTRACE(1, "erorr in opening file : " << FileName );
    return false;
 } 
 EraseData();
 string Str0, ThePeer;
 vector<string> tokens;
 unsigned int p0, p1;
 ThePeer = "";
 while(!f.eof())
 {
   tokens.clear();
   getline(f, Str0);
   wtrim(Str0);
   if ( Str0.empty() || Str0[0] == '#')
      continue;
   p0 = Str0.find('[');
   p1 = Str0.find(']');
   if ( (p0 != string::npos && p1 == string::npos) || (p0 == string::npos && p1 != string::npos) )
   {
      cerr << "erorr in [section] syntax: " << endl <<  Str0 << endl;
      continue;
   }
   if ( p0 != string::npos && p1 != string::npos)
   {
      ThePeer = Str0.substr(p0+1,p1-1);
      if ( List[ThePeer] == NULL )
      {
         TVoicePeer * mPeer = new TVoicePeer();
         List[ThePeer] = mPeer;
      }
      continue;
   }
   if ( ThePeer == "" )
   {
      cerr << "Empty PEER defenition " << endl;
      continue;
   }
   Tokenize(Str0, tokens, "=");
   if ( ! AddPair( ThePeer, tokens[0], tokens[1]) )
   {
        cerr << "An adding field fails. PeerID = " << ThePeer << " Key = " << tokens[0] << " Field " << tokens[1] << endl;
        continue;
   }
 } // while
 return true;
}

bool TDialPeers::AddPair(const string Peer, string & Key, const string Field)
{
 transform(Key.begin(), Key.end(), Key.begin(), ::tolower );
 if (Key == "target" )
 {
   List[Peer] -> TargetGW = Field ;
   return true;
 }
 if (Key == "destinationpattern" )
 {
    List[Peer] -> DestPattern = Field ;
    return true;
 }
 if (Key == "sourcepattern" )
 {
    List[Peer] -> SrcPattern = Field ;
    return true;
 }
 if (Key == "preference" )
 {
   List[Peer] -> Preference = strtoll( Field.c_str(), NULL,10) ;
   return true;
 }
 if (Key == "huntstop" )
 {
   List[Peer] -> HuntStop = strtoll( Field.c_str(), NULL,10) ;
   return true;
 }
 if (Key == "proxy" )
 {
   List[Peer] -> Proxy = strtoll( Field.c_str(), NULL,10) ;
   return true;
 }
 if (Key == "dst_translate" )
 {
   List[Peer] -> dst_translate = Field ;
   return true;
 }
 if (Key == "dst_bill_translate" )
 {
   List[Peer] -> dst_bill_translate = Field ;
   return true;
 }
 if (Key == "src_translate" )
 {
   List[Peer] -> src_translate = Field ;
   return true;
 }
 if (Key == "src_bill_translate" )
 {
   List[Peer] -> src_bill_translate = Field ;
   return true;
 }
 return false;
}

