/*
*
* versuch project
* H245 handler & thread
* $Id: h245.cxx,v 1.2 2006/05/06 13:17:54 warlock_kg Exp $
* warlock copyright since 2005 Nov
*
*/

#include "h245.h"
#include "AddrUtils.h"

#include <sstream>
#include <algorithm>

static void SendH245Mesg( H245_MultimediaSystemControlMessage & Mesg, WTCPSocket * Destination )
{
  WSocket::Address    PeerAddr;
  unsigned            PeerPort;
  Destination->GetPeerAddress( PeerAddr, PeerPort );
  PPER_Stream WriteStream;
  Mesg.Encode( WriteStream );
  WriteStream.CompleteEncoding();
  SendMsg( WriteStream, Destination );
}


void * h245_thread_function( void *ptr )
{
 struct T_h245_thread_data * thread_data;
 WTRACE(1, "H245\tEntering to h245_thread_function. ptr= " << ptr);
 thread_data = (struct T_h245_thread_data *) ptr;
 thread_data -> ch -> Main();
 WTRACE(1, "H245\tExiting from h245_thread_function");
 pthread_exit(0);
}

bool StartH245Thread(pthread_t & H245ThreadID, TGlobal & Environment, TCallPool * vCallPool, 
	unsigned CallRef, const H225_TransportAddress & Address, bool FromCaller) {
 H245Thread * ch;
 pthread_attr_t attr;
 ch = new H245Thread( Environment, vCallPool, CallRef, Address , FromCaller);
 vCallPool->H245ThreadList[CallRef] = ch;
 ch -> h245_tdata -> ch = ch;

 WTRACE(1, "H245\tStarting H245 thread");

 pthread_attr_init(&attr);
 pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
 if ( pthread_create( &H245ThreadID,  NULL, h245_thread_function, (void *) ch -> h245_tdata  ) != 0 )
 {
   cerr << endl << "Error in H245 thread creating. errno = " << errno << ". text is " << strerror(errno) << endl;
   return false;
 }
 pthread_attr_destroy(&attr);
 WTRACE(1, "H245\tThread started");
 return true;
}

H245Thread::H245Thread(TGlobal & AkaEnviron, TCallPool * vCallPool, unsigned vCallRef, const H225_TransportAddress & TransportAddress, bool FromCaller )
    : H245Handler(vCallPool->Parent), env(AkaEnviron), SessionEnded(false)
{
  Running = false; // When H245 is running
  Stop = false;    // Stop request
  if (pipe(pipes) == -1) {
    cerr << "error in H245 pipes create. errno = " << errno << endl << flush;
    Stop = true;
    return;
  }
  CallPool = vCallPool;
  CallRef = vCallRef;
  inSocket = new WTCPSocket;
  outSocket = new WTCPSocket;
  InitCaller = FromCaller;
  SetTransportAddress(TransportAddress, FromCaller);
  // Iniialize oppozit Listener
  if ( ! FromCaller ) 
  {
    WTRACE(1, "inListener init");
    if ( !inListener.Listen( (in_addr) SigThread->IncLocalAddr, env.H245PortRange->GetFreePort() ) ) {
        WTRACE( 1, "Error\tListen failed on adress " << SigThread->IncLocalAddr);
    }
  }
  else {
    WTRACE(1, "outListener init");
    if ( !outListener.Listen( (in_addr) SigThread->CallPool->OutLocalAddr[CallRef], env.H245PortRange->GetFreePort() ) ) {
        WTRACE( 1, "Error\tListen failed on adress " << SigThread->CallPool->OutLocalAddr[CallRef] );
    }
  }
  h245_tdata = ( struct T_h245_thread_data *) malloc (sizeof (struct T_h245_thread_data));
  Running = true;
}

H245Thread::~H245Thread()
{
  WTRACE(1,"H245\tThread destroy" << ", CallRef = " << CallRef);
  close (pipes[0]);
  close (pipes[1]);
  if (inSocket != NULL)
    delete inSocket;
  if (outSocket != NULL)
    delete outSocket;
  if (h245_tdata != NULL) 
  {
    WTRACE(1, "Free pointer " << h245_tdata );
     free(h245_tdata);
  }
}

void H245Thread::ReceiveMesg( WTCPSocket * From, WTCPSocket * To, bool FromCaller )
{
  WTRACE( 3, "Received H245 message from " << (FromCaller ? "Caller" : "Called"));
  BYTE tpkt[4];
  if (!From->ReadBlock( tpkt, sizeof(tpkt) ) )
  {
    // Assume that the connection has been closed....
    Close();
    return;
  }
  if ( tpkt[0] != 3 )
  {
     WTRACE( 1, "Fail TPKT in H.245" );
     return;
  }
  int BufferSize = ((tpkt[2] << 8)|tpkt[3]) - 4;
  PBYTEArray  Buffer(BufferSize);
  PPER_Stream ReadStream(Buffer);
  if ( From->ReadBlock( ReadStream.GetPointer(), BufferSize ) )
  {
    WSocket::Address    PeerAddr;
    unsigned            PeerPort;
    From->GetPeerAddress( PeerAddr, PeerPort );
    H245_MultimediaSystemControlMessage H245Mesg;
    H245Mesg.Decode( ReadStream );
    ProcessMesg( H245Mesg, To, FromCaller );
  }
}

bool H245Thread::Connecting()
{
 WTRACE(1, "H245\tH245Thread::Connecting" << ", CallRef = " << CallRef);
 int plus_one, c, inLhandle, outLhandle;
 fd_set readfds;
 struct timeval timeout;
 bool Waiting = true;
 plus_one = 0;
 while ( Waiting && ! Stop && Running) 
 {
  timeout.tv_usec = 0;
  timeout.tv_sec = 20;
  Waiting = false;
  inLhandle = inListener.GetHandler();
  outLhandle = outListener.GetHandler();
  FD_ZERO(&readfds);
  if (inLhandle > 1) 
  {
    FD_SET(inLhandle, &readfds);
    if ( plus_one < inLhandle ) 
        plus_one = inLhandle;
  }
  if (outLhandle > 1) 
  {
    FD_SET(outLhandle, &readfds);
    if ( plus_one < outLhandle )
        plus_one = outLhandle;
  }
  FD_SET(pipes[0], &readfds);
  if ( plus_one < pipes[0] )
       plus_one = pipes[0];
  if ( ::select( ++plus_one, &readfds, NULL, NULL, &timeout) == -1 ) 
  {
    WTRACE(1, "Error\tConnection select error. errno = " << errno << " string: " << strerror(errno));
    Stop = true;
    SigThread -> CallPool -> InternTC[CallRef] = TCallPool::InternalError;
    SigThread -> Stop = true;
    return false;
  }
  if ((inLhandle > 1 ) && FD_ISSET(inLhandle, &readfds)) 
     return MakeConnection( true );
  if (outLhandle > 1) {
// for debug
    if ( FD_ISSET(outLhandle, &readfds) )
      return MakeConnection( false );
  }
  if (FD_ISSET(pipes[0], &readfds) ) {
     WTRACE(1, "Pipe signal got" << ", CallRef = " << CallRef);
     Waiting = true;
     if (read(pipes[0], &c, 1) ==-1 )
         WTRACE(1, "Fail to read from pipe. errno = " << errno);
  }
 }  // While
 return false;
}

void H245Thread::Main()
{
 WTRACE(1, "H245\tH245Thread::Main" << ", CallRef = " << CallRef);
 ThreadID = pthread_self();
 int timeout = 10;
 if ( Connecting() ) {
   while ( !Stop && !SessionEnded && CallPool->cStatesList[CallRef] != TCallPool::ShuttingDownConnection
        && inSocket->IsOpen() && outSocket->IsOpen() )
   {
    int Selection = WTCPSocket::Select( *inSocket, *outSocket , timeout); 
    switch( Selection )
    {
      case -1 :   // Data available on inSocket
                  ReceiveMesg( inSocket, outSocket, true );
                  break;
      case -2 :   // Data available on outSocket
                  ReceiveMesg( outSocket, inSocket, false );
                  break;
      case -3 :   // Data available on both
                  ReceiveMesg( inSocket, outSocket, true );
                  ReceiveMesg( outSocket, inSocket, false );
                  break;
      case 0  :   // Timeout
                  break;
      default :   // Error on select...
                  WTRACE( 1, "Error " << Selection << " on H245 select : " << strerror(Selection) << " . Thread is stoping");
                  Stop = true;
                  SigThread -> CallPool -> InternTC[CallRef] = TCallPool::InternalError;
                  SigThread -> Stop = true;
                  break;
       } // switch
   } // main while
 } // if Connecting succeful
 else
 {
   WTRACE (1, "Connecting fail");
 }
 WTRACE(1, "H245\tThread closing" << ", CallRef = " << CallRef);
 WTRACE(5, "Stop = " << Stop << ", SessionEnded = " << SessionEnded << ", inSocket->IsOpen()" 
	 << inSocket->IsOpen() << ", outSocket->IsOpen() = " << outSocket->IsOpen() << ", cStatesList = " 
	 << CallPool->cStatesList[CallRef]);
 Close();
}

void H245Thread::Close()
{
 WTRACE(1, "H245\tThread closing ID= " << ThreadID << ", CallRef = " << CallRef);
 if (inListener.IsOpen() )
    inListener.Close();
 if (outListener.IsOpen() )
    outListener.Close();
 if (inSocket != NULL) 
    inSocket->Close();
 if (outSocket != NULL) {
    outSocket->Close();
 }
 Running = false;
 WTRACE(3,"H245Thread closed");
}

void H245Thread::ProcessMesg( H245_MultimediaSystemControlMessage & Mesg, WTCPSocket * Target, bool FromCaller )
{
  HandleMesg(Mesg, CallRef, FromCaller);
  if (Target != NULL ) {
    if ( Target->IsOpen() )
      SendH245Mesg( Mesg, Target );
  } // If Target is not NULL
}

bool H245Thread::MakeConnection(bool IsInListener)
{
  WSocket::Address ipAddr;
  unsigned Port;
  WTRACE(1, "H245\tStart H245 channel establish procedure from " << (IsInListener ? "IsInListener" : "OutInListener" ) << ", CallRef = " << CallRef);
  if (IsInListener) {
    if ( inSocket->Accept( inListener ) ) 
    {
      inSocket->SetReadTimeout(30);
      WTRACE(4,"H245\tinSocket dump after Accept" << endl << inSocket->OnPrint() );
      if (outListener.IsOpen()) 
      {
        WTRACE(1, "Connect with Accept on inListener and outListener is open");
        GetListenerAddress(false, ipAddr, Port);
        WTRACE(3, "H245\toutListener dump" << outListener.OnPrint());
        outListener.Close();
        WTRACE(3, "H245\toutListener dump after close" << outListener.OnPrint());
        if ( !outSocket -> Bind(ipAddr, Port) ) {
           WTRACE(1, "outSocket bind fail. errno = " << errno);
           return false;
        }
        WTRACE(3,"H245\toutSocket dump after Bind" << endl << outSocket->OnPrint() );
        AddrUtils::ConvertToIPAddress( outTransportAddress, ipAddr, Port );
        if (outSocket -> Connect(ipAddr, Port) )
        {
          WTRACE(3,"H245\toutSocket dump after Connect" << endl << outSocket->OnPrint() );
          return true;
        }
        else {
          WTRACE( 1, "H245\tError\tFailed to connect to H245 target for outSocket. errno = " 
			  << errno << " str= " << strerror(errno));
          return false;
        }
      }
      else 
      {
        WTRACE(1, "Connect with Accept on inListener and outListener is close");
        ipAddr = (in_addr) SigThread->CallPool->OutLocalAddr[CallRef];
        Port = env.H245PortRange->GetFreePort();
        outSocket -> Bind(ipAddr, Port);
        AddrUtils::ConvertToIPAddress( outTransportAddress, ipAddr, Port );
        if (outSocket -> Connect(ipAddr, Port) )
          return true;
        else
          WTRACE( 1, "H245\tError\tFailed to connect to H245 target for outSocket" );
      }
    }
    else 
    {
      WTRACE( 1, "H245\tError\tAccept failed for inSocket. errno= " << errno );     
    }
  }
  else
  {
    if ( outSocket->Accept( outListener ) )
    {
      outSocket->SetReadTimeout(30);
      WTRACE(1,"H245\toutSocket dump after Accept" << endl << outSocket->OnPrint() );
      if (inListener.IsOpen() )
      {
        WTRACE(1, "Connect with Accept on outListener and inListener is open");
        GetListenerAddress(true, ipAddr, Port);
        WTRACE(1, "H245\tinListener dump" << endl << inListener.OnPrint());
        inListener.Close();
        if (  !inSocket -> Bind(ipAddr, Port) ) {
           WTRACE(1, "inSocket bind fail. errno = " << errno);
           return false;
        }
        WTRACE(1,"H245\tinSocket dump after Bind" << endl << inSocket->OnPrint() );
        AddrUtils::ConvertToIPAddress( inTransportAddress, ipAddr, Port );
        if (inSocket -> Connect(ipAddr, Port) ) 
        {
          WTRACE(1,"H245\tinSocket dump after Connect" << endl << inSocket->OnPrint() );
          return true;
        }
        else
        {
          WTRACE( 1, "H245\tError\tFailed to connect to H245 target for inSocket. errno = " << errno );
          WTRACE( 1, "H245\tError\tDetails: peerAddr = " << ipAddr << ", Port = " << Port);
        }
      }
      else
      {
        WTRACE(1, "Connect with Accept on outListener and inListener is close");
        ipAddr = (in_addr) SigThread->IncLocalAddr;
        Port = env.H245PortRange->GetFreePort();
        inSocket -> Bind(ipAddr, Port);
        AddrUtils::ConvertToIPAddress( inTransportAddress, ipAddr, Port );
        if (inSocket -> Connect(ipAddr, Port) )
          return true;
        else
          WTRACE( 1, "H245\tError\tFailed to connect to H245 target for inSocket. errno = " << errno << " : " << strerror(errno) );
      }
    }
    else
    {
      WTRACE( 1, "H245\tError\tAccept failed for outSocket. errno= " << errno );
    }
  }
  return false;
}

void H245Thread::HandleCommand( H245_CommandMessage & Command, unsigned CallRef, bool FromCaller )
{
 H245Handler::HandleCommand( Command, CallRef, FromCaller );
 if ( Command.GetTag() == H245_CommandMessage::e_endSessionCommand )
 {
  SessionEnded = true;
 }
}

bool H245Thread::GetListenerAddress( bool IsInListener, WSocket::Address & Addr, unsigned & Port)
{
  if (IsInListener) 
  {
    if (! inListener.IsOpen() )
       return false;
    inListener.GetLocalAddress( Addr, Port );
    WTRACE(1, "Getting inListenerAddress. Addr = " << Addr << " , Port = " << Port);
  }
  else 
  {
    if (! outListener.IsOpen() )
       return false;
    outListener.GetLocalAddress( Addr, Port );
    WTRACE(1, "Getting outListenerAddress. Addr = " << Addr << " , Port = " << Port);
  }
  
  return true;
}

bool H245Thread::GetListenerAddress( bool IsInListener, H225_TransportAddress & TransAddr)
{
  WSocket::Address mAddr;
  unsigned mPort;
  if( GetListenerAddress(IsInListener, mAddr, mPort) )  
  {
    TransAddr = AddrUtils::ConvertToH225TransportAddr(mAddr, mPort);
    return true;
  }
  else 
  {
    return false;
  }
}


void H245Thread::SetTransportAddress(const H225_TransportAddress & TransportAddress, bool FromCaller)
{
  if (FromCaller) 
   inTransportAddress = TransportAddress; 
  else
   outTransportAddress = TransportAddress;
 return ;
}

void H245Thread::WaitForTermination()
{
 struct timeval timeout;
 unsigned int i, Limit;
 Limit = 200; // 20 sec
 i = 0;
 Stop = true;
 WTRACE (1, "H245\tWaitForTermiantion. Running=" << Running << ", CallRef = " << CallRef);
 while (Running && (i < Limit) ) {
  timeout.tv_sec = 0;
  timeout.tv_usec = 100000;  //    1/10 sec
  pthread_yield();
  select(0, NULL, NULL, NULL, &timeout);
  // i++;
 }
 if ( i == Limit ) 
   WTRACE(1, "Error in H245Thread::WaitForTermination waiting timeout");
 WTRACE(1, "H245\tThread Terminated" << ", CallRef = " << CallRef);
}

void H245Thread::AllocateSecondListener(bool IsInListener)
{
  if ( IsInListener )
  {
    if ( inListener.IsOpen() ) 
      return;
    WTRACE(1, "Request for inListener");
    if ( !inListener.Listen( (in_addr) SigThread->IncLocalAddr, env.H245PortRange->GetFreePort() ) ) {
        WTRACE( 1, "Error\tListen failed on adress " << SigThread->IncLocalAddr);
    }
  }
  else {
    if ( outListener.IsOpen() )
      return;
    WTRACE(1, "Request for outListener");
    if ( !outListener.Listen( (in_addr) SigThread->CallPool->OutLocalAddr[CallRef], env.H245PortRange->GetFreePort() ) ) {
        WTRACE( 1, "Error\tListen failed on adress " << SigThread->CallPool->OutLocalAddr[CallRef] );
    }
  }
  int c;
  if (write(pipes[1], &c, 1) == -1 ) {
    WTRACE(1, "Error\tWrite to pipe with handle=" << pipes[1] << " fails. errno = " << errno << " string: " << strerror(errno));
  }
}

void H245Thread::UnregCallPool()
{
 CallPool->H245ThreadList[CallRef] = NULL;
}

/////////////////////////////////////////////////////////////////////////////////////////////
////////////     H245 handler
////////////////////////////////////////////////////////////////////////////////////////////


H245Handler::H245Handler(SignallingThread * SigThread)
{
  this->SigThread = SigThread;
}

H245Handler::~H245Handler()
{
 WTRACE(3,"H245Handler destroy");
}

void H245Handler::HandleMesg( H245_MultimediaSystemControlMessage & Mesg, unsigned CallRef, bool FromCaller )
// Handle the given H.245 message
{
 switch( Mesg.GetTag() )
 {
   case H245_MultimediaSystemControlMessage::e_request     :
       HandleRequest( Mesg, CallRef, FromCaller );
       break;
   case H245_MultimediaSystemControlMessage::e_response    :
       HandleResponse( Mesg, CallRef, FromCaller );
       break;
   case H245_MultimediaSystemControlMessage::e_command     :
       HandleCommand( Mesg, CallRef, FromCaller );
       break;
   case H245_MultimediaSystemControlMessage::e_indication  :
       HandleIndication( Mesg, CallRef, FromCaller );
       break;
   default :
       WTRACE( 1, "Unknown H245 message type not handled : " << Mesg.GetTag() );
       break;
 }
 SigThread->CallPool->EstablishedConnectionCheck(CallRef);
}

void H245Handler::HandleRequest( H245_RequestMessage & Request, unsigned CallRef, bool FromCaller )
{
 WTRACE(2, "H245\tHandle Request from " << (FromCaller ? "Caller" : "Called") << " CallRef=" << CallRef);
 switch( Request.GetTag() )
 {
   case H245_RequestMessage::e_openLogicalChannel       :
      WTRACE(2, "H245\tReceiving : openLogicalChannel");
      SigThread->MessLog->Add("H245","Request_openLogicalChannel",FromCaller);
      HandleOpenLogicalChannel( Request, CallRef, FromCaller );
      break;
   case H245_RequestMessage::e_nonStandard              :
      SigThread->MessLog->Add("H245","Request_nonStandard",FromCaller);
      WTRACE(2, "H245\tReceiving : nonStandard");
      break;
   case H245_RequestMessage::e_masterSlaveDetermination :
      SigThread->MessLog->Add("H245","Request_masterSlaveDetermination",FromCaller);
      WTRACE(2, "H245\tReceiving : masterSlaveDetermination");
      SigThread->CallPool->DeterminedSet(CallRef);
      break;
   case H245_RequestMessage::e_terminalCapabilitySet    :
      WTRACE(2, "H245\tReceiving : terminalCapabilitySet");
      SigThread->MessLog->Add("H245","Request_terminalCapabilitySet",FromCaller);
      HandleTCSRequest(Request, SigThread -> env.BlockG711);
      if( FromCaller )
        SigThread->CallPool->InCapability[CallRef] = true;
      else
        SigThread->CallPool->OutCapability[CallRef] = true;
      break;
   case H245_RequestMessage::e_closeLogicalChannel      :
      SigThread->MessLog->Add("H245","Request_closeLogicalChannel",FromCaller);
      WTRACE(2, "H245\tReceiving : closeLogicalChannel");
      break;
   case H245_RequestMessage::e_requestChannelClose      :
      WTRACE(2, "H245\tReceiving : requestChannelClose");
      SigThread->MessLog->Add("H245","Request_requestChannelClose",FromCaller);
      break;
   case H245_RequestMessage::e_multiplexEntrySend       :
      SigThread->MessLog->Add("H245","Request_multiplexEntrySend",FromCaller);
      WTRACE(2, "H245\tReceiving : multiplexEntrySend");
      break;
   case H245_RequestMessage::e_requestMode              :
      WTRACE(2, "H245\tReceiving : requestMode");
      SigThread->MessLog->Add("H245","Request_requestMode",FromCaller);
      break;
   case H245_RequestMessage::e_roundTripDelayRequest    :
      WTRACE(2, "H245\tReceiving : roundTripDelayRequest");
      SigThread->MessLog->Add("H245","Request_roundTripDelayRequest",FromCaller);
      break;
   case H245_RequestMessage::e_maintenanceLoopRequest   :
      WTRACE(2, "H245\tReceiving : maintenanceLoopRequest");
      SigThread->MessLog->Add("H245","Request_roundTripDelayRequest",FromCaller);
      break;
   case H245_RequestMessage::e_communicationModeRequest :
      WTRACE(2, "H245\tReceiving : communicationModeRequest");
      SigThread->MessLog->Add("H245","Request_communicationModeRequest",FromCaller);
      break;
   case H245_RequestMessage::e_conferenceRequest        :
      WTRACE(2, "H245\tReceiving : conferenceRequest");
      SigThread->MessLog->Add("H245","Request_conferenceRequest",FromCaller);
      break;
   case H245_RequestMessage::e_multilinkRequest         :
      WTRACE(2, "H245\tReceiving : multilinkRequest");
      SigThread->MessLog->Add("H245","Request_multilinkRequest",FromCaller);
      break;
   case H245_RequestMessage::e_logicalChannelRateRequest :
      WTRACE(2, "H245\tReceiving : logicalChannelRateRequest");
      SigThread->MessLog->Add("H245","Request_logicalChannelRateRequest",FromCaller);
      break;
   default :
      WTRACE(1, "H245\tReceiving : unknown PDU " << Request.GetTag());
      SigThread->MessLog->Add("H245","Request_unknownPDU",FromCaller);
      break;
 }
}

void H245Handler::HandleResponse( H245_ResponseMessage & Response, unsigned CallRef, bool FromCaller )
{
 WTRACE(2, "H245\tHandle Response from " << (FromCaller ? "Caller" : "Called") << " CallRef=" << CallRef );
 switch( Response.GetTag() )
 {
 case H245_ResponseMessage::e_openLogicalChannelAck             :
    WTRACE(2, "H245\tReceiving : openLogicalChannelAck");
    SigThread->MessLog->Add("H245","Response_openLogicalChannelAck",FromCaller);
    HandleOpenLogicalChannelAck( Response, CallRef,  FromCaller );
    break;
 case H245_ResponseMessage::e_openLogicalChannelReject          :
    WTRACE(2, "H245\tReceiving : openLogicalChannelReject");
    SigThread->MessLog->Add("H245","Response_openLogicalChannelReject",FromCaller);
    break;
 case H245_ResponseMessage::e_nonStandard                       :
    WTRACE(2, "H245\tReceiving : nonStandard");
    SigThread->MessLog->Add("H245","Response_nonStandard",FromCaller);
    break;
 case H245_ResponseMessage::e_masterSlaveDeterminationAck       :
    WTRACE(2, "H245\tReceiving : masterSlaveDeterminationAck");
    SigThread->MessLog->Add("H245","Response_masterSlaveDeterminationAck",FromCaller);
    SigThread->CallPool->DeterminedSet(CallRef);
    break;
 case H245_ResponseMessage::e_masterSlaveDeterminationReject    :
    WTRACE(2, "H245\tReceiving : masterSlaveDeterminationReject");
    SigThread->MessLog->Add("H245","Response_masterSlaveDeterminationReject",FromCaller);
    break;
 case H245_ResponseMessage::e_terminalCapabilitySetAck          :
    WTRACE(2, "H245\tReceiving : terminalCapabilitySetAck");
    SigThread->MessLog->Add("H245","Response_terminalCapabilitySetAck",FromCaller);
    if(FromCaller)
      SigThread->CallPool->InCapability[CallRef] = true;
    else
      SigThread->CallPool->OutCapability[CallRef] = true;
    break;
 case H245_ResponseMessage::e_terminalCapabilitySetReject       :
    WTRACE(2, "H245\tReceiving : terminalCapabilitySetReject");
    SigThread->MessLog->Add("H245","Response_terminalCapabilitySetReject",FromCaller);
    break;
 case H245_ResponseMessage::e_closeLogicalChannelAck            :
    WTRACE(2, "H245\tReceiving : closeLogicalChannelAck");
    SigThread->MessLog->Add("H245","Response_closeLogicalChannelAck",FromCaller);
    break;
 case H245_ResponseMessage::e_requestChannelCloseAck            :
    WTRACE(2, "H245\tReceiving : requestChannelCloseAck");
    SigThread->MessLog->Add("H245","Response_requestChannelCloseAck",FromCaller);
    break;
 case H245_ResponseMessage::e_requestChannelCloseReject         :
    WTRACE(2, "H245\tReceiving : requestChannelCloseReject");
    SigThread->MessLog->Add("H245","Response_requestChannelCloseReject",FromCaller);
    break;
 case H245_ResponseMessage::e_multiplexEntrySendAck             :
    WTRACE(2, "H245\tReceiving : multiplexEntrySendAck");
    SigThread->MessLog->Add("H245","Response_multiplexEntrySendAck",FromCaller);
    break;
 case H245_ResponseMessage::e_multiplexEntrySendReject          :
    WTRACE(2, "H245\tReceiving : multiplexEntrySendReject");
    SigThread->MessLog->Add("H245","Response_multiplexEntrySendReject",FromCaller);
    break;
 case H245_ResponseMessage::e_requestMultiplexEntryAck          :
    WTRACE(2, "H245\tReceiving : requestMultiplexEntryAck");
    SigThread->MessLog->Add("H245","Response_requestMultiplexEntryAck",FromCaller);
    break;
 case H245_ResponseMessage::e_requestMultiplexEntryReject       :
    WTRACE(2, "H245\tReceiving : requestMultiplexEntryReject");
    SigThread->MessLog->Add("H245","Response_requestMultiplexEntryReject",FromCaller);
    break;
 case H245_ResponseMessage::e_requestModeAck                    :
    WTRACE(2, "H245\tReceiving : requestModeAck");
    SigThread->MessLog->Add("H245","Response_requestModeAck",FromCaller);
    break;
 case H245_ResponseMessage::e_requestModeReject                 :
    WTRACE(2, "H245\tReceiving : requestModeReject");
    SigThread->MessLog->Add("H245","Response_requestModeReject",FromCaller);
    break;
 case H245_ResponseMessage::e_roundTripDelayResponse            :
    WTRACE(2, "H245\tReceiving : roundTripDelayResponse");
    SigThread->MessLog->Add("H245","Response_roundTripDelayResponse",FromCaller);
    break;
 case H245_ResponseMessage::e_maintenanceLoopAck                :
    WTRACE(2, "H245\tReceiving : maintenanceLoopAck");
    SigThread->MessLog->Add("H245","Response_maintenanceLoopAck",FromCaller);
    break;
 case H245_ResponseMessage::e_maintenanceLoopReject             :
    WTRACE(2, "H245\tReceiving : maintenanceLoopReject");
    SigThread->MessLog->Add("H245","Response_maintenanceLoopReject",FromCaller);
    break;
 case H245_ResponseMessage::e_communicationModeResponse         :
    WTRACE(2, "H245\tReceiving : communicationModeResponse");
    SigThread->MessLog->Add("H245","Response_communicationModeResponse",FromCaller);
    break;
 case H245_ResponseMessage::e_conferenceResponse                :
    WTRACE(2, "H245\tReceiving : conferenceResponse");
    SigThread->MessLog->Add("H245","Response_conferenceResponse",FromCaller);
    break;
 case H245_ResponseMessage::e_multilinkResponse                 :
    WTRACE(2, "H245\tReceiving : multilinkResponse");
    SigThread->MessLog->Add("H245","Response_multilinkResponse",FromCaller);
    break;
 case H245_ResponseMessage::e_logicalChannelRateAcknowledge     :
    WTRACE(2, "H245\tReceiving : logicalChannelRateAcknowledge");
    SigThread->MessLog->Add("H245","Response_logicalChannelRateAcknowledge",FromCaller);
    break;
 case H245_ResponseMessage::e_logicalChannelRateReject          :
    WTRACE(2, "H245\tReceiving : logicalChannelRateReject");
    SigThread->MessLog->Add("H245","Response_logicalChannelRateReject",FromCaller);
    break;
 default :
    WTRACE(1, "H245\tReceiving : unknown PDU " << Response.GetTag());
    SigThread->MessLog->Add("H245","Response_uknownPDU",FromCaller);
    break;
 }
}

void H245Handler::HandleIndication( H245_IndicationMessage & Indication, unsigned CallRef, bool FromCaller )
{
 WTRACE(2, "H245\tHandle Indication from " << (FromCaller ? "Caller" : "Called") << " CallRef=" << CallRef );
 switch( Indication.GetTag() )
 {
   case H245_IndicationMessage::e_nonStandard                   :
      WTRACE(2, "H245\tReceiving : nonStandard");
      SigThread->MessLog->Add("H245","Indication_nonStandard",FromCaller);
      break;
   case H245_IndicationMessage::e_functionNotUnderstood         :
      WTRACE(2, "H245\tReceiving : functionNotUnderstood");
      SigThread->MessLog->Add("H245","Indication_functionNotUnderstood",FromCaller);
      break;
   case H245_IndicationMessage::e_masterSlaveDeterminationRelease :
      WTRACE(2, "H245\tReceiving : masterSlaveDeterminationRelease");
      SigThread->MessLog->Add("H245","Indication_masterSlaveDeterminationRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_terminalCapabilitySetRelease  :
      WTRACE(2, "H245\tReceiving : terminalCapabilitySetRelease");
      SigThread->MessLog->Add("H245","Indication_terminalCapabilitySetRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_openLogicalChannelConfirm     :
      WTRACE(2, "H245\tReceiving : openLogicalChannelConfirm");
      SigThread->MessLog->Add("H245","Indication_openLogicalChannelConfirm",FromCaller);
      break;
   case H245_IndicationMessage::e_requestChannelCloseRelease    :
      WTRACE(2, "H245\tReceiving : requestChannelCloseRelease");
      SigThread->MessLog->Add("H245","Indication_requestChannelCloseRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_multiplexEntrySendRelease     :
      WTRACE(2, "H245\tReceiving : multiplexEntrySendRelease");
      SigThread->MessLog->Add("H245","Indication_multiplexEntrySendRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_requestMultiplexEntryRelease  :
      WTRACE(2, "H245\tReceiving : requestMultiplexEntryRelease");
      SigThread->MessLog->Add("H245","Indication_requestMultiplexEntryRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_requestModeRelease            :
      WTRACE(2, "H245\tReceiving : requestModeRelease");
      SigThread->MessLog->Add("H245","Indication_requestModeRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_miscellaneousIndication       :
      WTRACE(2, "H245\tReceiving : miscellaneousIndication");
      SigThread->MessLog->Add("H245","Indication_miscellaneousIndication",FromCaller);
      break;
  case H245_IndicationMessage::e_jitterIndication              :
      WTRACE(2, "H245\tReceiving : jitterIndication");
      SigThread->MessLog->Add("H245","Indication_jitterIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_h223SkewIndication            :
      WTRACE(2, "H245\tReceiving : h223SkewIndication");
      SigThread->MessLog->Add("H245","Indication_h223SkewIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_newATMVCIndication            :
      WTRACE(2, "H245\tReceiving : newATMVCIndication");
      SigThread->MessLog->Add("H245","Indication_newATMVCIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_userInput             :
      WTRACE(2, "H245\tReceiving : userInput");
      SigThread->MessLog->Add("H245","Indication_userInput",FromCaller);
      break;
   case H245_IndicationMessage::e_h2250MaximumSkewIndication    :
      WTRACE(2, "H245\tReceiving : h2250MaximumSkewIndication");
      SigThread->MessLog->Add("H245","Indication_h2250MaximumSkewIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_mcLocationIndication          :
      WTRACE(2, "H245\tReceiving : mcLocationIndication");
      SigThread->MessLog->Add("H245","Indication_mcLocationIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_conferenceIndication          :
      WTRACE(2, "H245\tReceiving : conferenceIndication");
      SigThread->MessLog->Add("H245","Indication_conferenceIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_vendorIdentification          :
      WTRACE(2, "H245\tReceiving : vendorIdentification");
      SigThread->MessLog->Add("H245","Indication_vendorIdentification",FromCaller);
      break;
   case H245_IndicationMessage::e_functionNotSupported          :
      WTRACE(2, "H245\tReceiving : functionNotSupported");
      SigThread->MessLog->Add("H245","Indication_functionNotSupported",FromCaller);
      break;
   case H245_IndicationMessage::e_multilinkIndication           :
      WTRACE(2, "H245\tReceiving : multilinkIndication");
      SigThread->MessLog->Add("H245","Indication_multilinkIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_logicalChannelRateRelease     :
      WTRACE(2, "H245\tReceiving : logicalChannelRateRelease");
      SigThread->MessLog->Add("H245","Indication_logicalChannelRateRelease",FromCaller);
      break;
   case H245_IndicationMessage::e_flowControlIndication         :
      WTRACE(2, "H245\tReceiving : flowControlIndication");
      SigThread->MessLog->Add("H245","Indication_flowControlIndication",FromCaller);
      break;
   case H245_IndicationMessage::e_mobileMultilinkReconfigurationIndication      :
      WTRACE(2, "H245\tReceiving : mobileMultilinkReconfigurationIndication");
      SigThread->MessLog->Add("H245","Indication_",FromCaller);
      break;
   default :
      WTRACE(1, "H245\tReceiving : unknown PDU " << Indication.GetTag());
      SigThread->MessLog->Add("H245","Indication_unknownPDU",FromCaller);
      break;
 }
}

void H245Handler::HandleCommand( H245_CommandMessage & Command, unsigned CallRef, bool FromCaller )
{
 WTRACE(2, "H245\tHandle Command from " << (FromCaller ? "Caller" : "Called") << " CallRef=" << CallRef );
 switch( Command.GetTag() )
 {
   case H245_CommandMessage::e_nonStandard                      :
      WTRACE(2, "H245\tReceiving : nonStandard");
      SigThread->MessLog->Add("H245","Command_nonStandard",FromCaller);
      break;
   case H245_CommandMessage::e_maintenanceLoopOffCommand        :
      WTRACE(2, "H245\tReceiving : maintenanceLoopOffCommand");
      SigThread->MessLog->Add("H245","Command_maintenanceLoopOffCommand",FromCaller);
      break;
   case H245_CommandMessage::e_sendTerminalCapabilitySet        :
      WTRACE(2, "H245\tReceiving : sendTerminalCapabilitySet");
      SigThread->MessLog->Add("H245","Command_sendTerminalCapabilitySet",FromCaller);
      break;
   case H245_CommandMessage::e_encryptionCommand                :
      WTRACE(2, "H245\tReceiving : encryptionCommand");
      SigThread->MessLog->Add("H245","Command_encryptionCommand",FromCaller);
      break;
   case H245_CommandMessage::e_flowControlCommand               :
      WTRACE(2, "H245\tReceiving : flowControlCommand");
      SigThread->MessLog->Add("H245","Command_flowControlCommand",FromCaller);
      break;
   case H245_CommandMessage::e_endSessionCommand                :
      WTRACE(2, "H245\tReceiving : endSessionCommand");
      SigThread->MessLog->Add("H245","Command_endSessionCommand",FromCaller);
      if (SigThread->CallPool->StopTime[CallRef] == 0) {
         SigThread->CallPool->StopTime[CallRef] = time(NULL);
         WTRACE(2, "Fixing of the moment of StopTime" << PTime(SigThread->CallPool->StopTime[CallRef]) << " :  CallRef " << CallRef);
      }
      break;
   case H245_CommandMessage::e_miscellaneousCommand             :
      WTRACE(2, "H245\tReceiving : miscellaneousCommand");
      SigThread->MessLog->Add("H245","Command_miscellaneousCommand",FromCaller);
      break;
   case H245_CommandMessage::e_communicationModeCommand         :
      WTRACE(2, "H245\tReceiving : communicationModeCommand");
      SigThread->MessLog->Add("H245","Command_communicationModeCommand",FromCaller);
      break;
   case H245_CommandMessage::e_conferenceCommand                :
      WTRACE(2, "H245\tReceiving : conferenceCommand");
      SigThread->MessLog->Add("H245","Command_conferenceCommand",FromCaller);
      break;
   case H245_CommandMessage::e_h223MultiplexReconfiguration     :
      WTRACE(2, "H245\tReceiving : h223MultiplexReconfiguration");
      SigThread->MessLog->Add("H245","Command_h223MultiplexReconfiguration",FromCaller);
      break;
   case H245_CommandMessage::e_newATMVCCommand                  :
      WTRACE(2, "H245\tReceiving : newATMVCCommand");
      SigThread->MessLog->Add("H245","Command_newATMVCCommand",FromCaller);
      break;
   case H245_CommandMessage::e_mobileMultilinkReconfigurationCommand      :
      WTRACE(2, "H245\tReceiving : mobileMultilinkReconfigurationCommand");
      SigThread->MessLog->Add("H245","Command_mobileMultilinkReconfigurationCommand",FromCaller);
      break;
   default :
      WTRACE(1, "H245\tReceiving : unknown PDU " << Command.GetTag());
      SigThread->MessLog->Add("H245","Command_unknownPDU",FromCaller);
      break;
 }
}

void H245Handler::HandleOpenLogicalChannel( H245_OpenLogicalChannel & olc, unsigned CallRef, bool FromCaller )
{
 if (SigThread->CallPool->MediaProxing[CallRef] == false )
   return;
 WTRACE(2, "OpenLogicalChannel handle : CallRef=" << CallRef << ", from " << (FromCaller ? " Caller" : "Called"));
 if (FromCaller)
   WSocket::Address LocalAddr = SigThread->CallPool->OutLocalAddr[CallRef];
 else
   WSocket::Address LocalAddr = SigThread->IncLocalAddr;

 // Starting MediaThread in CallPool
 if (SigThread->CallPool->NewMediaThread(CallRef) )
 {
  int FreeUDPport;
  FreeUDPport = SigThread->env.RTPPortRange->GetFreePort();
  SigThread->CallPool->BindRTPSockets(CallRef, FreeUDPport, FreeUDPport+2, SigThread->IncLocalAddr, 
		  SigThread->CallPool->OutLocalAddr[CallRef] );
  SigThread->CallPool->BindRTCPSockets(CallRef, FreeUDPport+1, FreeUDPport+3, SigThread->IncLocalAddr, 
		  SigThread->CallPool->OutLocalAddr[CallRef] );
  WTRACE(1, "Start MediaThread. SigThread->LocalAddr " << SigThread->IncLocalAddr 
		  << " : SigThread->OutLocalAddr " << SigThread->CallPool->OutLocalAddr[CallRef] );
 }

 if ( olc.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters) )
 {
  if( olc.m_reverseLogicalChannelParameters.m_dataType.GetTag() == H245_DataType::e_audioData )
  {
   if ( ! FromCaller )
   {
    ostringstream sCodec;
    vector<string> tokens;
    H245_AudioCapability AC = olc.m_reverseLogicalChannelParameters.m_dataType;
    sCodec << AC;
    Tokenize(sCodec.str(), tokens, " ");
    SigThread->CallPool->IncCDR[CallRef].Codec = tokens[0];
   }
   if (olc.m_reverseLogicalChannelParameters.HasOptionalField(H245_OpenLogicalChannel_reverseLogicalChannelParameters::e_multiplexParameters))
   {
      H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters & ackparam =
                        olc.m_reverseLogicalChannelParameters.m_multiplexParameters;
      if (ackparam.GetTag() == H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters::e_h2250LogicalChannelParameters )
      {
         H245_H2250LogicalChannelParameters & h225Params = ackparam;
         // Control channel
         if ( h225Params.HasOptionalField( H245_H2250LogicalChannelParameters::e_mediaControlChannel ) )
         {
            H245_TransportAddress & Addr = h225Params.m_mediaControlChannel;
            WTRACE(2, "reverseMediaControlChannel adress is " << Addr);
            SigThread->CallPool->SetControlChannel(CallRef, Addr, FromCaller);
         } // if ::e_mediaControlChannel
         if ( h225Params.HasOptionalField( H245_H2250LogicalChannelParameters::e_mediaChannel ) )
         {
            H245_TransportAddress & Addr = h225Params.m_mediaChannel;
            WTRACE(2, "reverseMediaChannel adress is " << Addr);
            SigThread->CallPool->SetMediaChannel(CallRef, Addr, FromCaller);
         } // if ::e_mediaChannel
       } // if ::e_h2250LogicalChannel
   } // if ::e_multiplexParameters
  }
 }  // if ::e_reverseLogicalChannelParameters

 if(olc.m_forwardLogicalChannelParameters.m_dataType.GetTag() == H245_DataType::e_audioData)
 {
  if ( ! FromCaller )
  {
    ostringstream sCodec;
    vector<string> tokens;
    H245_AudioCapability AC = olc.m_forwardLogicalChannelParameters.m_dataType;
    sCodec << AC;
    Tokenize(sCodec.str(), tokens, " ");
    SigThread->CallPool->OutCDR[CallRef].Codec = tokens[0];
  }
  // get multiplexParameters field
   H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters & MultiParams =
        olc.m_forwardLogicalChannelParameters.m_multiplexParameters;
   if ( MultiParams.GetTag() ==
        H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::e_h2250LogicalChannelParameters )
   {
   // multiplexParameters is a h2250LogicalChannelParameters
    H245_H2250LogicalChannelParameters & H225Params = MultiParams;
    // handle the control channel if present
    if ( H225Params.HasOptionalField( H245_H2250LogicalChannelParameters::e_mediaControlChannel ) )
    {
      H245_TransportAddress & Addr = H225Params.m_mediaControlChannel;
      WTRACE(2, "forwardMediaControlChannel address is " << Addr << ":" << olc.m_forwardLogicalChannelNumber);
      SigThread->CallPool->SetControlChannel(CallRef, Addr, FromCaller);
      WTRACE(2, "forwardMediaControlChannel address after is " << Addr << ":" << olc.m_forwardLogicalChannelNumber);
    } // if media control channel address

    // handle the media channel if present
    if ( H225Params.HasOptionalField( H245_H2250LogicalChannelParameters::e_mediaChannel ) )
    {
      H245_TransportAddress & Addr = H225Params.m_mediaChannel;
      WTRACE(2, "forwardMediaChannel adress is " << Addr << ":" << olc.m_forwardLogicalChannelNumber);
      SigThread->CallPool->SetMediaChannel(CallRef, Addr, FromCaller);
    } // if media channel address
   } // if MultiplexedParametr
 }
 // Check for ready of RTP proxing
 if( !(SigThread->CallPool->IsMediaRunning(CallRef)) )  // SigThread->CalledRTPport not need
 {
    SigThread->CallPool->StartMediaThread(CallRef);
    SigThread->CallPool->SetMediaRunning(CallRef, true);
 }
}

void H245Handler::HandleOpenLogicalChannelAck( H245_OpenLogicalChannelAck & msg , unsigned CallRef, bool FromCaller )
{
 if (SigThread->CallPool->MediaProxing[CallRef] == false)
   return;
 WTRACE(2, "OpenLogicalChannelAck handle :  CallRef=" << CallRef );
 if (FromCaller)
   WSocket::Address LocalAddr = SigThread->CallPool->OutLocalAddr[CallRef];
 else
   WSocket::Address LocalAddr = SigThread->IncLocalAddr;

 // For media channels
 if ( msg.HasOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters) )
 {
   H245_OpenLogicalChannelAck_forwardMultiplexAckParameters & ackparam = msg.m_forwardMultiplexAckParameters;
   if (ackparam.GetTag() == H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters )
   {
     H245_H2250LogicalChannelAckParameters & h225Params = ackparam;
     // Control channel
     if ( h225Params.HasOptionalField( H245_H2250LogicalChannelAckParameters::e_mediaControlChannel ) )
     {
      H245_TransportAddress & Addr = h225Params.m_mediaControlChannel;
      WTRACE(2, "forwardMediaControlChannelAck adress is " << Addr << ":" << msg.m_forwardLogicalChannelNumber);
      SigThread->CallPool->SetControlChannel(CallRef, Addr, FromCaller);
      WTRACE(2, "forwardMediaControlChannelAck adress is after " << Addr << ":" << msg.m_forwardLogicalChannelNumber);
     }
     // Media channel
     if ( h225Params.HasOptionalField( H245_H2250LogicalChannelAckParameters::e_mediaChannel ) )
     {
       H245_TransportAddress & Addr = h225Params.m_mediaChannel;
       WTRACE(2, "forwardMediaChannelAck adress is " << Addr << ":" << msg.m_forwardLogicalChannelNumber
                << "\t packet from" << (FromCaller? "Caller" : "Called"));
       SigThread->CallPool->SetMediaChannel(CallRef, Addr, FromCaller);
       SigThread->CallPool->SetMediaChannelOpened(CallRef, FromCaller);
     }
   } // if e_h2250LogicalChannelAckParameters
 } // if _forwardMultiplexAckParameters
}

void H245Handler::HandleTCSRequest(H245_RequestMessage & Request, bool BlockG711)
{
 WTRACE(3, "TCS got. BlockG711 is " << (BlockG711 ? "true" : "false" ) );
 if ( ! BlockG711 )
   return;
 vector<int> cap_entry_index;
 int i, j, k;
 H245_TerminalCapabilitySet & msg = Request;
 if ( msg.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityTable) )
 {
   H245_ArrayOf_CapabilityTableEntry & Table = msg.m_capabilityTable;
   for (i =0 ; i< Table.GetSize(); i++ ) 
   {
     H245_CapabilityTableEntry & TableEntry = Table[i];
     if (TableEntry.HasOptionalField(H245_CapabilityTableEntry::e_capability) ) {
       H245_Capability & Capability = TableEntry.m_capability;
       switch (Capability.GetTag()) {
	case H245_Capability::e_receiveAudioCapability:
        case H245_Capability::e_transmitAudioCapability:
	case H245_Capability::e_receiveAndTransmitAudioCapability:
	 H245_AudioCapability & audio = Capability;
	 switch (audio.GetTag()) {
	    case H245_AudioCapability::e_g711Alaw64k:
	    case H245_AudioCapability::e_g711Alaw56k:
	    case H245_AudioCapability::e_g711Ulaw64k:
	    case H245_AudioCapability::e_g711Ulaw56k:
		WTRACE(4, "g711 Capability found = " << Capability << " entryNumber = " << TableEntry.m_capabilityTableEntryNumber);
		cap_entry_index.push_back(TableEntry.m_capabilityTableEntryNumber);
		Table.RemoveAt(i--);
		break;
	 } // switch g711
         break;
       } // switch audio
     } // if HasOptionalField e_capability
   } // for
 } // if e_capabilityTable
 if ( msg.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityDescriptors) )
 {
   H245_ArrayOf_CapabilityDescriptor & Descriptors = msg.m_capabilityDescriptors;
   for (i =0 ; i< Descriptors.GetSize(); i++ )
   {
     H245_CapabilityDescriptor & Desc = Descriptors[i];
     if (Desc.HasOptionalField(H245_CapabilityDescriptor::e_simultaneousCapabilities) )
     {
	H245_ArrayOf_AlternativeCapabilitySet & SimCap = Desc.m_simultaneousCapabilities;
        for ( j=0; j < SimCap.GetSize(); j++ )
        {
	  H245_AlternativeCapabilitySet & ACapSet =  SimCap[j];
	  for (k=0; k < ACapSet.GetSize(); k++ )
	  {
	    vector<int>::iterator itr;
	    H245_CapabilityTableEntryNumber & Num = ACapSet[k];
	    int h;
	    h = Num;
	    itr = find(cap_entry_index.begin(), cap_entry_index.end(), h);
	    if (itr != cap_entry_index.end())
	    {
		ACapSet.RemoveAt(k--);
	    }
          }
        } // for H245_ArrayOf_AlternativeCapabilitySet
     } // if Has e_simultaneousCapabilities
   } // for H245_ArrayOf_CapabilityDescriptor
 }
}
